#undef MAIN
#define USER
#include <cstdlib>
#include <iostream>
#include <fstream>
#include<iomanip>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include "const.h"
#include "struct.h"
#include "model.h"
#include "user.h"
#include "funcs.h"
#include "para.h"

#define DEBUG                // comment in to display debug information
#define JMP_JUMPOUTS         // to jump out of odeint
//#define INTERMEDIATE_STORAGE // to store intermediate results
 
// File-scope variables
const int nvar = 1;    // Number of coupled equations to integrate
const int KMAXX = 60000;
const int NMAX = nvar;
int kount;        // number of good steps taken
double dxsav;

// Storage for intermediate results (used in function odeint)
#ifdef INTERMEDIATE_STORAGE
int kmax = KMAXX; // intermediate storage exists (KMAXX steps can be stored)
double xp[KMAXX];
double yp[NMAX][KMAXX];
#else
int kmax = 0;     // no intermediate storage exists
#endif

// Current gas values (cgs)
double rho,pre,temp,mu;

extern bool get_r(float &value);
extern bool get_i(int &value);
extern bool get_n(char key[]);
extern void assign_part(struct part_list *start);
extern void rm_part(struct part *pptr);
extern void report();
extern void write_var(char item[], float value);
extern void write_var(char item[], int size, float value[]); 
extern void write_var(char item[], int value);
extern void write_var(char item[], int size, int value[]);
extern void write_var(char item[], bool value);
extern void write_var(char item[], char value[]);
extern void read_var(float *value);
extern void read_var(int *value);
extern void read_var(bool *value);
extern void read_var(char *value);
extern void del_allpart(struct cell *cptr);
extern void merge_array(float buf[], int nvalues[]);
extern struct cell *find_fcell(int &ilev, float xc[]);
extern float global_sum(float value);
extern float global_max(float value);
extern void broadcast(bool *value);
extern void broadcast(float buf[], int nvalues);
//JMP added:
extern void inc_lev();
extern void dec_lev();
extern void broadcast(float *value);
//SVL SILO stuff:
extern int write_silo(int flev, char silofile[]);
extern void sys_cmd(char key[]);
void usrqis(float xc[], float qa[]);
float usrss(float d, float p);
void usrqp(float pa[], float qa[]);
void usrpq(float qa[], float pa[]);
void usrfxq(cell *cptr);
void usrfxp(float pa[]);
bool usrps(cell *cptr, float &s);
void usrsaf(float dt, cell *cptr, float so[]);
void usrref(cell* cptr, float qref[]);
float usrqsol( cell *cptr);
float usrte(float pa[]);
void usrmu(float pa[], float mu[]);
void usrbd(int id, int bound, join *jptr, float pl[], float pr[]);
bool usrcmd(char key[]);
void usrprt();
void usrwr();
bool usrrd(char vname[]);
void usrsum();
void usrmon();
void usrdef();
void usrkeps();
void usrsetg();
bool usroff();
float usrvel(int id, struct cell *cptr);
float usrbfield(int id, struct cell *cptr);
float usrgfield(int id, struct cell *cptr);
void usrga(int id, struct cell *cptr, struct cell *left,
  struct cell *right);
void usrprompt(char prompt[]);
void usrquit();
void usrpath(char path[]);
float usrdmdt(float pmass, float pa[]);
struct part *usrmkpart(float xp[]);
float usrdmdt(float pmass, float pa[]);
void usrprate(struct part *pptr, float dt, float pa[], float dvp[],
  float &den, float &dm);
void usrpartpop(int ilev, float tcur);
bool usrpartps(char pitem[], struct part *pptr, float &s); 
//JMP defined:
float usravgmass(float pa[]);
float radiate(float pa[], float dt);
float radiate_heatcool(float pa[], float dt);
//float radiate_heatcool(float qa[], float pa[], float xc[], float dt, float &temp_init, float &temp_new);
void readin_cooling_curve();
void coolc(float t, float tmin, float logt[], float dlogt, int ntmax, 
	   float lamb_tot[], float &lambda);
int find_wind(float xc[], float d1_stag);
void orbit_calc();
void map_wind(bool init, int wnd, float xc[], float pa[], bool &update);
void read_tempmu_data();
void usrmonitor();
float usrpdtmp(float pdcell);
float usrpdmu(float pdcell);
float usrpdg(float pdcell);
float usrpdh2(float pdcell);
float usrpdhi(float pdcell);
float usrpdhii(float pdcell);
float usrtmpmu(float temp);
float usrtmpg(float temp);
float usriedg(float iedcell);
float usrteqa(float qa[]);
float usriedqa(float qa[]);
float usriedtmp(float iedcell);
float usriedmu(float iedcell);

double calc_lambda(double temp);
inline double edot(double rho, double lambda_cool, double gamma_heat);
inline double eint(double rho, double temp);
void heatcool(double x0, double* f, double* fprime);
void odeint(double ystart[], int nvar, double x1, double x2, double eps, double* h1, double hmin, int* nok, int* nbad, void (*derivs)(double, double*, double*), void (*rkqc)(double*, double*, int, double*, double, double, double*, double*, double*, void (*derivs)(double, double*, double*)));
void quit();
void rk4(double y[], double dydx[], int n, double x, double h, double yout[], void (*derivs)(double, double*, double*));
void rkqc(double y[], double dydx[], int n, double* x, double htry, double eps, double yscal[], double* hdid, double* hnext, void (*derivs)(double, double*, double*));
void str_error(string error_text);

void adiabatic_shk_jump(float mach, float& jumpr, float& jumpp);
void isothermal_shk_jump(float mach, float& jumpr, float& jumpp);
void cloud_soft_edge(float xc[], float ramb, float chi, float rhofudge, float& rho, float& col);
void setup_filament(float xc[], float fil_theta, float fil_phi, float rx0, float ry0, float rz0, float ramb, float chi, float pa[]);
void setup_postshock_pa(float pa[], float cs, float mach);
void setup_preshock_pa(float pa[], float ramb, float ambp, float& cs);

//file usrqis.c
/*
+ usrqis - classify and initialise cell, given coordinates.
     Description:
        This initialises qa in  the cell with coordinates xc according
        to cas.
*/
void usrqis(float xc[], float qa[])
{
  int i, id;
  float pa[MAXQ];
  //JMP defined:
  bool alignaxis0,alignaxis1;
  bool init,update;
  bool rzp,zrp,rtp,xyz,insidecloud;
  int n,cnt,nstars_check,nholes_check;
  int wnd;
  int col, ncol, modstep;
  float r,rs;
  float mdotwind,vinfwind,tempwind,avgmasswind;
  float xdist,ydist,zdist,avgm;
  float dsepx,dsepy,dsepz,xstag,ystag,d1_stag,rcal,drmin;
  float u0, r2, ri, ai, rc, rosq, x;
  float tv, rhos, ps, p1, p2, cs, ambp, jumpr, jumpp, vs, rbub, zbub;
  float delx, dely, delz, dist, dum, ygridwidth, zgridwidth, distproper;
  float xlc,xrc;
  float rho,pre,rhocloud,pcloud,rmaxcloud;

  //cout << "usrqis: nholes = " << nholes << "\n";
  //quit();

  //cout << "usrqis start\n";
  if (initialsetup){    //JMP initial setup
    if (tempmu){
      //Calculate avgmassamb and muamb etc...
      if (strcmp(cas, "SHELL_BLOWOUT") == 0){
        muamb = usrpdmu(pamb/ramb);        // pamb and ramb are specified so compute muamb
      }
      else if ((strcmp(cas, "CLUMP") == 0)||(strcmp(cas, "FILAMENT")) == 0){
        muamb = usrtmpmu(tamb);            //calculate mu given tamb
      }
      avgmassamb = muamb*massh;
      cout << "muamb        = " << muamb << "\n";
      cout << "avgmassamb   = " << avgmassamb << "\n";
      //cout << "tamb         = " << tamb << "\n";
    }
    else{
      //Specify default avgmassamb
      if (strcmp(cas, "WIND") == 0){
        avgmassamb = massh/(2.0*xamb + 0.75*yamb + 0.5*zamb); //assumes complete ionization!
      }
      else avgmassamb = 0.61*massh; //default
    }

    // CASES now follow.....

    if (strcmp(cas, "CWB") == 0){
      //remap1 = 0.01; //mg units
      //remap2 = 0.01;
      drmin = (xr[0]-xl[0])/float(nc[0])/pow(2.0,nlevs-1);
      //remap1 = remapcells*drmin; //mg units
      //remap2 = remapcells*drmin;
    }

    if ((strcmp(cas, "CWB") == 0)||(strcmp(cas, "WIND") == 0)){
      // Read in stellar data
      ifstream file("stars_info.dat");  
      if (!file) {
        cerr << "Failed to open file\n";
        exit(EXIT_FAILURE);
      } 
      file.seekg(0); // goto start of file
      //file >> header;
      file >> nstars_check;   // number of stars
      if (nstars_check != nstars){
        cout << "nstars does not match input file! ABORTING!\n";
        exit(EXIT_FAILURE);
      }
      else{
        cout << "nstars = " << nstars << "\n";
      }
      //Format of file is:
      //id, xpos, ypos, zpos, mass, mdot, vinf, twind, massfracx, massfracy, massfracz
      for (n = 0; n < nstars; n++){
        file >> cnt >> xpos[n] >> ypos[n] >> zpos[n] >> mass[n] >> mdot[n] >> vinf[n] >> tempw[n] >> massfracx[n] >> massfracy[n] >> massfracz[n];
      }
      file.close();
      //cout << "nstars = " << nstars << "\n";
      //cout << "mdot[0] = " << mdot[0] << "\n";
      cout << "mass[0] = " << mass[0] << "\n";
      cout << "mass[1] = " << mass[1] << "\n";
      for (n = 0; n < nstars; n++){ 
        mdot[n] *= msol/yr;
      }
    }//case CWB or WIND

    if (strcmp(cas, "CWB") == 0){
      avgmass1 = massh/(2.0*massfracx[0] + 0.75*massfracy[0] + 0.5*massfracz[0]); //assumes complete ionization!
      avgmass2 = massh/(2.0*massfracx[1] + 0.75*massfracy[1] + 0.5*massfracz[1]);
      xdist = xpos[1]-xpos[0];
      ydist = ypos[1]-ypos[0];
      zdist = zpos[1]-zpos[0];
      dsep = sqrt(xdist*xdist + ydist*ydist + zdist*zdist);
      eta = mdot[1]*vinf[1]/(mdot[0]*vinf[0]);
      rob = dsep*sqrt(eta)/(1.0 + sqrt(eta));
      //cout << "eta = " << eta << "\n";
      //cout << "rob = " << rob << "\n";
      xvel[0] = 0.0; //initialize velocities
      yvel[0] = 0.0;
      zvel[0] = 0.0;
      xvel[1] = 0.0;
      yvel[1] = 0.0;
      zvel[1] = 0.0;
      if (orbit) orbit_calc();  //update cell positions 
    }

    //Initial setup for many clouds in shock-cloud(s) simulation
    if (strcmp(cas, "MANYCLOUD") == 0){
      if (readinclouddata){
        ifstream file(cloudfilename);  
        if (!file) {
         cerr << "Failed to open file " << cloudfilename << "\n";
         exit(EXIT_FAILURE);
        } 
        file.seekg(0); // goto start of file
        //file >> header;
        file >> ncc;   // number of clouds
        cout << "ncc = " << ncc << "\n";
        for (n = 1; n <= ncc; n++){
          file >> cnt >> xc1[n] >> xc2[n] >> xc3[n] >> clr[n] >> cld[n] ;
          //>> clm[n] >> clmr[n] ;// >> dum >> dum;
          //cout << n << " " << xc1[n] << " " << xc2[n] << '\n';
        }
        file.close();
        readinclouddata = false;
      }
    }//end of cloud initial setup
	
    //Initial setup for FILAMENT simulation
    if (strcmp(cas, "FILAMENT") == 0){
      // Set the initial filament centre-of-mass coordinates
      fil_com_x = 0.0; fil_com_y = 0.0; fil_com_z = 0.0;
    }

    // Initial setup for SHELL_BLOWOUT simulation
    if (strcmp(cas, "SHELL_BLOWOUT") == 0){
      // Read hole data
      ifstream file("hole_info.dat");  
      if (!file) {
        cerr << "Failed to open file\n";
        exit(EXIT_FAILURE);
      } 
      file.seekg(0); // goto start of file
      //file >> header;
      file >> nholes_check;   // number of stars
      if (nholes_check != nholes){
        cout << "nholes does not match input file! nholes_check = " << nholes_check << "; nholes = " << nholes << "; ABORTING!\n";
        exit(EXIT_FAILURE);
      }
      else{
        cout << "nholes = " << nholes << "\n";
      }  
      // Format of file is: id, xpos, ypos, zpos, radius
      for (n = 0; n < nholes; n++){
        file >> cnt >> hole_xpos[n] >> hole_ypos[n] >> hole_zpos[n] >> hole_radius[n];
      }
      file.close();
    }

    initialsetup = false;
    //cout << "usrqis after initialsetup\n";
  }//initial setup

  xyz = (strcmp(geom, "XYZ") == 0);
  rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
  zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
  rtp = (strcmp(geom, "RTP") == 0);

  /* Defaults (safety) */
  pa[iqd] = 1.0;
  for (id = 1; id <= nd; id++)
    pa[iqu0+id] = 0.0;
  pa[iqe] = 1.0;
  for (i = iqe + 1; i < nq; i++)
    pa[i] = 0.0;
  /* Sod case */
  if (strcmp(cas, "SOD") == 0){
    pa[iqu0+1] = 0.0;
    if (xc[0] <= 0.0){
      pa[iqd] = 1.0;
      pa[iqe] = 1.0;
    }
    else{
      pa[iqd] = 0.125;
      pa[iqe] = 0.1;
    }
    if (inten)
      pa[iqie] = pa[iqe]/powf(pa[iqd], g);
  }
  else if (strcmp(cas, "SODXY") == 0){
    if (xc[0] <= 0.0){
      pa[iqd] = 1.0;
      pa[iqe] = 1.0;
    }
    else{
      pa[iqd] = 0.125;
      pa[iqe] = 0.1;
    }
  }
  else if (strcmp(cas, "SODYX") == 0){
    if (xc[1] <= 0.0){
      pa[iqd] = 1.0;
      pa[iqe] = 1.0;
    }
    else{
      pa[iqd] = 0.125;
      pa[iqe] = 0.1;
    }
  }
  else if (strcmp(cas, "CIRCLE") == 0){
    if (xc[0]*xc[0] + xc[1]*xc[1] <= 1.0){
      pa[iqd] = 1.0;
      pa[iqe] = 1.0;
      pa[iqal0+1] = 1.0;
    }
    else{
      pa[iqd] = 0.125;
      pa[iqe] = 0.1;
      pa[iqal0+1] = 0.0;
    }
  }
  else if (strcmp(cas, "SEDOV2DXY") == 0){          //JMP
    if (xc[0]*xc[0] + xc[1]*xc[1] <= 1.8687E-4){
      pa[iqd] = 1.0;
      pa[iqe] = 1135.59;
      pa[iqal0+1] = 1.0;
    }
    else{
      pa[iqd] = 1.0;
      pa[iqe] = 1.0E-5;
      pa[iqal0+1] = 0.0;
    }
  }
  /* Sam's SHOCK setup
  else if (strcmp(cas, "SHOCK") == 0){
    pa[iqd] = 1.0;
    pa[iqe] = 1.0;
    pa[iqu0+1] = 0.0;
    pa[iqu0+2] = 0.0;
    if (xc[0]*xc[0] + xc[1]*xc[1] <= 1.0){
      pa[iqd] = 1.0;
      pa[iqal0+1] = 1.0;
    }
    else
      pa[iqal0+1] = 0.0;
    if (xc[1] < -1.5){
      pa[iqd] = 5.0;
      pa[iqe] = 29.0;
      pa[iqu0+2] = 4.73282;
    }
  }
  */
  else if (strcmp(cas, "WAVE") == 0){
    pa[iqu0+1] = 0.0;
    if (xc[0] < 0.0){
      pa[iqd] = 1.0;
      pa[iqe] = 2.0;
      pa[iqal0+1] = 1.0;
    }
    else{
      pa[iqd] = 1.0;
      pa[iqe] = 1.0;
      pa[iqal0+1] = 0.0;
    }
  }
  else if (strcmp(cas, "ADVECT") == 0){
    if (xc[0] > -2.0 && xc[0] < 2.0) {
      pa[iqd] = chi*0.1;
      pa[iqe] = 1.0;
      pa[iqu0+1] = 1.0;
    }
      else {
	pa[iqd] = 0.1;
	pa[iqe] = 1.0;
	pa[iqu0+1] = 1.0;
      }
  }
  else if (strcmp(cas, "GRAVX") == 0){
    pa[iqu0+1] = 0.0;
    pa[iqal0+1] = 0.0;
    if (mabs(xc[0]) < rat)
      pa[iqd] = 1.0;
    else
      pa[iqd] = 1.0e-10;
    pa[iqe] = 1.0e-6;
  }
  else if (strcmp(cas, "GRAVXY") == 0){
    pa[iqu0+1] = 0.0;
    pa[iqu0+2] = 0.0;
    if (xc[0]*xc[0] + xc[1]*xc[1] < rat*rat){
      pa[iqd] = 1.0;
      pa[iqe] = 0.1*pa[iqd];
    }
    else{
      pa[iqd] = 1.0e-4;
      pa[iqe] = 0.1*pa[iqd];
    }
  }
  else if (strcmp(cas, "EGRAVXYZ") == 0){
    pa[iqu0+1] = 0.0;
    pa[iqu0+2] = 0.0;
    pa[iqu0+3] = 0.0;
    pa[iqe] = pfloor;
    r = sqrt(xc[0]*xc[0] + xc[1]*xc[1] + xc[2]*xc[2]);
    if (r < rat){
      pa[iqd] = 1.0;
    if (grav)
      pa[iqpot] = 4*1.0642*gkpi*(r*r/6.0 - 0.5*rat*rat)/gconst;
    }
    else{
      pa[iqd] = dfloor;
      if (grav)
	pa[iqpot] = - 4*1.0642*gkpi*rat*rat*rat/(3.0*r)/gconst;
    }
  }
  else if (strcmp(cas, "EGRAVRZ") == 0){
    pa[iqu0+1] = 0.0;
    pa[iqu0+2] = 0.0;
    pa[iqe] = pfloor;
    float r;
    r = sqrt(xc[0]*xc[0] + xc[1]*xc[1]);
    if (r < rat){
      pa[iqd] = 1.0;
    if (grav)
      pa[iqpot] = 4*1.0642*gkpi*(r*r/6.0 - 0.5*rat*rat)/gconst;
    }
    else{
      pa[iqd] = dfloor;
      if (grav)
	pa[iqpot] = - 4*1.0642*gkpi*rat*rat*rat/(3.0*r)/gconst;
    }
  }
  else if (strcmp(cas, "GRAVXYZ") == 0){
    pa[iqu0+1] = 0.0;
    pa[iqu0+2] = 0.0;
    pa[iqu0+3] = 0.0;
    pa[iqe] = pfloor;
    if (xc[0]*xc[0] + xc[1]*xc[1] + xc[2]*xc[2] < rat*rat)
      pa[iqd] = 1.0;
    else
      pa[iqd] = dfloor;
  }
  /*else if (strcmp(cas, "SNR") == 0){
    // Check there is at least 1 radial G1 cell in the ejecta
    float dr0 = lev[0]->dxbar[0]; // coarse grid cell size
    if (ej_radius/rscale/dr0 < 0.5){
      cout << "error: ejecta radius (ej_radius) must be at least as large as a G1 grid cell. Aborting!\n";
      quit();
    }
    if      (nd == 1) r = xc[0];
    else if (nd == 2) r = sqrt(xc[0]*xc[0] + xc[1]*xc[1]);
    else              r = sqrt(xc[0]*xc[0] + xc[1]*xc[1] + xc[2]*xc[2]);
    double ej_vol  = (4.0/3.0)*gkpi*pow(ej_radius,3); // cm
    double ej_dens = ej_mass/ej_vol;                  // g/cm^3
    double ej_pre  = (g - 1.0)*(ej_energy/ej_vol);    // dyn/cm^2 
    if (r <= ej_radius/rscale){
      pa[iqd] = ej_dens;
      pa[iqe] = ej_pre;
      pa[iqal0+1] = 1.0;
    }
    else{
      pa[iqd] = ramb;
      pa[iqe] = pa[iqd]*tamb*(kboltz/avgmass);
      pa[iqal0+1] = 0.0;
    }
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
  }*/
  else if (strcmp(cas, "WIND") == 0){ //JMP
    //Start by assuming that the cell is in the ambient medium. Write over 
    //with stellar wind stuff if not.
    pa[iqd] = ramb;
    pa[iqe] = pa[iqd]*tamb*(kboltz/avgmassamb);
    pa[iqu0+1] = 0.0;
    if (nd >= 2) pa[iqu0+2] = 0.0;
    if (nd >= 3) pa[iqu0+3] = 0.0;

    //If the ambient medium is a fractal cloud then set this up
    //See notes on p32 of "Multiple Colliding Winds"
    //Note: we do not have parent child info in this subroutine,
    //so in practice we need to setup the cloud in a stand-alone code
    //if (fractal){
    //}
    //Map on the stellar winds.
    //There are nstars advected scalars for the wind colours, 
    //which precede the 3 advected scalars for the mass fractions
    pa[iqal0+nstars+1] = xamb; //H mass fraction
    pa[iqal0+nstars+2] = yamb; //He mass fraction
    pa[iqal0+nstars+3] = zamb; //metals mass fraction
    for (n = 0; n < nstars; n++){
      pa[iqal0+n+1] = 0.0;         //colour
      ydist = 0.0;
      zdist = 0.0;
      //search to see if cell position is within wind remap
      xdist = xc[0]-xpos[n];
      if (nd > 1) ydist = xc[1]-ypos[n];
      if (nd > 2) zdist = xc[2]-zpos[n];
      r = sqrt(xdist*xdist + ydist*ydist + zdist*zdist);
      rs = r*rscale;
      //drmin = (xr[0]-xl[0])/float(nc[0])/pow(2.0,nlevs-1);
      /*
      if (remapcells*drmin > 1.5){
        cout << "xr[0] = " << xr[0] << "\n";
        cout << "xl[0] = " << xl[0] << "\n";
        cout << "nc[0] = " << nc[0] << "\n";
        cout << "nlevs = " << nlevs << "\n";
        cout << "drmin = " << drmin << "\n";
        exit(EXIT_FAILURE);
      }
      */
      //struct level *lptr;

      //if (r <= remapcells*drmin){
      if (r <= remap1){
      //if (r <= 0.02){
        //mdotwind = 1.0;
        //vinfwind = 1.0;
        mdotwind = mdot[n];
        vinfwind = vinf[n];
        tempwind = tempw[n];
        rho = mdotwind/(4.0*gkpi*rs*rs*vinfwind);
        //cout << "rho = " << rho << "\n";
        pa[iqd] = rho;
        //cout << "rho = " << pa[iqd] << "\n";
        //exit(EXIT_FAILURE);
                     pa[iqu0+1] = vinfwind*xdist/r;
        if (nd >= 2) pa[iqu0+2] = vinfwind*ydist/r;
        if (nd >= 3) pa[iqu0+3] = vinfwind*zdist/r;
        //pa[iqal0+1] = 1.0;         //colour
        pa[iqal0+n+1] = 1.0;         //colour
        pa[iqal0+nstars+1] = massfracx[n]; //H mass fraction
        pa[iqal0+nstars+2] = massfracy[n]; //He mass fraction
        pa[iqal0+nstars+3] = massfracz[n]; //metals mass fraction
        avgm = usravgmass(pa);
        //cout << "avgm = " << avgm;
        //exit(EXIT_FAILURE);
        //pa[iqe] = 1.0e-3*pa[iqd];
        pa[iqe] = pa[iqd]*tempwind*(kboltz/avgm);
      }
    }//loop over stars
    if (inten)
      pa[iqie] = pa[iqe]/powf(pa[iqd], g); //this is the entropy (=p/rho^g)
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    if (inten) pa[iqie] /= pscale;
    pa[iqu0+1] /= uscale;
    if (nd >= 2) pa[iqu0+2] /= uscale;
    if (nd >= 3) pa[iqu0+3] /= uscale;
  }
  /*  else if (strcmp(cas, "CWB") == 0){ //JMP
    //cout << "usrqis CWB case\n";
    //compute mtm balance surface
    rcal = sqrt(mdot[0]*vinf[0]/(mdot[1]*vinf[1])); 
    dsepx = xpos[1]-xpos[0];
    dsepy = ypos[1]-ypos[0];
    dsepz = zpos[1]-zpos[0];
    dsep = sqrt(dsepx*dsepx + dsepy*dsepy + dsepz*dsepz);
    //d1_stag = rcal*dsep/(1.0+rcal);
    d1_stag = 0.5*dsep; //use when winds blowing into constant density ambient
                        //medium
    wnd = find_wind(xc, d1_stag);//returns variable wnd
    init = true;
    update = false;
    // populate pa array of primitive variables
    //cout << "usrqis call map_wind\n";
    map_wind(init, wnd, xc, pa, update);

    //cout << "r = " << xc[0] << "\n";
    //cout << "z = " << xc[1] << "\n";
    //cout << "wnd = " << wnd << "\n";
    //cout << "rho = " << pa[iqd] << "\n";
    //cout << "r = " << xc[0] << ", z = " << xc[1] << ", wnd = " << wnd << ", rho = " << pa[iqd] << "\n";
    //exit(EXIT_FAILURE);
  }//case CWB  */       
  else if (strcmp(cas, "BESPHERE") == 0){ //JMP
    //Setup a Bonnor-Ebert sphere
    //Start with mapping on the ambient background
    pa[iqd] = ramb;
    pa[iqe] = pa[iqd]*tamb*(kboltz/avgmassamb);
    pa[iqu0+1] = 0.0;
    if (nd >= 2) pa[iqu0+2] = 0.0;
    if (nd >= 3) pa[iqu0+3] = 0.0;
    pa[iqal0+1] = 0.0; //colour
    pa[iqal0+2] = xamb; //H mass fraction
    pa[iqal0+3] = yamb; //He mass fraction
    pa[iqal0+4] = zamb; //metals mass fraction
    //vt = sqrt(kboltz*tamb/avgmassamb);    //isothermal speed
    pamb = ramb*(kboltz/avgmassamb)*tamb; //ambient pressure
    //gconst3 = pow(gconstcgs,3);
    //mbe = cbe*pow(vt,4)/sqrt(pamb*gconst3); //maximum mass of a hydrostatic
                                              //Bonnor-Ebert sphere (cgs)
    //cloudmass = be_massfrac*mbe;          //mass of cloud (cgs)
    xdist = xc[0];
    ydist = 0.0;
    zdist = 0.0;
    if (nd > 1) ydist = xc[1];
    if (nd > 2) zdist = xc[2];
    r = sqrt(xdist*xdist + ydist*ydist + zdist*zdist);
    rhocloud = 10.0*ramb;                     //cloud central density
    pcloud = rhocloud*(kboltz/avgmassamb)*tamb; //cloud central pressure
    rmaxcloud = rcloud*(sqrt(pcloud/pamb) - 1.0);//max cloud radius (mg units)
    //cout << rcloud << " " << pcloud << " " << pamb << "\n";
    //cout << "r = " << r << ", rmaxcloud = " << rmaxcloud << "\n";
    //exit(EXIT_FAILURE);
    if (r < rmaxcloud){
      rho = rhocloud/pow(1.0 + r/rcloud,2); // density of cloud (cgs)
      pre = rho*(kboltz/avgmassamb)*tamb;   // pressure of cloud (cgs)
      pa[iqd] = rho;
      pa[iqe] = pre;
      pa[iqal0+1] = 1.0; //colour
    }
    if (inten)
      pa[iqie] = pa[iqe]/powf(pa[iqd], g); //this is the entropy (=p/rho$(gamma-1))
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    if (inten) pa[iqie] /= pscale;
    pa[iqu0+1] /= uscale;
    if (nd >= 2) pa[iqu0+2] /= uscale;
    if (nd >= 3) pa[iqu0+3] /= uscale;
  }//case BESPHERE    

  else if (strcmp(cas, "SINGLECLOUD") == 0){
    float col = 0.0; //cloud colour
    float v_cloud; //cloud velocity
    //ambp = ramb*tamb*(kboltz/massh); //use massh for radiative runs 
    ambp = ramb*tamb*(kboltz/avgmass); //use avgmass for adiabatic runs 
                                     
    // Setup pre-shock quantities 
    setup_preshock_pa(pa,ramb,ambp,cs);  

    // Setup cloud 
    //cloud_soft_edge(xc,ramb,chi,rhofudge,rho,col); //comment out if cloud does not have soft edges 
      //pa[iqd] = rho;
      //pa[iqal0+1] = col;

    //sharp edged setup: first determine if cell is in cloud, then set cell parameters equal to cloud parameters 
    float r0 = 0.0; float z0 = 0.0; //coords of centre of cloud
    float dist = 0.0; //radial distance from cloud centre to cell 
    float distr = 0.0; float distz = 0.0; //r and z distance of cell from cloud 
      distr = xc[0] - r0; 
      distz = xc[1] - z0; 
        dist = sqrt(distr*distr + distz*distz); 
     

    if (dist <= rcloud){ //cell is in the cloud 
      rho = ramb*chi; //density of cell is the density of the cloud 
      v_cloud = 0.0; //cloud not moving so cell velocity is zero 
      col = 1.0; //colour of cell - different to medium 
      pa[iqd] = rho;
      pa[iqu0+2] = v_cloud;
      pa[iqal0+1] = col; //colour of wind 
    }  
    else{  // Setup flow outside of cloud...
    //map on wind (i.e. post-shock flow) for ALL cells, then set up postshock quantities behind the wind boundary
    if (xc[1] > 1.0){ //position of shock (COMMENT OUT if using a wind)
    //if (xc[0] < -10.0){ //use xc[0] for xyz geometry
      setup_postshock_pa(pa,cs,mach);
      //pa[iqe] /= 10.0; // reduce post-shock (wind) pressure to mimic a higher Mach number wind...(COMMENT OUT IF USING VELOCITY)
      //float vb = mach*cs/uscale; 
      //pa[iqu0+2] = sqrt(1000.0)*(0.74*vb); //change the post-shock (wind) velocity (COMMENT OUT IF REDUCING PRESSURE)
      //pa[iqu0+2] *= sqrt(1000.0); //change the post-shock (wind) velocity (COMMENT OUT IF REDUCING PRESSURE)
    } //COMMENT OUT if using a wind
   }


   //rescale 
    pa[iqd]    /= rhoscale; 
    pa[iqe]    /= pscale; 
    pa[iqu0+1] /= uscale; 
    pa[iqu0+2] /= uscale; 
    if (nd == 3){ 
      pa[iqu0+3] /= uscale; 
    } 

  //float vb = mach*cs/uscale;               //speed of shock in medium (mg_gt units)
  //float tcc = sqrt(chi)/vb;                //cloud crushing timescale (radius=1)
  //float machsq = mach*mach; 
  //float gp1 = g+1.0; 
  //float gm1 = g-1.0; 
  //float jumpr = gp1*machsq/(gm1*machsq + 2.0); // density jump
  //float vps = pa[iqu0+1]/vb; //post-shock flow velocity
  //float M_wind = ;

  //cout << "shock speed = " << vb << ", tcc = " << tcc << ", density jump = " << jumpr << ", vps = " << vps << "\n"; 
  //cout << "M_wind = " << NEED CALCULATION OF M USING THE PRESSURE << "\n";
  //quit();

  } //end of SINGLECLOUD

    
    /*
    //Superbubble setup (hot interior pushing isothermal shell)
    //pressure and velocity within interior match that of isothermal shell
    isothermal_shk_jump(mach,jumpr,jumpp);  //postshock quantities for isothermal shocks

    //rbub = 15.0;
    //zbub = 3.0 + rbub; //centre of bubble
    //rc = sqrt(xc[0]*xc[0] +(xc[1]-zbub)*(xc[1]-zbub));//dist from bubble centre
    //if (rc < rbub){            //interior of superbubble
    if (((rzp) && (xc[1] > 7.0)) || ((xyz) && (xc[0] < -7.0))){
                //interior of superbubble (>4 for 2pc,
                //                         >7 for 8pc,
                //                         >19 for 32pc)
      pa[iqd] = ramb/100.0; //set appropriately low density
      pa[iqe] = jumpp*ambp; 
      //planar shell
      if (rzp) {
	//        pa[iqu0+2] = -1.21*(1.0 - 1.0/jumpr)*mach*cs; //planar shell
        pa[iqu0+2] = -(1.0 - 1.0/jumpr)*mach*cs; //planar shell
      }
      else if (xyz) {
        //pa[iqu0+1] = 1.21*(1.0 - 1.0/jumpr)*mach*cs; //planar shell
        pa[iqu0+1] = (1.0 - 1.0/jumpr)*mach*cs; //planar shell
      }
      //pa[iqu0+1] = (xc[0]/rbub)*(1.0 - 1.0/jumpr)*mach*cs; //spherical shell
      //pa[iqu0+2] = ((xc[1]-zbub)/rbub)*(1.0 - 1.0/jumpr)*mach*cs; 
      pa[iqal0+1] = 0.0;  // colour
    }
    //else if (rc < rbub+1.0){       //isothermal shell
    else if (((rzp) && (xc[1] > 3.0)) || ((xyz) && (xc[0] < -3.0))){ 
      //isothermal shell (leading edge)
      pa[iqd] = jumpr*ramb;
      //pa[iqd] = jumpr*ramb*tamb/tmin;//cooling to T=tmin
      pa[iqe] = jumpp*ambp; 
      if (rzp) {
         pa[iqu0+2] = -(1.0 - 1.0/jumpr)*mach*cs;  //planar shell
      }
      else if (xyz) {
         pa[iqu0+1] = (1.0 - 1.0/jumpr)*mach*cs;  //planar shell
      }
      //pa[iqu0+1] = (xc[0]/rbub)*(1.0 - 1.0/jumpr)*mach*cs; //spherical shell
      //pa[iqu0+2] = ((xc[1]-zbub)/rbub)*(1.0 - 1.0/jumpr)*mach*cs; 
      pa[iqal0+1] = 0.0;  //colour
    }
    */     


    //mtm driven bubble (ram pressure on back of shell must equal the shell 
    //thermal pressure)
    //pa[iqe] = pa[iqd]*tamb*(kboltz/avgmass);
    //isothermal_shk_jump(mach,jumpr,jumpp);  //postshock quantities for isothermal shocks
    //vs = -(1.0 - 1.0/jumpr)*mach*cs;//see Dyson & Williams book Eq 6.83
    //pa[iqu0+2] = vs - sqrt(ps/pa[iqd]);  //pushed shell
    //pa[iqu0+2] = vs;                     //coasting shell 


    
  else if (strcmp(cas, "SLAB") == 0){
    //RA: change to make a uniform density slab of "radius" = 200
    float col = 0.0;
    //ambp = ramb*tamb*(kboltz/massh); //use massh for radiative runs
    ambp = ramb*tamb*(kboltz/avgmass); //use avgmass for adiabatic runs
                                       //(this was also used for all of the sbs
                                       //runs)
    // Setup pre-shock quantities
    setup_preshock_pa(pa,ramb,ambp,cs);

    // Setup cloud
    cloud_soft_edge(xc,ramb,chi,rhofudge,rho,col);
    pa[iqd] = rho;
    pa[iqal0+1] = col;

    //***********************************************************
    //** NOTE!!!!
    //** If gamma has a stupid value, and, for instance the postshock
    //** density is non-sensible, then probably gamma has been overwritten!
    //** Check that the file name for the stat files is not too long!!
    //**
    //************************************************************
        
    // Set postshock quantities behind shock (ADIABATIC *OR* ISOTHERMAL)
    //if (xc[1] > 3.0){//3.5 for p2=1 runs, 3.0 for all other runs
    if (xc[0] < -2.0){//3.5 for p2=1 runs, 3.0 for all other runs
      setup_postshock_pa(pa,cs,mach);
    }
     
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    pa[iqu0+1] /= uscale;
    pa[iqu0+2] /= uscale;
  }//slab setup
  else if (strcmp(cas, "MANYCLOUD") == 0){
    //manycloud with distribution ending at xcmin
    //Setup default ambient medium
    if (nal > 1){ 
      pa[iqal0+1] = 1.0; //set ambient scalar to 1.0
      for (n = 2; n <= nal; n++){
        pa[iqal0+n] = 0.0;  //set cloud scalar to 0.0
      }
    }
    else pa[iqal0+1] = 0.0;         //set ambient scalar to 0.0
    ambp = ramb*tamb*(kboltz/avgmass); //use avgmass for adiabatic runs
                                       //(this was also used for all of the sbs
                                       //runs)
    // Setup pre-shock quantities
    setup_preshock_pa(pa,ramb,ambp,cs);

    //Map on shock   
    if (xc[0] < xcmin-3.0){ // xcmin = -10 (-18) for r_clumpyregion = 10 (18)
      setup_postshock_pa(pa,cs,mach);
    }
    
    //sharp setup
    //Loop through cloud list, seeing if the current cell falls within one
    /*for (n = 1; n <= ncc; n++){
      insidecloud = false;
      delx = xc[0]-xc1[n];   //xc = cell coord, xc1 = x-pos of cloud
      dely = xc[1]-xc2[n];
      if (nd == 2){
        dist = sqrt(delx*delx + dely*dely);//distance of cell from clump center
      }
      else if (nd == 3){
        delz = xc[2]-xc3[n];
        dist = sqrt(delx*delx + dely*dely + delz*delz);
      }
      if (dist <= clr[n]) insidecloud = true;
      //If periodic, also check whether the cloud could appear the other side
      //of the grid
      if (bcl[1] == bcper){//periodic boundary on 2nd axis
        ygridwidth = xr[1]-xl[1]; //width of grid in y-direction
	if (nd == 2){
          dely = xc[1]-ygridwidth-xc2[n]; //first check if it is just off -ymin
          dist = sqrt(delx*delx + dely*dely);
          if (dist <= clr[n]) insidecloud = true;
          dely = xc[1]+ygridwidth-xc2[n]; //now check if it is just off +ymax
          dist = sqrt(delx*delx + dely*dely);
          if (dist <= clr[n]) insidecloud = true;
        }
        else if (nd == 3){
          dely = xc[1]-ygridwidth-xc2[n]; //first check if it is just off -ymin
          dist = sqrt(delx*delx + dely*dely + delz*delz);
          if (dist <= clr[n]) insidecloud = true;
          dely = xc[1]+ygridwidth-xc2[n]; //now check if it is just off +ymax
          dist = sqrt(delx*delx + dely*dely + delz*delz);
          if (dist <= clr[n]) insidecloud = true;
	}
      }
      if ((nd == 3)&&(bcl[2] == bcper)){//periodic boundary on 3rd axis
        zgridwidth = xr[2]-xl[2]; //width of grid in z-direction
        delz = xc[2]-zgridwidth-xc3[n]; //first check if it is just off -zmin
        dist = sqrt(delx*delx + dely*dely + delz*delz);
        if (dist <= clr[n]) insidecloud = true;
        delz = xc[2]+zgridwidth-xc3[n]; //now check if it is just off +zmax
        dist = sqrt(delx*delx + dely*dely + delz*delz);
        if (dist <= clr[n]) insidecloud = true;
      }
      if (insidecloud){
        pa[iqd] = cld[n]*ramb;
        if (nal == 1) pa[iqal0+1] = 1.0;
        else if (nal > 1){
          //set different colours to trace different clouds or regions of
          //clouds
          //First set the colour for separate regions of clouds

          
          xlc = xcmin;
          xrc = xcmax;
          col = int(4.0*(xc1[n]-xlc)/(xrc-xlc)) + 2;
          pa[iqal0+col] = 1.0;  //set "block" colour for cloud
          for (ncol = 1; ncol <= nal; ncol++){
            if (ncol != col) pa[iqal0+ncol] = 0.0; //set other colours to zero
          }
          
          //Then pick single clouds within each region to have a 
          //different colour. The cloud list must be re-ordered so that
          //the clouds are in order on their x-position. We can
          //then simply pick roughly the middle cloud in each "block" using the
          //modulus operator (%) - see Sec 3.1.4 of C++ book. This will need
          //a minimum of 5 clouds per block or so to work correctly.
          modstep = ncc/4; //divide number of clouds by number of cloud blocks
          if (n%modstep == modstep/2) pa[iqal0+2+4+n/modstep] = 1.0;
	  }       */
    //shallow setup, RA
   for (n = 1; n <= ncc; n++){
      insidecloud = false;
      delx = xc[0]-xc1[n];   //xc = cell coord, xc1 = x-pos of cloud
      dely = xc[1]-xc2[n];
      if (nd == 2){
        dist = sqrt(delx*delx + dely*dely);//distance of cell from clump center
	distproper = dist;
      }
      else if (nd == 3){
        delz = xc[2]-xc3[n];
        dist = sqrt(delx*delx + dely*dely + delz*delz); //dist is the r^2.
	distproper = dist;
      }
      if (dist <= 2.5*clr[n]) insidecloud = true;  //changed to 2.0 to allow max radius 2.0, in reality, radius is sqrt(2) because of cld1.
      //but overwrites. Cloud centres have to be separated by min 4!
      //If periodic, also check whether the cloud could appear on the other side
      //of the grid
      if (bcl[1] == bcper){//periodic boundary on 2nd axis
        ygridwidth = xr[1]-xl[1]; //width of grid in y-direction
	if (nd == 2){
          dely = xc[1]-ygridwidth-xc2[n]; //first check if it is just off -ymin
          dist = sqrt(delx*delx + dely*dely);
	  distproper = min(distproper, dist);
          if (dist <= 2.0*clr[n]) insidecloud = true;
          dely = xc[1]+ygridwidth-xc2[n]; //now check if it is just off +ymax
          dist = sqrt(delx*delx + dely*dely);
	  distproper = min(distproper, dist);
          if (dist <= 2.0*clr[n]) insidecloud = true;
        }
        else if (nd == 3){
          dely = xc[1]-ygridwidth-xc2[n]; //first check if it is just off -ymin
          dist = sqrt(delx*delx + dely*dely + delz*delz);
	  distproper = min(distproper, dist);
          if (dist <= 2.0*clr[n]) insidecloud = true;
          dely = xc[1]+ygridwidth-xc2[n]; //now check if it is just off +ymax
          dist = sqrt(delx*delx + dely*dely + delz*delz);
	  distproper = min(distproper, dist);
          if (dist <= 2.0*clr[n]) insidecloud = true;
	}
      }
      if ((nd == 3)&&(bcl[2] == bcper)){//periodic boundary on 3rd axis
        zgridwidth = xr[2]-xl[2]; //width of grid in z-direction
        delz = xc[2]-zgridwidth-xc3[n]; //first check if it is just off -zmin
        dist = sqrt(delx*delx + dely*dely + delz*delz);
	distproper = min(distproper, dist);
        if (dist <= 2.0*clr[n]) insidecloud = true;
        delz = xc[2]+zgridwidth-xc3[n]; //now check if it is just off +zmax
        dist = sqrt(delx*delx + dely*dely + delz*delz);
	distproper = min(distproper, dist);
        if (dist <= 2.0*clr[n]) insidecloud = true;
      }
      if (insidecloud){
	//determine proper distance first.
	dist = distproper;	
	tv = dist - clr[n]; 
	//tv = 1.0e3;
	tv = exp((double)min(cld1,cld2*tv));  //tv = exp(min(cld1 (~20),cld2(~10)*(r^2-1))  (0; 0.5G)
	tv = 0.5*(1.0 + (tv - 1.0)/(tv + 1.0));  //tv - (0;1)
	//	cout << "x = " << xc[0] << " y = " <<  xc[1] << " z = " << xc[2] << endl;
	//	cout << "current pa[iqd] = " << pa[iqd] << " new = " << cld[n]*ramb + (ramb - cld[n]*ramb)*tv 
	//	     << " supposed = " <<  max(pa[iqd], cld[n]*ramb + (ramb - cld[n]*ramb)*tv);
	pa[iqd] = max(pa[iqd], cld[n]*ramb + (ramb - cld[n]*ramb)*tv); 
	//	cout << "Real = " << pa[iqd] << endl;
        if (nal == 1) pa[iqal0+1] = 1.0;
	  else if (nal > 1){
          //set different colours to trace different clouds or regions of
          //clouds
          //First set the colour for separate regions of clouds
	  //note: colour do not follow shallow profile.
          xlc = xcmin;
          xrc = xcmax;
	  if (xcmin != xcmax) col = int(4.0*(xc1[n]-xlc)/(xrc-xlc)) + 2;  //proper way
	  else { 
	    col = int(4*n/ncc) + 2;
	    if (col == 6) col--;
	  }
	  pa[iqal0+col] = 1.0;  //set "block" colour for cloud
          for (ncol = 1; ncol <= nal; ncol++){
            if (ncol != col) pa[iqal0+ncol] = 0.0; //set other colours to zero
          }
          
          //Then pick single clouds within each region to have a 
          //different colour. The cloud list must be re-ordered so that
          //the clouds are in order on their x-position. We can
          //then simply pick roughly the middle cloud in each "block" using the
          //modulus operator (%) - see Sec 3.1.4 of C++ book. This will need
          //a minimum of 5 clouds per block or so to work correctly.
	  if (!cust_al[0]) {
	    modstep = ncc/4; //divide number of clouds by number of cloud blocks
	    if (n%modstep == modstep/2) pa[iqal0+2+4+n/modstep] = 1.0;
	  }
	  else {
	    for (int a=0; a < 4; a++) {
	      if ( n  == cust_al[a] ) pa[iqal0+2+4+a] =  1.0;
	    }
	  }   
	  }
        if (keps){
          pa[iqk] = kmin; //kmin
          pa[iqeps] = epsmin;  //epsmin
        }
	 break;   //removed to stop overwrite. bring it back for performance.
      }//inside cloud
    }
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    pa[iqu0+1] /= uscale;
    pa[iqu0+2] /= uscale;
  }//multiple cloud setup
  else if (strcmp(cas, "SHOCK") == 0){
    float col = 0.0;
    //ambp = ramb*tamb*(kboltz/massh); //use massh for radiative runs
    ambp = ramb*tamb*(kboltz/avgmass); //use avgmass for adiabatic runs (this was also used for all of the sbs runs)
    // Setup pre-shock quantities
    setup_preshock_pa(pa,ramb,ambp,cs);

    // Set postshock quantities behind shock (ADIABATIC *OR* ISOTHERMAL)
    //if (xc[1] > 3.0){//3.5 for p2=1 runs, 3.0 for all other runs
    if (xc[0] < -2.0){//3.5 for p2=1 runs, 3.0 for all other runs
      setup_postshock_pa(pa,cs,mach);
    }
    
    /*
    //Superbubble setup (hot interior pushing isothermal shell)
    //pressure and velocity within interior match that of isothermal shell
    //isothermal compression (here mach is the adiabatic mach number, 
    //M_A - the compression is also = (M_0)^2, where M_O is the 
    //isothermal mach number (M_O^2 = g M_A^2)
    isothermal_shk_jump(mach,jumpr,jumpp);  //postshock quantities for isothermal shocks
    if (xc[0] < 2.0){            //interior of superbubble
      pa[iqd] = ramb/100.0; //set appropriately low density
      pa[iqe] = jumpp*ambp; 
      pa[iqu0+1] = 1.21*(1.0 - 1.0/jumpr)*mach*cs; //planar shell
      pa[iqal0+1] = 0.0;
    }
    else if (xc[0] < 3.0){       //isothermal shell (leading edge)
      pa[iqd] = jumpr*ramb;
      //pa[iqd] = jump*ramb*tamb/tmin;//cooling to T=tmin
      pa[iqe] = jumpp*ambp; 
      pa[iqu0+1] = (1.0 - 1.0/jumpr)*mach*cs;  //planar shell
      pa[iqal0+1] = 0.0;
    }
    */

    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    pa[iqu0+1] /= uscale;
  }//end of SHOCK case
  else if (strcmp(cas, "SHOCK1D") == 0){
    ambp = ramb*tamb*(kboltz/avgmass); //use avgmass for adiabatic runs (this was also used for all of the sbs runs)
    // Setup pre-shock quantities
    setup_preshock_pa(pa,ramb,ambp,cs);

    // Set postshock quantities behind shock (ADIABATIC *OR* ISOTHERMAL)
    if (xc[0] < -5.0){
      setup_postshock_pa(pa,cs,mach);
    }
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    pa[iqu0+1] /= uscale;
    //cout << "Postscale: pa[iqe] = " << pa[iqe] << "; pscale = " << pscale << "\n"; exit(EXIT_FAILURE);
  }
  else if (strcmp(cas, "FILAMENT") == 0){
    //ambp = ramb*tamb*(kboltz/massh); //use massh for radiative runs
    ambp = ramb*tamb*(kboltz/avgmass); //use avgmass for adiabatic runs (this was also used for all of the sbs runs)
    // Setup pre-shock quantities
    setup_preshock_pa(pa,ramb,ambp,cs);

    // Setup filament
    float rx0 = 0.0; float ry0 = 0.0; float rz0 = 0.0; // coords of centre of filament
    setup_filament(xc,fil_theta,fil_phi,rx0,ry0,rz0,ramb,chi,pa);

    // Set postshock quantities behind shock (ADIABATIC *OR* ISOTHERMAL)
    if (xc[0] < -10.0){
      setup_postshock_pa(pa,cs,mach);
    }

    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    pa[iqu0+1] /= uscale;
  }
  else if (strcmp(cas, "LEAKYSHELL") == 0){
    ambp = ramb*tamb*(kboltz/avgmass); //use avgmass for adiabatic runs
    pa[iqd] = ramb;
    pa[iqe] = ambp;
    cs = sqrt(g*kboltz*tamb/avgmass);  //use ionized mu for sound/shock speed

    if (((rzp) && (xc[0] < holedr)) || ((xyz) && (mabs(xc[1]) < holedr))){
      // void in ambient medium - has the same initial pressure, but this soon
      // radiates and creates a collapse
      pa[iqd] *= 0.5;
    } 

    //***********************************************************
    //** NOTE!!!!
    //** If gamma has a stupid value, and, for instance the postshock
    //** density is non-sensible, then probably gamma has been overwritten!
    //** Check that the file name for the stat files is not too long!!
    //**
    //************************************************************
            
    //Superbubble setup (hot interior pushing isothermal shell)
    //pressure and velocity within interior match that of isothermal shell
    isothermal_shk_jump(mach,jumpr,jumpp);  //postshock quantities for isothermal shocks
    //rbub = 15.0;
    //zbub = 3.0 + rbub; //centre of bubble
    //rc = sqrt(xc[0]*xc[0] +(xc[1]-zbub)*(xc[1]-zbub));//dist from bubble centre
    //if (rc < rbub){            //interior of superbubble
    if (((rzp) && (xc[1] < -shelldz)) || ((xyz) && (xc[0] < -shelldz))){
                //interior of superbubble (>4 for 2pc,
                //                         >7 for 8pc,
                //                         >19 for 32pc)
      pa[iqd] = ramb/100.0; //set appropriately low density
      pa[iqe] = jumpp*ambp; 
      //planar shell
      if      (rzp) pa[iqu0+2] = (1.0 - 1.0/jumpr)*mach*cs; //planar shell
      else if (xyz) pa[iqu0+1] = (1.0 - 1.0/jumpr)*mach*cs; //planar shell
      //pa[iqu0+1] = (xc[0]/rbub)*(1.0 - 1.0/jumpr)*mach*cs; //spherical shell
      //pa[iqu0+2] = ((xc[1]-zbub)/rbub)*(1.0 - 1.0/jumpr)*mach*cs; 
      pa[iqal0+1] = 0.0;  // colour
    }
    //else if (rc < rbub+1.0){       //isothermal shell
    else if (((rzp) && (xc[1] < 0.0)) || ((xyz) && (xc[0] < 0.0))){
      //isothermal shell (leading edge)
      pa[iqd] = jumpr*ramb;
      //pa[iqd] = jumpr*ramb*tamb/tmin;//cooling to T=tmin
      pa[iqe] = jumpp*ambp; 
      if      (rzp) pa[iqu0+2] = (1.0 - 1.0/jumpr)*mach*cs;  //planar shell
      else if (xyz) pa[iqu0+1] = (1.0 - 1.0/jumpr)*mach*cs;  //planar shell
      //pa[iqu0+1] = (xc[0]/rbub)*(1.0 - 1.0/jumpr)*mach*cs; //spherical shell
      //pa[iqu0+2] = ((xc[1]-zbub)/rbub)*(1.0 - 1.0/jumpr)*mach*cs; 
      pa[iqal0+1] = 0.0;  //colour
      if (((rzp) && (xc[0] < holedr)) || ((xyz) && (mabs(xc[1]) < holedr))){
        // hole in shell - has the same velocity and pressure as the shell, but 
        // keeps the ambient density
        pa[iqd] = ramb;
      } 
    }
    // Change reference frame to that of the shell
    if (rzp)      pa[iqu0+2] -= (1.0 - 1.0/jumpr)*mach*cs;  //planar shell
    else if (xyz) pa[iqu0+1] -= (1.0 - 1.0/jumpr)*mach*cs;  //planar shell
  
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    pa[iqu0+1] /= uscale;
    pa[iqu0+2] /= uscale;
    //set entropy only after rescaling!!
    if (inten) pa[iqie] = pa[iqe]/powf(pa[iqd], g); //entropy =p/rho^g)
  }// leaky shell setup
  else if (strcmp(cas, "SHELL_BLOWOUT") == 0){
    // Setup ambient medium (default for all cells)
    tamb = (pamb/ramb)*(avgmassamb/kboltz); 
    setup_preshock_pa(pa,ramb,pamb,cs);

    // Calculate hot bubble parameters (in pressure equilibrium with ambient medium)
    rbub = ramb/100.0;                     // set appropriately low density in the hot bubble
    float mu_bub = (pamb/rbub);            // mu in hot bubble
    float avgmass_bub = mu_bub*massh;      // avgmass of gas in hot bubble
    float tbub = (pamb/rbub)*(avgmass_bub/kboltz); // temperature in hot bubble
    cs = sqrt(g*kboltz*tbub/avgmass_bub);  // sound speed in hot bubble

    if (((rzp) && (xc[1] < shockpos)) || ((xyz) && (xc[0] < shockpos))){
      pa[iqd] = rbub;
      //Setup shock running through hot bubble using postshock quantities for adiabatic shocks
      setup_postshock_pa(pa,cs,mach);
      pa[iqal0+1] = 1.0;  //colour
    }
    else if (((rzp) && (xc[1] < -shelldz)) || ((xyz) && (xc[0] < -shelldz))){
      // interior of superbubble
      pa[iqd] = rbub; 
    }
    else if (((rzp) && (xc[1] < 0.0)) || ((xyz) && (xc[0] < 0.0))){
      //isothermal shell
      pa[iqd] = rshell;
      // Create shell hole(s)
      if (nholes == 1){
        if (((rzp) && (xc[0] < holedr)) || ((xyz) && (mabs(xc[1]) < holedr))){
          // hole in shell - density is that of the ambient medium
          pa[iqd] = ramb;
        }
        else{
          if (nal >= 2) pa[iqal0+2] = 1.0;  //colour
        } 
      }
      else{
        // More than one hole - loop over holes (geometry is assumed to be xyz cartesian)
        if (nal >= 2) pa[iqal0+2] = 1.0;        // default colour
        for (int n = 0; n < nholes; n++){
          if (nd == 2){
            float lft_edge = hole_ypos[n] - hole_radius[n];
            float rgh_edge = hole_ypos[n] + hole_radius[n];
            if ((lft_edge < xc[1])&&(rgh_edge > xc[1])){// in hole
              pa[iqd] = ramb;
              if (nal >= 2) pa[iqal0+2] = 0.0;  //colour
            }
          }
          else if (nd == 3){
            float dy = xc[1] - hole_ypos[n];
            float dz = xc[2] - hole_zpos[n];
            float dr = sqrt(dy*dy + dz*dz);
            if (dr < hole_radius[n]){// in hole
              pa[iqd] = ramb;
              if (nal >= 2) pa[iqal0+2] = 0.0;  //colour
            }
          }
        }
      }
    }
  
    //rescale
    pa[iqd] /= rhoscale;
    pa[iqe] /= pscale;
    pa[iqu0+1] /= uscale;
    pa[iqu0+2] /= uscale;
    //set entropy only after rescaling!!
    if (inten) pa[iqie] = pa[iqe]/powf(pa[iqd], g); //entropy =p/rho^g)
  }// SHELL_BLOWOUT setup
     

  usrqp(pa, qa);
  return;
}
//end usrqis.c

//file usrss.c
/*
+ usrss - sound speed
*/
float usrss(float d, float p)
{
  return (sqrt(g*p/d));
}
//end usrss.c

//file usrqp.c
/*
+ usrqp - conserved  variables from primitive
*/
void usrqp(float pa[], float qa[])
{
  float rho, divrho, ke, u;
  int i, id;
  rho = pa[iqd];
  qa[iqd] = rho;
  ke = 0.0;
  for (id = 1; id <= nd; id++){
    u = pa[iqu0+id];
    qa[iqu0+id] = u*rho;
    ke += u*u;
  }
  qa[iqe] = pa[iqe]/(g - 1.0) + 0.5*rho*ke;
  for (i = iqe+1; i <= iqal0 + scad; i++)
    qa[i] = rho*pa[i];
  for (i = iqal0 + scad + 1; i < nq; i++)
    qa[i] = pa[i];
  if (inten)
    qa[iqie] = rho*pa[iqie];
 
  //cout << "usrqp: pa[iqd] = " << pa[iqd] << "; qa[iqd] = " << qa[iqd] << "\n";

  return;
}
//end usrqp.c

//file usrpq.c
/*
+ usrpq - primitive variables from conserved
*/
void usrpq(float qa[], float pa[])
{
  float rho, divrho, ke, u;
  int i, id;
  rho = max(qa[iqd], dfloor);
  divrho = 1.0/rho;
  pa[iqd] = rho;
  ke = 0.0;
  for (id = 1; id <= nd; id++){
    u = qa[iqu0+id]*divrho;
    pa[iqu0+id] = u;
    ke += u*u;
  }
  ke *= 0.5*rho;
  pa[iqe] = (g - 1.0)*(qa[iqe] - ke);
  for (i = iqe + 1; i <= iqal0 + scad; i++)
    pa[i] = qa[i]*divrho;
  for (i = iqal0 + scad + 1; i < nq; i++)
    pa[i] = qa[i];
  if (fxp)
    usrfxp(pa);
  else
    pa[iqe] = max(pfloor,pa[iqe]);
  if (inten)
    pa[iqie] = qa[iqie]*divrho;
  return;
}
//end usrpq.c

//file usrfxq.c
/*
+ usrfxq - Fix up solution.
     Description :
        This fixes the  solution in a cell to  ensure energy, pressure
        and density are positive. It  is given the cell coordinates in
        xc,  so  that   it  can  also  be  used   for  other  purposes
        e.g. forcing  the solution  to take a  given value  at certain
        points.
*/
void usrfxq(struct cell * cptr)
{
  int i, id;
  float ke, rho, divrho, mu;
  //JMP defined:
  bool init,update;
  int n,wnd;
  float rs,pre,tmpold,col,tfloor,r,x,y,z,vx,vy,vz,v2,inte,etot;
  float xdist,ydist,zdist,mdotwind,vinfwind,tempwind,drmin;
  float avgm,totcolwind,pa[MAXQ];
  float rcal,d1_stag;

  //Map on CWB wind 
  //Do this first, because it involves a conversion from pa to qa
  //if ((strcmp(cas, "CWB") == 0) && (windy)){
  if (strcmp(cas, "CWB") == 0){
    if (orbit) orbit_calc();  //update cell positions     
    //compute mtm balance surface
    rcal = sqrt(mdot[0]*vinf[0]/(mdot[1]*vinf[1])); 
    d1_stag = rcal*dsep/(1.0+rcal);
    wnd = find_wind(cptr->xc, d1_stag);
    init = false;
    update = false;
    map_wind(init, wnd, cptr->xc, cptr->pa, update);
    // update qa if update occured in map_wind
    if (update) usrqp(cptr->pa, cptr->qa);// this is necessary (also to
                                          // avoid segmentation faults)
  }

  // Map on wind(s)  
  if (strcmp(cas, "WIND") == 0){
    for (n = 0; n < nstars; n++){      
      ydist = 0.0;
      zdist = 0.0;
      xdist = cptr->xc[0]-xpos[n];
      if (nd > 1) ydist = cptr->xc[1]-ypos[n];
      if (nd > 2) zdist = cptr->xc[2]-zpos[n];
      r = sqrt(xdist*xdist + ydist*ydist + zdist*zdist);
      rs = r*rscale;
      //drmin = (xr[0]-xl[0])/float(nc[0])/pow(2.0,nlevs-1);
      if (r <= remap1){
      //if (r <= remapcells*drmin){
        mdotwind = mdot[n];
        vinfwind = vinf[n];
        tempwind = tempw[n];
        //mdotwind = 1.0;
        //vinfwind = 1.0;
        rho = mdotwind/(4.0*gkpi*rs*rs*vinfwind);
        //cout << "rho = " << rho << "\n";
        cptr->qa[iqd] = rho;
        //cout << "rho = " << pa[iqd] << "\n";
        //exit(EXIT_FAILURE);
                     cptr->qa[iqu0+1] = rho*vinfwind*xdist/r;
        if (nd >= 2) cptr->qa[iqu0+2] = rho*vinfwind*ydist/r;
        if (nd >= 3) cptr->qa[iqu0+3] = rho*vinfwind*zdist/r;
        cptr->qa[iqal0+n+1] = rho;         //colour
        cptr->qa[iqal0+nstars+1] = rho*massfracx[n]; //H mass fraction
        cptr->qa[iqal0+nstars+2] = rho*massfracy[n]; //He mass fraction
        cptr->qa[iqal0+nstars+3] = rho*massfracz[n]; //metals mass fraction
        usrpq(cptr->qa,pa); //calculate primitive variables (needed for usravgmass)
        avgm = usravgmass(pa);
        //cout << "avgm = " << avgm;
        //exit(EXIT_FAILURE);
        ke = 0.0;
        for (id = 1; id <= nd; id++){
          ke += cptr->qa[iqu0+id]*cptr->qa[iqu0+id];
        }
        ke *= 0.5/rho;
        //pre = 1.0e-3*rho;
        pre = rho*tempwind*(kboltz/avgm);
        //cout << " rho = " << rho << ", tempwind = " << tempwind << ", avgm = " << avgm << "\n";
        //exit(EXIT_FAILURE);
        cptr->qa[iqe] = ke + pre/(g-1.0);
        if (inten) cptr->qa[iqie] = pre/powf(rho, g - 1.0);
        //rescale
        cptr->qa[iqd] /= rhoscale;
        cptr->qa[iqe] /= pscale;
        if (inten) cptr->qa[iqie] /= pscale;
        cptr->qa[iqu0+1] /= rhoscale*uscale;
        if (nd >= 2) cptr->qa[iqu0+2] /= rhoscale*uscale;
        if (nd >= 3) cptr->qa[iqu0+3] /= rhoscale*uscale;
        //cout << "nal = " << nal << "\n";
        //cout << "iqd = " << iqd << ", iqu0+1 = " << iqu0+1 << ", iqe = " << iqe  << ", iqie = " << iqie << "\n";
        //cout << "iqal0 = " << iqal0 << ", scad = " << scad << ", nq = " << nq << "\n";
        //exit(EXIT_FAILURE);
        for (i = iqal0 + 1; i <= iqal0 + nal; i++)
          cptr->qa[i] /= rhoscale;
      } //inside remap radius
    }//loop over stars
  }//case WIND

  /* Density */
  rho = max(cptr->qa[iqd], dfloor);
  divrho = 1.0/rho;
  cptr->qa[iqd] = rho;
  /* Thermal energy */
  ke = 0.0;
  for (id = 1; id <= nd; id++){
    mu = cptr->qa[iqu0+id];
    ke += mu*mu;
  }
  ke *= 0.5*divrho;
  col = divrho*cptr->qa[iqal0+1];

  //JMP: Thermal floor for CWB and WBB sims
  //if ((strcmp(cas, "CWB")==0)||(strcmp(cas, "WBB")==0)){
  if ((strcmp(cas, "CWB")==0)||(strcmp(cas, "WIND")==0)){
    //Put in a temperature floor
    //Calculate the avgmass in the cell
    if (strcmp(cas, "CWB")==0){
      if (col > 0.5) avgm = avgmass1;
      else avgm = avgmass2;
    }
    else{ //WIND case
      usrpq(cptr->qa, pa); //get primitive variables
      avgm = usravgmass(pa);
    }
    //First fixup qa[iqe]
    pre = (g-1.0)*(cptr->qa[iqe] - ke);
    tmpold = (pre/rho)*(avgm/kboltz)*(pscale/rhoscale);
    //if (windid == 0) tfloor = tamb;
    //else tfloor = tempw[windid-1];
    //tfloor = min(tamb,tempw[0]);
    tfloor = tmin;
    if (tmpold <= tfloor){
      pre = tfloor*rho*rhoscale*(kboltz/avgm);
      cptr->qa[iqe] = pre/(pscale*(g-1.0)) + ke; //makes qa[iqe] positive
    }
    //Now fixup qa[iqie]
    if (inten){
      pre = cptr->qa[iqie]*powf(rho, g - 1.0);
      tmpold = (pre/rho)*(avgm/kboltz)*(pscale/rhoscale);
      if (tmpold <= tfloor){
        pre = tfloor*rho*rhoscale*(kboltz/avgm);
        cptr->qa[iqie] = pre/(pscale*powf(rho, g - 1.0));
      }
    } 
  }
  // Thermal floor for CLUMP and SINGLECLOUD or MANYCLOUD sims
  else if ((strcmp(cas, "CLUMP") == 0)||(strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "MANYCLOUD") == 0)||(strcmp(cas, "LEAKYSHELL") == 0)){
    if (cptr->qa[iqe] - ke <= pfloor/(g - 1.0))
      cptr->qa[iqe] = ke + pfloor/(g - 1.0);
  }
  // Floor in original code is a fraction pfloor of ke
  else {
    if (cptr->qa[iqe] - ke <= pfloor/(g - 1.0))
    cptr->qa[iqe] = ke + pfloor/(g - 1.0);
  }

  //Put in a temperature floor for the radiative CLOUD and other sims
  
  if (((strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "MANYCLOUD") == 0)||(strcmp(cas, "LEAKYSHELL") == 0)||(strcmp(cas, "SHELL_BLOWOUT") == 0))&&(cool)){
    avgm = avgmassamb;
    pre = (g-1.0)*(cptr->qa[iqe] - ke)*pscale;
    rho = cptr->qa[iqd] * rhoscale;
    tmpold = (pre/rho)*(avgm/kboltz);
    col = cptr->qa[iqal0+1]/cptr->qa[iqd];
    //if (col < 2.0e-3) tfloor = col*tmin + (1.0-col)*tamb;
    //else tfloor = tmin;
    tfloor = tmin;
    if (tmpold <= tfloor){
      pre = tfloor*rho*(kboltz/avgm);
      cptr->qa[iqe] = pre/(pscale*(g-1.0)) + ke;
    }
  }

  if (keps){
    cptr->qa[iqk] = max(cptr->qa[iqk], rho*kmin);
    cptr->qa[iqeps] = max(cptr->qa[iqeps], rho*epsmin);
  }
  /* Scalars */
  for (i = iqal0+1; i <= iqal0+scad; i++)
    cptr->qa[i] = min(cptr->qa[iqd], max(cptr->qa[i], 0.0));
  return;
}
//end usrfxq.c

//file usrfxp.c
/*
+ usrfxp - Fix up primitive solution.
     Description :
        This  fixes the  solution in  pa to  ensure that  pressure and
        density are positive.
*/
void usrfxp(float pa[])
{
  int i;
  /* Fix density */
  pa[iqd] = max(pa[iqd], dfloor);
  /* Fix pressure */
  pa[iqe] = max(pa[iqe], pfloor);
  if (keps){
    pa[iqk] = max(pa[iqk], kmin);
    pa[iqeps] = max(pa[iqeps], epsmin);
  }
  /* Scalars */
  for (i = iqal0+1; i <= iqal0+scad; i++)
    pa[i] = min(1.0, max(pa[i], 0.0));
  return;
}
//end usrfxp.c

//file usrps.c
/* 
+ usrps - construct plot scalar
     Description :
        This  returns  the value  of  a plot  scalar  in  s given  the
        conserved solution  in qa. The  scalar is defined by  ITEM and
        returns TRUE if the plot item is found, FALSE otherwise.
*/
bool usrps(cell *cptr, float &s)
{
  int i, j, id;
  bool found;
  float qref[MAXQ], pa[MAXQ], tv[MAXQ], temp;
  struct cell *left, *right;
  //JMP added
  struct join *jptr;
  float dens, pre, maxtemp, col, xfrac;
  float c,v,edot,divdx,divv,rc;
  float avgm;
  double nenh, gr, em;
  double logte,logemisshalpha_caseA,htotcolionrate,htotrecombcaseA,
    ratio,onesubx,ionfrac;
  bool rzp,zrp,rtp;

  found = false;
  if (strncmp(item, "QA", 2) ==0){
    sscanf(&item[2], "%d", &i);
    if ((i < 1) || (i > nq))
      found = false;
    else{
      s = cptr->qa[i-1];
      found = true;
    }
  }
  else if (strncmp(item, "QR", 2) ==0){
    sscanf(&item[2], "%d", &i);
    if ((i < 1) || (i > nq))
      found = false;
    else{
      s = cptr->dqdt[i-1]/cptr->vol;
      found = true;
    }
  }
  else if (strcmp(item, "RHO") == 0){
    s = cptr->qa[iqd];
    found = true;
  }
  else if (strcmp(item, "RHOSCALED") == 0){
    s = cptr->qa[iqd]/(ramb/rhoscale);
    found = true;
  }
  else if (strcmp(item, "U1") == 0){
    s = cptr->qa[iqu0+1]/cptr->qa[iqd];
    found = true;
  }
  else if (strcmp(item, "U2") == 0){
    if (nd >= 2){
      s = cptr->qa[iqu0+2]/cptr->qa[iqd];
      found = true;
    }
    else
      found = false;
  }
  else if (strcmp(item, "U3") == 0){
    if (nd == 3){
      s = cptr->qa[iqu0+3]/cptr->qa[iqd];
      found = true;
    }
    else
      found = false;
  }
  else if (strcmp(item, "UT") == 0){
    s = 0.0;
    for (id = 1; id <= nd; id++){
      temp = cptr->qa[iqu0+id]/cptr->qa[iqd];
      s += temp*temp;
    }
    s = sqrt(s);
    found = true;
  }
  else if (strcmp(item, "VEL") == 0){
    s = 0.0;
    for (id = 1; id <= nd; id++){
      s += cptr->qa[iqu0+id] * cptr->qa[iqu0+id];
    }
    s = sqrt(s)/cptr->qa[iqd];
    found = true;
  }
  else if (strcmp(item, "POT") == 0){
    if (grav){
      s = gconst*cptr->qa[iqpot];
      found = true;
    }
    else
      found = false;
  }
  else if (strncmp(item, "GA", 2) ==0){
    sscanf(&item[2], "%d", &id);
    if ((!grav) || (id < 1) || (id > nd))
      found = false;
    else{
      i = 0;
      s = 0.0;
      left = cptr->ljoin[id-1]->left;
      if (left != NULL){
	s += cptr->qa[iqpot] - left->qa[iqpot];
	i++;
      }
      right = cptr->rjoin[id-1]->right;
      if (right != NULL){
	s +=  right->qa[iqpot] - cptr->qa[iqpot];
	i++;
      }
      if (i == 0)
	found = false;
      else
	s /= - (float)(i)*
	  (cptr->rjoin[id-1]->xj[id-1] - cptr->ljoin[id-1]->xj[id-1])
	  /gconst;
      found = true;
    }
  }
  else if (strcmp(item, "QSOL") == 0){
    s = cptr->qsol;
    found = true;
  }
  else if (strcmp(item, "ICELL") == 0){
    s = cptr->icell;
    found = true;
  }
  else if (strcmp(item, "PROC") == 0){
    s = (float)mype;
    found = true;
  }
  else if (strcmp(item, "WORK") == 0){
    i = (int)((cptr->xc[nd-1] - xl[nd-1])/lev[nlevs-1]->dxbar[nd-1]);
    s = (float)wka[i];
    found = true;
  }
  else{
    usrpq(cptr->qa, pa);
    if (strncmp(item, "PA", 2) ==0){
      sscanf(&item[2], "%d", &i);
      if ((i < 1) || (i > nq))
      found = false;
      else{
	s = pa[i-1];
	found = true;
      }
    }
    else if (strcmp(item, "KEFRAC") == 0){
      s = 0.0;
      for (id = 1; id <= nd; id++){
        s += cptr->qa[iqu0+id] * cptr->qa[iqu0+id];
      }
      s = 0.5*s/cptr->qa[iqd];
      s = s/cptr->qa[iqe]; //KE fraction
      found = true;
    }
    else if (strcmp(item, "IEFRAC") == 0){
      s = 0.0;
      for (id = 1; id <= nd; id++){
        s += cptr->qa[iqu0+id] * cptr->qa[iqu0+id];
      }
      s = 0.5*s/cptr->qa[iqd];  //KE
      s = (cptr->qa[iqe] - s)/cptr->qa[iqe]; //IE fraction
      found = true;
    }
    else if (strncmp(item, "AL", 2) ==0){
      sscanf(&item[2], "%d", &i);
      if ((i < 1) || (i > nal))
      found = false;
      else{
	s = pa[iqal0+i];
	found = true;
      }
    }
    else if (strncmp(item, "ARHO", 2) ==0){
      sscanf(&item[4], "%d", &i);
      if ((i < 1) || (i > nal))
      found = false;
      else{
	s = pa[iqal0+i]*cptr->qa[iqd];
	found = true;
      }
    }
    else if (strcmp(item, "PG") == 0){
      s = pa[iqe];
      found = true;
    }
    else if (strcmp(item, "PGSCALED") == 0){
    float ambp = ramb*tamb*(kboltz/avgmass);
      s = pa[iqe]/(ambp/pscale);
      found = true;
    }
    else if (strcmp(item, "PGINT") == 0){
      if (inten){
	s = pa[iqie]*powf(pa[iqd], g);
	found = true;
      }
      else
	found = false;
    }
    else if (strcmp(item, "ENT") == 0){
      s = pa[iqe]/powf(pa[iqd], g);
      found = true;
    }
    else if (strcmp(item, "ENTIE") == 0){
      if (inten){
	s = pa[iqie];
	found = true;
      }
    }
    else if (strcmp(item, "C") == 0){
      s = usrss(pa[iqd], pa[iqe]);
      found = true;
    }
    else if (strcmp(item, "MACH") == 0){
      s = 0.0;
      for (id = 1; id <= nd; id++)
	s += pa[iqu0+id]*pa[iqu0+id];
      s = sqrt(s)/usrss(pa[iqd], pa[iqe]);
      found = true;
    }
    else if (strcmp(item, "TE") == 0){
      s = usrte(pa);
      found = true;
    }
    else if (strcmp(item, "TEMP") == 0){
      if ((strcmp(cas, "WIND") == 0)||(strcmp(cas, "BOWSHOCK") == 0)){//calculate from cell mass fracs
        avgm = usravgmass(pa);
      }
      else if (strcmp(cas, "CWB") == 0){
        if (nal >= 1){
          col = pa[iqal0+1];
          if (col > 0.5) avgm = avgmass1;
          else avgm = avgmass2;
        }
        else avgm = avgmass1;
      }
      else{
        avgm = avgmass;
      }
      if (tempmu){
        // use temperature dependent mu
        float mu = usrpdmu(pa[iqe]*pscale/(pa[iqd]*rhoscale));
        avgm = mu*massh;
      }
      s = usrte(pa)*(pscale/rhoscale)*(avgm/kboltz);
      //cout << "te = " << usrte(pa) << "; avgm = " << avgm << "\n";
      //quit();
      found = true;
    }
    else if (strcmp(item, "TEMPMU") == 0){
      float mu;
      if (strcmp(cas, "WIND") == 0){//calculate avgmass from cell mass fracs
        mu = usravgmass(pa)/massh;
      }
      else if (strcmp(cas, "CWB") == 0){
        if (nal >= 1){
          col = pa[iqal0+1];
          if (col > 0.5) mu = avgmass1/massh;
          else mu = avgmass2/massh;
        }
        else mu = avgmass1/massh;
      }
      else{
        mu = avgmass/massh;
      }
      if (tempmu){
        // use temperature dependent mu
        mu = usrpdmu(pa[iqe]*pscale/(pa[iqd]*rhoscale));
      }
      s = mu;
      found = true;
    }
    else if (strcmp(item, "K") == 0){
      if (!keps)
	found = false;
      else{
	s = pa[iqk];
	found = true;
      }
    }
    else if (strcmp(item, "EPS") == 0){
      if (!keps)
	found = false;
      else{
	s = pa[iqeps];
	found = true;
      }
    }
    else if (strcmp(item, "TL") == 0){
      if (!keps)
	found = false;
      else{
	s = ceps*pa[iqk]*sqrt(pa[iqk])/max(pa[iqeps],epsmin);
	found = true;
      }
    }
    else if (strcmp(item, "TCOOL") == 0){
      usrsaf(1.0, cptr, tv); // pass in a dummy dt = 1.0
      s = pa[iqe]/max(1.0e-10, mabs(tv[iqe]));
      found = true;
    }
    else if (strncmp(item, "MU", 2) ==0){
      sscanf(&item[2], "%d", &i);
      if ((i < 1) || (i > nq))
      found = false;
      else{
	usrmu(pa, tv);
	s = tv[i];
	found = true;
      }
    }
    else if (strncmp(item, "SA", 2) ==0){
      sscanf(&item[2], "%d", &i);
      if ((i < 1) || (i > nq))
      found = false;
      else{
	for (j = 0; j < nq; j++)
	  cptr->pa[j] = pa[j];
	usrsaf(0.0, cptr, tv);
	s = tv[i-1];
	found = true;
      }
    }
    else if (strcmp(item, "PRAM") == 0){
      s = 0.0;
      for (id = 1; id <= nd; id++){
        s += cptr->qa[iqu0+id] * cptr->qa[iqu0+id];
      }
      s *= 0.5/pa[iqd];
      found = true;
    }
    else if (strcmp(item, "DIVV") == 0){//cell struct no longer contains divv
      //For some reason this doesn't work properly
      rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
      zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
      rtp = (strcmp(geom, "RTP") == 0);
      //Cartesian divergence 
      divv = 0.0;
      for (i = 0; i < nd; i++){
        divv += cptr->dudx[i][i];
      }
      if (rzp){ // Cylindrical polars xc[0] = R 
        rc = cptr->xc[0];
        divv += cptr->pa[iqu0+1]/rc;
      }
      else if (zrp){ // Cylindrical polars xc[1] = R
        rc = cptr->xc[1];
        divv += pa[iqu0+2]/rc;
      }
      else if (rtp){ //Spherical polars 
        rc = cptr->xc[0];
        divv += 2.0*pa[iqu0+1]/rc;
      }
      s = divv;
      found = true;
    }
    else if (strcmp(item, "MOM1") == 0){
      s = cptr->qa[iqu0+1];
      found = true;
    }
    else if (strcmp(item, "MOM2") == 0){
      if (nd >= 2){
	s = cptr->qa[iqu0+2];
	found = true;
      }
      else
	found = false;
    }
    else if (strcmp(item, "MOM3") == 0){
      if (nd == 3){
	s = cptr->qa[iqu0+3];
	found = true;
      }
      else
	found = false;
    }
    else if (strcmp(item, "MOMT") == 0){
      s = 0.0;
      for (id = 1; id <= nd; id++){
	temp = cptr->qa[iqu0+id];
	s += temp*temp;
      }
      s = sqrt(s);
      found = true;
    }
    
    else if (strcmp(item, "EDOT") ==0){
      for (j = 0; j < nq; j++)
        cptr->pa[j] = pa[j];
      usrsaf(1.0, cptr, tv); // pass in a dummy dt = 1.0
      s = tv[iqe];
      found = true;
    }

    else if (strcmp(item, "LJEANS") ==0) { //RA
      //Try and plot sqrt(15/4 kT/G mju rho pi)
      //Ideally rescaled to 2pc.
      //k = kboltz (cgs)
      //mju: to do properly need avgmass. But what if i used 1.4massh...
      //pi = gkpi
      //G = gconstcgs
      // get T get rho.     
      //avgm = avgmass;
      avgm = massh;
      //      cout << avgm << endl;
      temp = usrte(pa)*(pscale/rhoscale)*(avgm/kboltz);
      s = cptr->qa[iqd];
      s = temp/s;
      s = sqrt(15*kboltz*s/(4*avgm*gkpi));  //cgs?
      //cout << s << endl;
      s /= rscale;
      found = true;
    }
    //else if (strcmp(item, "TCOOL") ==0){//JMP
    //	cptr->dqdt[iqe] = 0.0;
    //  divdx = 1.0/(cptr->rjoin[0]->xj[0] - cptr->ljoin[0]->xj[0]);
    //	usrsaf(0.0, divdx, cptr);
    //        edot = (cptr->dqdt[iqe]/cptr->vol)*pscale*uscale/rscale;//Edot(erg/cm^3/s)
    //	s = pa[iqe]*pscale/(g-1.0)/edot; //tcool (s)
    //	found = true;
    //}
    else if (strcmp(item, "HALPHA") ==0){//JMP 23/05/08

      //rescale to different values (best to set absolutely due to
      //possible repeated calls to this routine)
      //rscale = 0.1*pc;
      //uscale = sqrt(800.0)*1.0e6;
      //pscale = rhoscale*uscale*uscale;
      //tscale = rscale/uscale;

      dens = pa[iqd]*rhoscale;
      col = pa[iqal0+1];
      //	  rhosqV = (dens/massh)*(dens/massh)*col*cvol2;
      temp = (pa[iqe]/pa[iqd])*(pscale/rhoscale)*(avgmass/kboltz);

      //The fit in DKS's routine is specifically for 500 < Te < 30000,
      //but Figure 2 of Sutherland & Dopita (1993) indicates that a 
      //power-law extrapolation works well up to 1e5 K.
      //log10 emsvty = m log10 t_e + n

      s = 0.0;
      //if ((temp > 1.05e4)&&(temp < 1.0e5)){
        logte = log10(temp);
        logemisshalpha_caseA = -20.822 + logte*(-0.956);
        //logemisshalpha_caseB = -20.927 + logte*(-0.884);
        s = pow(10.0,logemisshalpha_caseA);// (erg cm^3/s)
        //Now calculate the fractional ionization
        htotcolionrate = 2.5e-10*(1.0+(temp/7.8945e4))*
                            sqrt(temp)*exp(-(1.5789e5/temp));
        htotrecombcaseA = 4.11e-13*pow((1.0e-4*temp),-0.698);
        ratio = htotcolionrate/htotrecombcaseA;
        onesubx = 1.0/(1.0+ratio);
        ionfrac = 1.0-onesubx;
        //Multiply the emissivity by the ionization fraction to obtain
        //the true temperature dependence
        s *= ionfrac;
	//}
      found = true;
    }     
  }
  return(found);
}
//end usrps.c

//file usrsaf.c
/*
+ usrsaf - source terms.
*/
void usrsaf(float dt, cell *cptr, float so[])
{
  int i, j;
  float rc, dudx[MAXD][MAXD], pa[MAXQ], mu[MAXQ], vcof, tv, divv, pp;
  bool rzp, zrp, rtp;
  //JMP added
  int n;
  float vol,fac,rho,dr,dz,radc,dx,dy,temp,alpha;
  float tinj,coolinj,coolamb,eloss;
  float tv1,tv2,tv3,pdot;

  /* Zero so (safety) and copy primitives */
  for (i = 0; i < nq; i++){
    so[i] = 0.0;
    pa[i] = cptr->pa[i];
  }
  /* Geometry */
  rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
  zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
  rtp = (strcmp(geom, "RTP") == 0);
  /* Geometric source term (if any) */
  if (rzp){
    /* Cylindrical polars xc[0] = R */
    rc = cptr->xc[0];
    so[iqu0+1] += pa[iqe]/rc;
  }
  else if (zrp){
    /* Cylindrical polars xc[1] = R */
    rc = cptr->xc[1];
    so[iqu0+2] += pa[iqe]/rc;
  }
  else if (rtp){
    /* Spherical polars */
    rc = cptr->vol/(cptr->rjoin[0]->area - cptr->ljoin[0]->area);
    so[iqu0+1] += pa[iqe]/rc;
  }
  if (viscous){
    /* Diffusion coefficients etc */
    usrmu(pa, mu);
    vcof = mu[iqu0+1];
    divv = 0.0;
    /*Cartesian divergence */
    for (i = 0; i < nd; i++){
      divv += cptr->dudx[i][i];
    }
    /* Geometric source term (if any) */
    if (rzp){
      /* Cylindrical polars xc[0] = R */
      divv += pa[iqu0+1]/rc;
      so[iqu0+1] += 2.0*vcof*(divv/3.0 - pa[iqu0+1]/rc)/rc;
    }
    else if (zrp){
      /* Cylindrical polars xc[1] = R */
      divv += pa[iqu0+2]/rc;
      so[iqu0+2] += 2.0*vcof*(divv/3.0 - pa[iqu0+2]/rc)/rc;
    }
    else if (rtp){
      /* Spherical polars */
      divv += 2.0*pa[iqu0+1]/rc;
      so[iqu0+1] += 4.0*vcof*(divv/3.0 - pa[iqu0+1]/rc)/rc;
    }
    if (keps){
      /* Velocity gradients */
      for (i = 0; i < nd ; i++)
	for (j = 0; j < nd; j++)
	  dudx[i][j] = cptr->dudx[i][j];
      /* Production term */
      if (rzp){
	/* Turbulent pressure r-momentum source */
	so[iqu0+1] += 2.0*pa[iqd]*pa[iqk]/(3.0*rc);
	/* Geometric production term */
	tv = pa[iqu0+1]/rc;
	pp = 2.0*tv*tv;
      }
      else if (zrp){
	/* Turbulent pressure r-momentum source */
	so[iqu0+2] += 2.0*pa[iqd]*pa[iqk]/(3.0*rc);
	/* Geometric production term */
	tv = pa[iqu0+2]/rc;
	pp = 2.0*tv*tv;
      }
      else if (rtp){
	/* Turbulent pressure r-momentum source */
	so[iqu0+1] += 4.0*pa[iqd]*pa[iqk]/(3.0*rc);
	/* Geometric production term */
	tv = pa[iqu0+1]/rc;
	pp = 4.0*tv*tv;
      }
      else
	pp = 0.0;
      for (i = 0; i < nd; i++){
	for (j = 0; j < nd; j++)
	  pp += dudx[i][j]*(dudx[i][j] + dudx[j][i]);
      }
      pp = vcof*pp - 2.0*divv*(pa[iqd]*pa[iqk] + divv*vcof)/3.0;
      /* k source */
      so[iqk] += pp - pa[iqd]*pa[iqeps];
      /* Energy source */
      so[iqe] += (pa[iqd]*pa[iqeps] - pp);
      /* Epsilon source */
      so[iqeps] += pa[iqeps]*(ce1*pp - pa[iqd]*ce2*pa[iqeps])/pa[iqk]; 
    }
  }
  /*
     Physical  sources  (per unit  volume).   Note  primitives are  in
     pa. If the entropy is used, you also need to add in its sources
  */
 // if (strcmp(cas, "CLUMP") == 0){
    // Add in mass and energy injection region(s) for clump(s)
    // This code computes the Reynold's stress. For the k-epsilon
    // model, index iqt+0 is the old k in COBRA. 

    //Ed's gravity terms for his cluster-clump sim
    //so[iqu0+2] += (gravit)*rho;
    //so[iqe]+= (gravit)*pa[iqu0+2]*rho;
 /*   for (n = 0; n < ncc; n++){  //loop over clumps
      if (nd == 2){
        dr = cptr->xc[0] - xc1[n];
        dz = cptr->xc[1] - xc2[n];
        radc = sqrt(dr*dr + dz*dz);
      }
      else if (nd == 3){
        dx = cptr->xc[0] - xc1[n];
        dy = cptr->xc[1] - xc2[n];
        dz = cptr->xc[2] - xc3[n];
        radc = sqrt(dx*dx + dy*dy + dz*dz);
      }
      if (radc <= 1.0){ //inject mass and energy
	//tv2 = mdot*(1.+(t*(gravit)/uamb))*(1.+(t*(gravit)/uamb)); //(also need
	//gravity terms above)
        tv2 = mdotcl; //(no gravity)
        so[iqd] = tv2;
	so[iqe] += tv2*tinj/(g-1.0);
        so[iqal0+1] = tv2; //colour injection (1 tracer only)
	//        so[iqal0+n] = so[iqd]; 
        //colour injection (multiple tracers)
        if (keps){
	  so[iqk] += 0.01*tv2; //this is the k source term
          so[iqeps] += sqrt(0.01*tv2)*0.01*tv2/mtl;//set by k/mtl
        }
      }
    } //loop over clumps
    // Heating/cooling for the flow
    /* calculate pressure at old time (tv1) */
 /*   tv1 = 0.0;
    tv2 = cptr->qa[iqd];
    for (i = 1; i <= nd; i ++){
      tv3 = cptr->qa[iqu0+i];
      tv1 += tv3*tv3;
    }
    tv1 = (g - 1.0)*(cptr->qa[iqe] - 0.5*tv1/tv2); //tv1 = pressure
    /* exponential extrapolation of temperature */
  /*  alpha = pa[iqal0+1]; // single colour variable
    tv3 = alpha*(coolinj - coolamb) + coolamb; //cooling rate
    temp = tinj + (tv1/tv2 - tinj)*(float)
	    exp((double)(-tv3*(g - 1.0)*dt));

    // Ambient flow adiabatic, injected gas isothermal
    so[iqe] -= rho*(alpha*(coolinj-coolamb) + coolamb)*(temp-tinj);
    // Both flows have some degree of cooling between adiabatic 
    // (rat and/or
    // acte=0.0), or isothermal (at different temperatures)
    // rtau = 10.0;
    // acte2 = acte*exp(-rc/rtau);
    // so[iqe] = so[iqe] - rho*alpha*coolinj*(temp-1.0);
    // so[iqe] = so[iqe] - rho*(1.0-alpha)*coolamb*(temp-2.4);
    // Both flows cooling isothermally (to different temperatures)
    // so[iqe] = so[iqe] - rho*alpha*coolinj*(temp-1.0);
    // so[iqe] = so[iqe] - rho*(1.0-alpha)*coolamb*(temp-1.0);
  }
  /* Sam's shock code
  else if (strcmp(cas, "SHOCK") == 0 ){
    // Exponential integration
    int id;
    float al1, fact;
    al1 = pa[iqal0+1];
    fact = exp(-min(rat*al1*dt, 20.0));
    if (fact < 0.9){
      for (id = 1; id <= nd; id++){
	so[iqu0+id] = cptr->qa[iqu0+id]*(fact - 1.0)/dt;
	so[iqe] += pa[iqu0+id]*so[iqu0+id];
      } 
    }
    else{
      for (id = 1; id <= nd; id++){
	so[iqu0+id] = - rat*al1*pa[iqd]*pa[iqu0+id];
      }
    }
  }
  */
  else if (strcmp(cas, "WAVE") == 0 ){
    if (pa[iqe] > 1.1){
      so[iqal0+1] = rat*pa[iqd]*(1.0 - pa[iqal0+1]);
      so[iqe] = heat*so[iqal0+1];
    }
  }
  //else if (strcmp(cas, "GRAVXY") == 0){
  //  if (cptr->xc[0]*cptr->xc[0] + cptr->xc[1]*cptr->xc[1] < rat*rat)
  //    so[iqd] = heat;
  //}

  // Calculate cooling (no heating)
  if ((strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "SHOCK") == 0)||(strcmp(cas, "MANYCLOUD") == 0)||(strcmp(cas, "LEAKYSHELL") == 0)||(strcmp(cas, "WIND") == 0)){
    if (cool){
      so[iqe] += radiate(pa,dt);  //JMP's idealized cooling curve    
      /*  
    // calculate pressure at old time (tv1) 
    tv1 = 0.0;
    tv2 = cptr->qa[iqd];
    for (i = 1; i <= nd; i ++){
      tv3 = cptr->qa[iqu0+i];
      tv1 += tv3*tv3;
    }
    tv1 = (g - 1.0)*(cptr->qa[iqe] - 0.5*tv1/tv2); //tv1 = pressure

    coolinj = 50.0;   //cooling rate of injected material per unit volume
    coolamb = 0.0;      //cooling rate of ambient  material per unit volume
    //tinj = temperature (= p/rho in mg units) of injected material
    // we will set tinj to tcloud
    tinj = (tamb/chi)*(kboltz/avgmass)*(rhoscale/pscale);
    //cout << "tinj = " << tinj << "\n";
    alpha = cptr->pa[iqal0+1]; // single colour variable
    temp = tv1/tv2; //old temp in units of p/rho
    // Ambient flow adiabatic, injected gas isothermal
    //so[iqe] -= rho*(alpha*(coolinj-coolamb) + coolamb)*(temp-tinj);

    // exponential extrapolation of temperature
    tv3 = alpha*(coolinj - coolamb) + coolamb; //cooling rate
    //cout << "tv3 = " << tv3 << "\n";
    temp = tinj + (tv1/tv2 - tinj)*(float)exp((double)(-tv3*(g - 1.0)*dt));//new temp
    //pdot = Tdot/rho    (tv1/tv2 = old temp in units of p/rho)
    pdot = ((temp-tv1/tv2)/tv2)/dt;
    //Edot = pdot/(g-1.0)
    so[iqe] += pdot/(g-1.0);
    //temp = pa[iqe]/pa[iqd];
    //cout << "temp = " << temp << ", tinj = " << tinj << "\n";

    //cout << "so[iqe] = " << so[iqe] << "\n";
    //exit(EXIT_FAILURE);
    */
    }//if cool
    
  }// cooling only
  else if ((cool)&&(heating)){
    so[iqe] += radiate_heatcool(pa,dt);
  }

  //if (((strcmp(cas, "WIND") == 0)||(strcmp(cas, "CWB") == 0)) && (cool)){
  /* - commented out, since otherwise it cools twice!
  if (cool){
    eloss = radiate(pa,dt); //eloss is negative
    so[iqe] += eloss;
    if (inten) so[iqie] += eloss*(g - 1.0)/powf(pa[iqd], g - 1.0);
  }
  */

  return;
}
//end usrsaf.c

//file usrref.c
/*
+ usrref - reference solution.
     Description:
        Provides a reference solution  for cell cptr. 
*/
void usrref(cell *cptr, float qref[])
{
  int i, id;
  float pa[MAXQ], rho, ss;
  for (i = 0; i < nq; i++)
    qref[i] = mabs(cptr->qa[i]);
  ss = sqrt(cptr->qa[iqe]*cptr->qa[iqd]);
  for (id = 1; id <= nd; id++)
    qref[iqu0+id] = mabs(cptr->qa[iqu0+id]) + ss;
  rho = cptr->qa[iqd];
  if (keps){
    qref[iqk] = 0.01*qref[iqe]/rho;
    qref[iqeps] = qref[iqk];
  }
  for (i = iqal0 + 1; i <= iqal0 + scad; i++)
    qref[i] = rho;
  for (i = iqal0 + scad + 1; i <= iqal0 + nal; i++)
    qref[i] = 1.0;
  if (grav)
    qref[iqpot] = big;
  return;
}
//end usrref.c

//file usrqsol.c
/*
+ usrqsol - Solution quality.
     Description:
        This returns  qsol = badcell if  the solution is  poor, qsol =
        good if  not. This is  determined by comparing  cptr->qa (fine
        cell   value),  with   cptr->pa  (mapped   down   coarse  cell
        value). usrref provides a reference solution.
*/
float usrqsol(int ilev, struct cell *cptr)
{
  int i;
  float qsol, qref[MAXQ];
  //JMP added
  int n, difflev, closestclump;
  float dist,xdist,ydist,zdist,r,rcoord,zcoord,rdist,rmin,col,axisdist;
  float drfine,drcurrent,ambcol;
  float cs,vb,initial_shock_pos,drcoarse,drlev;
  bool rzp,zrp,rtp,xyz;

  xyz = (strcmp(geom, "XYZ") == 0);
  rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
  zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
  rtp = (strcmp(geom, "RTP") == 0);

  usrref(cptr, qref);
  qsol = good;
  if (keps){
    for (i = 0; i < iqk; i++)
      if (mabs(cptr->qa[i] - cptr->pa[i]) > rtol*qref[i])
	qsol = badcell;
    for (i = iqal0+1; i < nq; i++)
      if (mabs(cptr->qa[i] - cptr->pa[i]) > rtol*qref[i])
	qsol = badcell;
  }
  else{
    for (i = 0; i <= iqd; i++){ // refinement based on density ONLY
      if (mabs(cptr->qa[i] - cptr->pa[i]) > rtol*qref[i])
	qsol = badcell;
    }
    /*
    for (i = 0; i < nq; i++){
      if (mabs(cptr->qa[i] - cptr->pa[i]) > rtol*qref[i])
	qsol = badcell;
    }
    */
  }

  //if (nd == 2) r = cptr->xc[0];
  if (nd == 2) r = sqrt(cptr->xc[0]*cptr->xc[0] + cptr->xc[1]*cptr->xc[1]);
  if (nd == 3) r = sqrt(cptr->xc[0]*cptr->xc[0] + cptr->xc[1]*cptr->xc[1] + cptr->xc[2]*cptr->xc[2]);

  if (r > 10.0){
    //cout << "qa[iqd] = " << cptr->qa[iqd] << "; pa[iqd] = " << cptr->pa[iqd] << "; qref = " << qref[iqd] << "; rtol = " << rtol << "; badcell = " << badcell << "\n";
    //quit();
  }
  //CWB case: I can't get it to work (even with cn=0.01 and viscous) if I limit
  //refinement at start-up (see p161 of my Notes on Developing Hydro Code book)
  if ((strcmp(cas, "CWB") == 0)&&(t != 0.0)){ //limit refinement far downstream
  //if ((strcmp(cas, "CWB") == 0)){ //limit refinement far downstream
    dist = 1.0e30;
    for (n = 0; n < nstars; n++){      
      ydist = 0.0;
      zdist = 0.0;
      xdist = cptr->xc[0]-xpos[n];
      if (nd > 1) ydist = cptr->xc[1]-ypos[n];
      if (nd > 2) zdist = cptr->xc[2]-zpos[n];
      r = sqrt(xdist*xdist + ydist*ydist + zdist*zdist);
      dist = min(r,dist);  // minimum distance of a cell from a star
    }
    //cout << "rob = " << rob << "\n";
    //cout << "nlevs = " << nlevs << "\n";
    //cout << "dist = " << dist << "\n";
    //cout << "ilev = " << ilev << "\n";
    //exit(EXIT_FAILURE);

    drlev = (xr[0]-xl[0])/float(nc[0])/pow(2.0,ilev); //current cell size
    if (dist > 50.0*drlev) qsol = good;

    //nlevs is the value of the finest grid (ie if G3, then nlevs=3) 
    //difflev = nlevs-ilev;  //difference between current level (ilev) and 
                           // finest level (nlevs)
    //if (dist >= 1.5*rob*pow(2.0,double(difflev-1))) qsol = good; //Good

    //if (dist >= 3.0*rob*pow(2.0,double(difflev-1))) qsol = good;
    //if ((dist >= 3.0*rob)&&(ilev==nlevs-1)) qsol = good;
    //else if ((dist >= 5.0*rob)&&(ilev==nlevs-2)) qsol = good;
    //else if ((dist >= 10.0*rob)&&(ilev==nlevs-3)) qsol = good;
  }

  
  if ((strcmp(cas, "CLUMP") == 0)||(strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "MANYCLOUD") == 0)||(strcmp(cas, "FILAMENT") == 0)){
 
  cs = sqrt(g*kboltz*tamb/avgmass);  //use ionized mu for sound/shock speed
  vb = mach*cs/uscale;
  drcoarse = (xr[0]-xl[0])/float(nc[0]); //coarsest cell size

  // nlevs is the value of the finest grid (ie if G3, then nlevs=3) 
  // tnlevs is the number of grids (i.e. the number after LEVELS in the start file)
  // If we have "LEVELS 4" then tnlevs = 4, giving grids G0 to G3. Grid G3 has ilev = 3.

  difflev = (tnlevs-1)-ilev; // difference between current level (ilev) and 
                             // desired finest level (tnlevs)
  drfine = (xr[0]-xl[0])/float(nc[0])/pow(2.0,tnlevs-1); // finest cell size
  drcurrent = drfine*pow(2.0,difflev);   // current cell size

  // JMP: Set refinement for 3D single cloud sims. To set the initial
  // refinement we make use of "tnlevs", the total number of levels
  // that the user wished for in the START file. I have added this
  // variable to model.h.
  if ((strcmp(cas, "SINGLECLOUD") == 0)&&(nd == 3)){
    initial_shock_pos = -2.0;
    col = cptr->qa[iqal0+1]/cptr->qa[iqd]; // cell colour

    //cout << "nlevs = " << nlevs << "; tnlevs = " << tnlevs; exit(EXIT_FAILURE);
    // Initially we wish to have 6 coarse cells per shell "width"
    if (t == 0.0){
      float drshell = 1.0; // initialize with radius of cloud 
      for (int i = tnlevs-1; i > ilev; i--){
        difflev = tnlevs-1-i;  // difference between current level (i) and 
                               // desired finest level (tnlevs-1)
        drlev = drfine*pow(2.0,difflev); //current cell size 
        drshell += 6.0*drlev;
        //cout << "i = " << i << "; difflev = " << difflev << "; drlev = " << drlev << "; drshell = " << drshell << "\n";
      }
      //exit(EXIT_FAILURE);
      if (r < drshell) qsol = badcell;
    }

    // Now limit the refinement based on off-axis distance and colour
    xdist = cptr->xc[0];
    ydist = cptr->xc[1];
    zdist = cptr->xc[2];
    axisdist = sqrt(ydist*ydist + zdist*zdist);
    if (axisdist > 2.0){
      difflev = (tnlevs-1)-ilev;
      if (difflev-1 < int(1.3*axisdist)/2){
        // cell is far enough off-axis to be potentially derefined. Now check colour.
        // cells with colour > 0.5 are not forced to derefine.
        if ((chi > 1.0e2)&&(col < 0.01)&&(ilev < tnlevs - 3)) qsol = good;
        else if ((col < 0.15)&&(ilev < tnlevs - 2)) qsol = good;
        else if (col < 0.5) qsol = good;
      }
    }

    // Also limit refinement for cells far downstream. Derefine up to 4 coarse cells into the
    // postshock flow once the shock gets past the specified minimum distance.
    xdist = cptr->xc[0];    
    if (xdist > max(20.0, vb*t + initial_shock_pos - 4.0*drcoarse)){
      int limlev = tnlevs - 2;
      if (ilev >= limlev) qsol = good; 
    }    
    if (xdist > max(40.0, vb*t + initial_shock_pos - 4.0*drcoarse)){ // this must be an if and not an elseif
      int limlev = tnlevs - 3;
      if (ilev >= limlev) qsol = good; 
    }

  } 
  else if ((strcmp(cas, "MANYCLOUD") == 0)&&(nd == 3)){
    // JMP refinement/derefinement criteria
    initial_shock_pos = -13.0; // -13.0 (-21) for r_clumpyregion = 10 (18)
    col = 1.0 - cptr->qa[iqal0+1]/cptr->qa[iqd]; // cloud colour (the ambient medium has iqal0+1 = 1 for its colour)

    // Initially we wish to have 6 coarse cells per shell "width"
    if (t == 0.0){
      float drshell = 1.0; // initialize with radius of cloud 
      for (int i = tnlevs-1; i > ilev; i--){
        difflev = tnlevs-1-i;  //difference between current level (i) and 
                               // desired finest level (tnlevs-1)
        drlev = drfine*pow(2.0,difflev); //current cell size 
        drshell += 6.0*drlev;
        //cout << "i = " << i << "; difflev = " << difflev << "; drlev = " << drlev << "; drshell = " << drshell << "\n";
      }
      //exit(EXIT_FAILURE);
      for (n = 1; n <= ncc; n++){  //loop over clouds
        xdist = cptr->xc[0] - xc1[n];
        ydist = cptr->xc[1] - xc2[n];
        zdist = cptr->xc[2] - xc3[n];
        r = sqrt(xdist*xdist + ydist*ydist + zdist*zdist);
        if (r < drshell) qsol = badcell;
      }
    } // t = 0.0

    // Now limit the refinement based on off-axis distance from the clumpy region (this
    // derefines the shock off-axis)
    xdist = cptr->xc[0];
    ydist = cptr->xc[1];
    zdist = cptr->xc[2];
    axisdist = sqrt(ydist*ydist + zdist*zdist);
    float rclumpy_region = 10.0;
    if (axisdist > rclumpy_region+2.0){
      difflev = (tnlevs-1)-ilev;
      if (int(1.3*axisdist)/(2.0*rclumpy_region) > difflev-1){
        // cell is far enough off-axis to be potentially derefined. Now check colour
        if ((col < 0.15)&&(ilev < tnlevs - 2)) qsol = good;
        else if (col < 0.5) qsol = good;
      }
    }

    // Also limit refinement for cells far downstream. Derefine up to 4 coarse cells into the
    // postshock flow once the shock gets past the specified minimum distance.
    xdist = cptr->xc[0];    
    if (xdist > max(rclumpy_region*2.0, vb*t + initial_shock_pos - 4.0*drcoarse)){
      int limlev = tnlevs - 2;
      if (ilev >= limlev) qsol = good; 
    }    
    if (xdist > max(rclumpy_region*4.0, vb*t + initial_shock_pos - 4.0*drcoarse)){ // this must be an if and not an elseif
      int limlev = tnlevs - 3;
      if (ilev >= limlev) qsol = good; 
    }
 
    // Finally, override all other refinement/derefinement criteria so that cells are
    // maximally refined if col > 0.5 
    //if (col > 0.5) qsol = badcell;
    //else if ((col > 0.1)&&(ilev < tnlevs - 1)) qsol = badcell;

  }
  else if ((strcmp(cas, "FILAMENT") == 0)&&(nd == 3)){
    initial_shock_pos = -10.0;
    col = cptr->qa[iqal0+1]/cptr->qa[iqd]; // cell colour
    //cout << "nlevs = " << nlevs << "; tnlevs = " << tnlevs; exit(EXIT_FAILURE);

    // change angles to radians
    float fil_theta_rad = fil_theta*gkpi/180.0;
    float fil_phi_rad   = fil_phi  *gkpi/180.0;

    // Calculate filament direction vector
    float a = cos(fil_phi_rad)*sin(fil_theta_rad);            // x-direction
    float b = sin(fil_phi_rad)*sin(fil_theta_rad);            // y-direction
    float c =                  cos(fil_theta_rad);            // z-direction

    // Set the refinement at initialization
    if (t == 0.0){

      float drshell = 1.0; // initialize with radius of filament
      for (int i = tnlevs-1; i > ilev; i--){
        difflev = tnlevs-1-i;  //difference between current level (i) and 
                             // desired finest level (tnlevs-1)
        drlev = drfine*pow(2.0,difflev); //current cell size 
        drshell += 6.0*drlev;
        //cout << "i = " << i << "; difflev = " << difflev << "; drlev = " << drlev << "; drshell = " << drshell << "\n";
      }

      // Now determine whether the cell lies within the filament and adjust its parameters appropriately
      // Distance of a point (xc[0],xc[1],xc[2]) from a line (specified in vector form as r = r0 + tA where t is 
      // a constant and A is a direction vector with components (a,b,c)). r0 is the coord of the centre of the filament.
      float rx0 = 0.0; float ry0 = 0.0; float rz0 = 0.0;
      float rpx = cptr->xc[0] - rx0;
      float rpy = cptr->xc[1] - ry0;
      float rpz = cptr->xc[2] - rz0;
      float x = rpy*c - rpz*b;
      float y = -(rpx*c - rpz*a);
      float z = rpx*b - rpy*a;
      float dist = sqrt(x*x + y*y + z*z); // radial distance from filament symmetry axis
      float dist2 = sqrt(rpx*rpx + rpy*rpy + rpz*rpz); // distance from centre of filament
      float dist3 = sqrt(dist2*dist2 - dist*dist);     // projected distance along symmetry axis from centre of filament
      float dist4 = dist3 - fil_length/2.0;            // projected distance beyond filament
      float distr = sqrt(dist*dist + dist4*dist4);     // spherical distance from centre of filament cap
      bool inside_fil = false;

      if (dist3 < fil_length/2.0){
        if (dist < drshell) qsol = badcell;
      }
      else if (distr < drshell) qsol = badcell;

    } // t=0.0


    // fil_com_x = 20.0; fil_com_y = 0.0; fil_com_z = -6.4;

    col = cptr->qa[iqal0+1]/cptr->qa[iqd];

    // Now limit the refinement based on off-axis distance from the filament
    // centre of mass (at the origin at t=0.0). This limits refinement at the shock
    // when t=0.0. Only limit refinement if col < 0.5 or col < 0.15 && ilev < tnlevs -2

    xdist = cptr->xc[0] - fil_com_x;
    ydist = cptr->xc[1] - fil_com_y;
    zdist = cptr->xc[2] - fil_com_z;
    axisdist = sqrt(ydist*ydist + zdist*zdist);
    float fil_size_y = max(2.0,fil_length*b); // projected length of filament on y-axis (2.0 is the diameter of the filament)
    float fil_size_z = max(2.0,fil_length*c); // projected length of filament on z-axis
    if ((fabs(ydist) > 2.0*fil_size_y)||(fabs(zdist) > 2.0*fil_size_z)){
      difflev = (tnlevs-1)-ilev;
      if (int(1.3*max(fabs(ydist/fil_size_y),fabs(zdist/fil_size_z))/2) > difflev-1){
        // cell is far enough off-axis to be potentially derefined. Now check colour
        if ((col < 0.15)&&(ilev < tnlevs - 2)) qsol = good;
        else if (col < 0.5) qsol = good;
      } // off-axis
      //cout << "fil_size_y = " << fil_size_y << "; fil_size_z = " << fil_size_z << "; tnlevs = " << tnlevs << "; ilev = " << ilev << "; difflev = " << difflev << "\n";
      //exit(EXIT_FAILURE);
    }

    // Also limit refinement for cells far downstream. Derefine up to 4 coarse cells into the
    // postshock flow once the shock gets past the specified minimum distance.
    xdist = cptr->xc[0];    
    if (xdist > max(20.0, vb*t + initial_shock_pos - 4.0*drcoarse)){
      int limlev = tnlevs - 2;
      if (ilev >= limlev) qsol = good; 
    }    
    if (xdist > max(40.0, vb*t + initial_shock_pos - 4.0*drcoarse)){ // this must be an if and not an elseif
      int limlev = tnlevs - 3;
      if (ilev >= limlev) qsol = good; 
    }
    
  } 
  
  }//CLOUD or CLUMP or FILAMENT case


  return(qsol);
}
//end usrqsol.c

//file usrte.c
/*
+ usrte - Temperature.
*/
float usrte(float pa[])
{
  // JMP added
  int id;
  float rho,pre,te,ke,preie;
  //return(pa[iqe]/pa[iqd]); // orig

  rho = pa[iqd];
  pre = pa[iqe];
  if (inten){
    ke = 0.0;
    for (id = 1; id <= nd; id++){
      ke += pa[iqu0+id]*pa[iqu0+id];
    }
    ke *= 0.5*rho;
    if (pre < 0.03*ke) pre = pa[iqie]*powf(rho, g); //use pressure from entropy instead
  }
  //use the minimum pressure when estimating the temperature
  //te = min(pre,preie)/rho;
  te = pre/rho;
  return(te);

}
//end usrte.c

//file usrmu.c
/*
+ usrmu - Diffusion coefficients.
*/
void usrmu(float pa[], float mu[])
{
  float rho, tv, tv1, tl, mut, c2, tm2;
  int i;
  /* Zero for safety */
  for (i = 0; i < nq; i++)
    mu[i] = 0.0;
  rho = pa[iqd];
  if (viscous){
    /* Laminar diffusion coefficients */
    for (i = iqu0 + 1; i <= iqu0 + nd; i++)
      mu[i] = rho*vis;
    mu[iqe] = rho*tcond;
    tv = rho*sdiff;
    for (i = iqal0 + 1; i <= iqal0 + scad; i++)
      mu[i] = tv;
    for (i = iqal0 + scad + 1; i <= iqal0 + nal; i++)
      mu[i] = sdiff;
  }
  if (keps){
    /* k-epsilon */
    tv = max(pa[iqk], kmin);
    tv1 = sqrt(tv);
    tl = tv*tv1/max(pa[iqeps],epsmin);
    tl = min(ceps*tl, mtl);
    mut = rho*tv1*tl;

    /* Dash & Wolffe compressibility term */   //JMP added
    mut /=1.0 + 19.0*pa[iqk]*rho/(g*pa[iqe]);//comment out for incompress. flow

    /* Turbulent viscosity */
    tv = cs*mut;
    for (i = iqu0 + 1; i <= iqu0 + nd; i++)
      mu[i] += tv;
    /* Turbulent thermal conduction */
    mu[iqe] += ct*mut;
    /* k diffusion */
    mu[iqk] = tv;
    /* epsilon diffusion */
    mu[iqeps] = ce*mut;
    /* Turbulent scalar diffusion */
    tv = cd*mut;
    for (i = iqal0 + 1; i <= iqal0 + scad; i++)
      mu[i] += tv;
  }
  return;
}
//end usrmu.c

//file usrbd.c
/*
+ usrbd - Boundary conditions.
     Description:
        Sets  the primitive arrays,  pl(left), pr(right)  according to
        the boundary condition specified by the boundary flag, bound.
*/
void usrbd(int id, int bound, join *jptr, float pl[], float pr[])
{
  int i;
  float qa[MAXQ], pa[MAXQ], xc[MAXQ];
  if (bound == lebound){
    /* Left boundary -- default is pl = pr */
    for (i = 0; i < nq; i++)
      pl[i] = pr[i];
    if (bcl[id-1] == bcref)
      /* Symmetry -- reverse normal velocity */
      pl[iqu0+id] = - pr[iqu0+id];
    else if (bcl[id-1] == bcwall)
      /* Wall -- reverse all velocities */
      for (i = 1; i <= nd; i++)
	pl[iqu0+i] = - pr[iqu0+i];
    else if (bcl[id-1] == bcfix){
      /* Fix -- fix solution */
      for (i = 0; i < nd; i++)
	xc[i] = jptr->xj[i];
      xc[id-1] = 2.0*xc[id-1] - jptr->right->xc[id-1];
      usrqis(xc, qa);
      usrpq(qa, pa);
      for (i = 0; i < nq; i++)
	pl[i] = pa[i];
    }
    if ((grav) &&(bcl[id-1] != bcref))
    /* Gravitational potential is zero unless symmetry */
      pl[iqpot] = -pr[iqpot];
  }
  else{
    /* Right boundary -- default is pr = pl */
    for (i = 0; i < nq; i++)
      pr[i] = pl[i];
    if (bcr[id-1] == bcref)
      /* Symmetry -- reverse normal velocity */
      pr[iqu0+id] = - pl[iqu0+id];
    else if (bcr[id-1] == bcwall)
      /* Wall -- reverse all velocities */
      for (i = 1; i <= nd; i++)
	pr[iqu0+i] = - pl[iqu0+i];
    else if (bcr[id-1] == bcfix){
      /* Fix -- fix solution */
      for (i = 0; i < nd; i++)
	xc[i] = jptr->xj[i];
      xc[id-1] = 2.0*xc[id-1] - jptr->left->xc[id-1];
      usrqis(xc, qa);
      usrpq(qa, pa);
      for (i = 0; i < nq; i++)
	pr[i] = pa[i];
    }
    if ((grav) &&(bcr[id-1] != bcref))
    /* Gravitational potential is zero unless symmetry */
      pr[iqpot] = -pl[iqpot];
  }
  return;
}
//end usrbd.c

//file usrcmd.c
/*
+ usrcmd - User commands.
    Description :
     This checks to see if it knows the command string in key, and and
     if necessary  performs the associated actions,  returning true to
     stop other command menus from seeing it.
*/
bool usrcmd(char key[])
{
  float tv;
  char filename[SZNAM];
  bool cmd, found, fail;

  cmd = true;
  if (strcmp(key, "PFLOOR") == 0){
    if (!get_r(pfloor)){
      cout << "Error : reading PFLOOR" << '\n';
      report();
    }
  }
  else if (strcmp(key, "DFLOOR") == 0){
    if (!get_r(dfloor)){
      cout << "Error : reading DFLOOR" << '\n';
      report();
    }
  }
  else if (strcmp(key, "TFLOOR") == 0){
    if (!get_r(tfloor)){
      cout << "Error : reading TFLOOR" << '\n';
      report();
    }
  }
  else if (strcmp(key, "AVISP") == 0){
    if (!get_r(avisp)){
      cout << "Error : reading AVISP" << '\n';
      report();
    }
  }
  else if (strcmp(key, "AVISE") == 0){
    if (!get_r(avise)){
      cout << "Error : reading AVISE" << '\n';
      report();
    }
  }
  else if (strcmp(key, "GAMMA") == 0){
    if (!get_r(g)){
      cout << "Error : reading G" << '\n';
      report();
    }
    else
      usrsetg();
  }
  else if (strcmp(key, "KEPS") == 0){
    keps = true;
    usrkeps();
  }
  else if (strcmp(key, "KMIN") == 0){
    if (get_r(tv) == 0){
      cout << "Error : reading KMIN" << '\n';
      report();
    }
    else{
      kmin = tv;
      if (!keps)
      cout << "Warning : not turbulent" << '\n';
    }
  }
  else if (strcmp(key, "EPSMIN") == 0){
    if (get_r(tv) == 0){
      cout << "Error : reading EPSMIN" << '\n';
      report();
    }
    else{
      epsmin = tv;
      if (!keps)
      cout << "Warning : not turbulent" << '\n';
    }
  }
  else if (strcmp(key, "MTL") == 0){
    if (get_r(tv) == 0){
      cout << "Error : reading MTL" << '\n';
      report();
    }
    else{
      mtl = tv;
      if (!keps)
      cout << "Warning : not turbulent" << '\n';
    }
  }
  else if (strcmp(key, "MON") == 0){
    if (get_n(filename) == 0){
      if (mype == 0)
	cout << "Error : reading monfile name" << '\n';
      report();
    }
    else{
      if (mype == 0){
	strcat(filename,".xq");
	fail = (monfile=fopen(filename, "r"));
	if (fail)
	  fclose(monfile);
      }
      broadcast(&fail);
      if (fail){
	if (mype == 0)
	  cout << "Error: file exists \n";
	report();
      }
      else{
	if (mype == 0)
	  fail = (!(monfile = fopen(filename, "w")));
	broadcast(&fail);
	if (fail){
	  if (mype == 0)
	    cout << "Error: unable to open monitor file \n";
	  report();
	}
	else
	  mon = true;
      }
    }
  }
  else if (strcmp(key, "NOMON") == 0){
    mon = false;
    if ((monfile != NULL) && (mype == 0))
      fclose(monfile);
  }
  else if (strcmp(key, "FXP") == 0)
    fxp = true;
  else if (strcmp(key, "NOFXP") == 0)
    fxp = false;
  else if (strcmp(key, "VISCOUS") == 0)
    viscous = true;
  else if (strcmp(key, "NOVISCOUS") == 0)
    viscous = false;
  else if (strcmp(key, "INTEN") == 0)
    inten = true;
  else if (strcmp(key, "NOINTEN") == 0)
    inten = false;
  else if (strcmp(key, "TCOND") == 0){
    if (get_r(tcond) == 0){
      cout << "Error : reading TCOND" << '\n';
      report();
    }
  }
  else if (strcmp(key, "SDIFF") == 0){
    if (get_r(sdiff) == 0){
      cout << "Error : reading SDIFF" << '\n';
      report();
    }
  } 
  else if (strcmp(key, "VIS") == 0){
    if (get_r(vis) == 0){
      cout << "Error : reading VIS" << '\n';
      report();
    }
  }
  else if (strcmp(key, "GCONST") == 0){
    if (get_r(gconst) == 0){
      cout << "Error : reading GCONST" << '\n';
      report();
    }
  }
  else if (strcmp(key, "PARTICLE") == 0)
    particle = true;
  else if (strcmp(key, "NOPARTICLE") == 0){
    int icell;
    for (icell = 1; icell <= celltop; icell++)
      del_allpart(carr[icell]);
    particle = false;
  }
  else if (strcmp(key, "RAT") == 0){
    if (get_r(rat) == 0){
      cout << "Error : reading RAT" << '\n';
      report();
    }
  } 
  else if (strcmp(key, "HEAT") == 0){
    if (get_r(heat) == 0){
      cout << "Error : reading HEAT" << '\n';
      report();
    }
  }
  //JMP commands
  else if (strcmp(key, "PREFX") == 0){
    if (get_n(prefx) == 0){
      if (mype == 0)
	cout << "Error : reading prefx name" << '\n';
      report();
    }
  }
  else if (strcmp(key, "MONSTATS") == 0){
    if (get_r(tstats) == 0){
      cout  << "Error : reading tstats value" << '\n';
      report();
    }
    else{
      monstats = true;
    }
    //if ((monstats)&&(mype==0)) usrsum();
    cout << "t = " << t << "\n";
    cout << "mype = " << mype  << "\n";
    if ((monstats)&&(t == 0.0)&&(mype==0)) usrmonitor();
  }
  else if (strcmp(key, "NOMONSTATS") == 0)
    monstats = false;
  else if (strcmp(key, "READINCLOUDDATA") == 0){
    if (get_n(cloudfilename) == 0){
      if (mype == 0)
	cout << "Error : reading cloud file name" << '\n';
      report();
    }
    readinclouddata = true;
    //cout << "Will read in " << cloudfilename << "\n";
  }
  else if (strcmp(key, "NCC") == 0){
    if (get_i(ncc) == 0){
      cout << "Error : reading NCC" << '\n';
      report();
    }
  }
  else if (strcmp(key, "NSTARS") == 0){
    if (get_i(nstars) == 0){
      cout << "Error : reading NSTARS" << '\n';
      report();
    }
  }
  else if (strcmp(key, "REMAPCELLS") == 0){
    if (get_i(remapcells) == 0){
      cout << "Error : reading REMAPCELLS" << '\n';
      report();
    }
  }
  else if (strcmp(key, "REMAP1") == 0){
    //cout << "remap1 = " << remap1 << "\n";
    if (get_r(remap1) == 0){
      cout << "Error : reading REMAP1" << '\n';
      report();
    }
  }
  else if (strcmp(key, "REMAP2") == 0){
    if (get_r(remap2) == 0){
      cout << "Error : reading REMAP2" << '\n';
      report();
    }
  }
  else if (strcmp(key, "NCELLSMAP") == 0){
    if (get_i(ncellsmap) == 0){
      cout << "Error : reading NCELLSMAP" << '\n';
      report();
    }
  }
  else if (strcmp(key, "RAMB") == 0){
    if (get_r(ramb) == 0){
      cout << "Error : reading RAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "PAMB") == 0){
    if (get_r(pamb) == 0){
      cout << "Error : reading PAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "TAMB") == 0){
    if (get_r(tamb) == 0){
      cout << "Error : reading TAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "XAMB") == 0){
    if (get_r(xamb) == 0){
      cout << "Error : reading XAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "YAMB") == 0){
    if (get_r(yamb) == 0){
      cout << "Error : reading YAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "ZAMB") == 0){
    if (get_r(zamb) == 0){
      cout << "Error : reading ZAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "UAMB") == 0){
    if (get_r(uamb) == 0){
      cout << "Error : reading UAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "KAMB") == 0){
    if (get_r(kamb) == 0){
      cout << "Error : reading KAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "EPSAMB") == 0){
    if (get_r(epsamb) == 0){
      cout << "Error : reading EPSAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "MACH") == 0){
    if (get_r(mach) == 0){
      cout << "Error : reading MACH" << '\n';
      report();
    }
  }
  else if (strcmp(key, "RSHELL") == 0){
    if (get_r(rshell) == 0){
      cout << "Error : reading RSHELL" << '\n';
      report();
    }
  }
  else if (strcmp(key, "SHOCKPOS") == 0){
    if (get_r(shockpos) == 0){
      cout << "Error : reading SHOCKPOS" << '\n';
      report();
    }
  }
  else if (strcmp(key, "SHELLDZ") == 0){
    if (get_r(shelldz) == 0){
      cout << "Error : reading SHELLDZ" << '\n';
      report();
    }
  }
  else if (strcmp(key, "HOLEDR") == 0){
    if (get_r(holedr) == 0){
      cout << "Error : reading HOLEDR" << '\n';
      report();
    }
  }
  else if (strcmp(key, "NHOLES") == 0){
    if (get_i(nholes) == 0){
      cout << "Error : reading NHOLES" << '\n';
      report();
    }
    //cout << "usrcmd: nholes = " << nholes << "\n";
  }
  else if (strcmp(key, "CHI") == 0){
    if (get_r(chi) == 0){
      cout << "Error : reading CHI" << '\n';
      report();
    }
  }
  else if (strcmp(key, "SLRADIUS") == 0) {
    if (get_r(slradius) == 0){
      cout << "Error: reading SLRADIUS" << '\n';
      report();
    }
    cout << "SLRADIUS: " << slradius << endl;
  }
  else if (strcmp(key, "RHOFUDGE") == 0){
    if (get_r(rhofudge) == 0){
      cout << "Error : reading RHOFUDGE" << '\n';
      report();
    }
    //cout << "rhofudge = " << rhofudge << '\n';
    //exit(EXIT_FAILURE);
  }
  else if (strcmp(key, "CLD1") == 0){
    if (get_r(cld1) == 0){
      cout << "Error : reading CLD1" << '\n';
      report();
    }
  }
  else if (strcmp(key, "CLD2") == 0){
    if (get_r(cld2) == 0){
      cout << "Error : reading CLD2" << '\n';
      report();
    }
  }
  else if (strcmp(key, "MCLOUD") == 0){
    if (get_r(mcloud) == 0){
      cout << "Error : reading MCLOUD" << '\n';
      report();
    }
  }
  else if (strcmp(key, "VCLOUD") == 0){
    if (get_r(vcloud) == 0){
      cout << "Error : reading VCLOUD" << '\n';
      report();
    }
  }
  else if (strcmp(key, "XCMIN") == 0){
    if (get_r(xcmin) == 0){
      cout << "Error : reading XCMIN" << '\n';
      report();
    }
    //cout << "xcmin = " << xcmin << "\n";
  }
  else if (strcmp(key, "XCMAX") == 0){
    if (get_r(xcmax) == 0){
      cout << "Error : reading XCMAX" << '\n';
      report();
    }
  }
  else if (strcmp(key, "YCMIN") == 0){
    if (get_r(ycmin) == 0){
      cout << "Error : reading YCMIN" << '\n';
      report();
    }
  }
  else if (strcmp(key, "YCMAX") == 0){
    if (get_r(ycmax) == 0){
      cout << "Error : reading YCMAX" << '\n';
      report();
    }
  }
  else if (strcmp(key, "ZCMIN") == 0){
    if (get_r(zcmin) == 0){
      cout << "Error : reading ZCMIN" << '\n';
      report();
    }
  }
  else if (strcmp(key, "ZCMAX") == 0){
    if (get_r(zcmax) == 0){
      cout << "Error : reading ZCMAX" << '\n';
      report();
    }
  }
  else if (strcmp(key, "DC") == 0){
    if (get_r(dc) == 0){
      cout << "Error : reading DC" << '\n';
      report();
    }
  }
  else if (strcmp(key, "RCMIN") == 0){
    if (get_r(rcmin) == 0){
      cout << "Error : reading RCMIN" << '\n';
      report();
    }
  }
  else if (strcmp(key, "RCMAX") == 0){
    if (get_r(rcmax) == 0){
      cout << "Error : reading RCMAX" << '\n';
      report();
    }
  }
  else if (strcmp(key, "CHIMIN") == 0){
    if (get_r(chimin) == 0){
      cout << "Error : reading CHIMIN" << '\n';
      report();
    }
  }
  else if (strcmp(key, "CHIMAX") == 0){
    if (get_r(chimax) == 0){
      cout << "Error : reading CHIMAX" << '\n';
      report();
    }
  }
  else if (strcmp(key, "MASSRATIO") == 0){
    if (get_r(massratio) == 0){
      cout << "Error : reading MASSRATIO" << '\n';
      report();
    }
  }
  else if (strcmp(key, "AVGMASSAMB") == 0){
    if (get_r(avgmassamb) == 0){
      cout << "Error : reading AVGMASSAMB" << '\n';
      report();
    }
  }
  else if (strcmp(key, "AVGMASS") == 0){
    if (get_r(avgmass) == 0){
      cout << "Error : reading AVGMASS" << '\n';
      report();
    }
  }
  else if (strcmp(key, "TINJ") == 0){
    if (get_r(tinj) == 0){
      cout << "Error : reading TINJ" << '\n';
      report();
    }
  }
  else if (strcmp(key, "TMIN") == 0){
    if (get_r(tmin) == 0){
      cout << "Error : reading TMIN" << '\n';
      report();
    }
  }
  else if (strcmp(key, "TSTATS") == 0){
    if (get_r(tstats) == 0){
      cout << "Error : reading TSTATS" << '\n';
      report();
    }
  }
  else if (strcmp(key, "MONDSB") == 0){
    if (get_r(dphase_dsbwr_ap) == 0){
      cout << "Error : reading dphase_dsbwr_ap value" << '\n';
      report();
    }
    else{
      mondsb = true;
      t_dsbwr = dphase_dsbwr_ap*period/tscale;
      cout << "t_dsbwr = " << t_dsbwr << "\n";
    }
  }
  else if (strcmp(key, "NOMONDSB") == 0)
    mondsb = false;
  else if (strcmp(key, "RSCALE") == 0){
    if (get_r(rscale) == 0){
      cout << "Error : reading RSCALE" << '\n';
      report();
    }
  }
  else if (strcmp(key, "USCALE") == 0){
    if (get_r(uscale) == 0){
      cout << "Error : reading USCALE" << '\n';
      report();
    }
  }
  else if (strcmp(key, "RHOSCALE") == 0){
    if (get_r(rhoscale) == 0){
      cout << "Error : reading RHOSCALE" << '\n';
      report();
    }
  }
  else if (strcmp(key, "PSCALE") == 0){
    if (get_r(pscale) == 0){
      cout << "Error : reading PSCALE" << '\n';
      report();
    }
  }
  else if (strcmp(key, "TSCALE") == 0){
    if (get_r(tscale) == 0){
      cout << "Error : reading TSCALE" << '\n';
      report();
    }
  }
  else if (strcmp(key, "PICT_COUNT") == 0){
    if (get_i(pict_count) == 0){
      cout << "Error : reading PICT_COUNT" << '\n';
      report();
    }
  }
  else if (strcmp(key, "PREFX") == 0){
    if (get_n(prefx) == 0){
      if (mype == 0)
	cout << "Error : reading prefx name" << '\n';
      report();
    }
  }
  else if (strcmp(key, "COOL") == 0){
    cool = true;
  }
  else if (strcmp(key, "NOCOOL") == 0){
    cool = false;
  }
  else if (strcmp(key, "COOLCURVE1") == 0){
    if (get_n(coolcurve1) == 0){
      if (mype == 0)
	cout << "Error : reading coolcurve1 name" << '\n';
      report();
    }
  }
  else if (strcmp(key, "COOLCURVE2") == 0){
    if (get_n(coolcurve2) == 0){
      if (mype == 0)
	cout << "Error : reading coolcurve2 name" << '\n';
      report();
    }
  }
  else if (strcmp(key, "HEATING") == 0){
    heating = true;
  }
  else if (strcmp(key, "NOHEATING") == 0){
    heating = false;
  }
  else if (strcmp(key, "GAMMAHEAT") == 0){
    if (get_r(gammaheat) == 0){
      cout << "Error : reading GAMMAHEAT" << '\n';
      report();
    }
  }
  else if (strcmp(key, "TEMPMU") == 0){
    tempmu = true;
    read_tempmu_data();
  }
  else if (strcmp(key, "ORBIT") == 0){
    orbit = true;
  }
  else if (strcmp(key, "FORCE") == 0){
    force = true;
  }
  else if (strcmp(key, "PERIOD") == 0){
    if (get_r(period) == 0){
      cout << "Error : reading PERIOD" << '\n';
      report();
    }
  }
  else if (strcmp(key, "ECC") == 0){
    if (get_r(ecc) == 0){
      cout << "Error : reading ECC" << '\n';
      report();
    }
  }
  else if (strcmp(key, "PHASE_OFFSET") == 0){
    if (get_r(phase_offset) == 0){
      cout << "Error : reading PHASE_OFFSET" << '\n';
      report();
    }
  }
  else if (strcmp(key, "RCLOUD") == 0){
    if (get_r(rcloud) == 0){
      cout << "Error : reading RCLOUD" << '\n';
      report();
    }
  }
  else if (strcmp(key, "FRACTAL") == 0){
    fractal = true;
  }
  else if (strcmp(key, "FRACDIM") == 0){
    if (get_r(fracdim) == 0){
      cout << "Error : reading FRACDIM" << '\n';
      report();
    }
    if (((fracdim > 2)&&(nd==2))||((fracdim > 3)&&(nd==3))){
      cout << "Fractal dimension is too high for physical dimension.\n";
      cout << " Fracdim = " << fracdim << "\n";
      cout << " Nd      = " << nd << "\n";
      exit(EXIT_FAILURE);
    }
  }
  else if (strcmp(key, "NCOARSE") == 0){
    if (get_i(ncoarse) == 0){
      cout << "Error : reading NCOARSE" << '\n';
      report();
    }
  }
  else if (strcmp(key, "BE_MASSFRAC") == 0){
    if (get_r(be_massfrac) == 0){
      cout << "Error : reading BE_MASSFRAC" << '\n';
      report();
    }
  }

  //RA: Custom AL
  else if (strcmp(key, "CUST_AL5") == 0) {
    if (get_i(cust_al[0]) == 0){
      cout << "Error : reading CUST_AL5" << '\n';
      report();
    }
  }
  else if (strcmp(key, "CUST_AL6") == 0) {
    if (get_i(cust_al[1]) == 0){
      cout << "Error : reading CUST_AL6" << '\n';
      report();
    }
  }
  else if (strcmp(key, "CUST_AL7") == 0) {
    if (get_i(cust_al[2]) == 0){
      cout << "Error : reading CUST_AL7" << '\n';
      report();
    }
  }
  else if (strcmp(key, "CUST_AL8") == 0) {
    if (get_i(cust_al[3]) == 0){
      cout << "Error : reading CUST_AL8" << '\n';
      report();
    }
  }
  //End of Custom AL

  else if (strcmp(key, "FIL_LENGTH") == 0) {
    if (get_r(fil_length) == 0){
      cout << "Error : reading FIL_LENGTH" << '\n';
      report();
    }
  }
  else if (strcmp(key, "FIL_RAD") == 0) {
    if (get_r(fil_rad) == 0){
      cout << "Error : reading FIL_RAD" << '\n';
      report();
    }
  }
  else if (strcmp(key, "FIL_THETA") == 0) {
    if (get_r(fil_theta) == 0){
      cout << "Error : reading FIL_THETA" << '\n';
      report();
    }
  }
  else if (strcmp(key, "FIL_PHI") == 0) {
    if (get_r(fil_phi) == 0){
      cout << "Error : reading FIL_PHI" << '\n';
      report();
    }
  }
 /* else if (strcmp(key, "EJ_MASS") == 0) {
    if (get_r(ej_mass) == 0){
      cout << "Error : reading EJ_MASS" << '\n';
      report();
    }
  }
  else if (strcmp(key, "EJ_ENERGY") == 0) {
    if (get_r(ej_energy) == 0){
      cout << "Error : reading EJ_ENERGY" << '\n';
      report();
    }
  }
  else if (strcmp(key, "EJ_RADIUS") == 0) {
    if (get_r(ej_radius) == 0){
      cout << "Error : reading EJ_RADIUS" << '\n';
      report();
    }
  } */
  // SVL SILO commands...
  else if (strcmp(key, "SILO")== 0){
    vectors = false;
    if (!get_n(filename)){
      cout << "Error : reading SILO filename" << '\n';
      report();
    }
    else{
      /* silo_out takes number of scalars to output           */
      /* If nothing specified: everything                     */
      if (get_i(silo_out) == 0)
         silo_out = nq - iqal0 + 1;
      else
         silo_out = min(silo_out, nq - iqal0 + 1);
      write_silo(plmax,filename);
    }
  }
  else if (strcmp(key, "SIMAGE")== 0){
    vectors = false;
    if (!get_n(silofile)){
      cout << "Error : reading SILO filename for imaging" << '\n';
      report();
    }
    else{
      if (get_r(idumpdt) == 0){
        cout << "Error : reading time interval for image output" << '\n';
        image = false;
        report();
      } else {
        /* silo_out takes number of scalars to output           */
        /* If nothing specified: everything                     */
        if (get_i(silo_out) == 0)
           silo_out = nq - iqal0 + 1;
        else
           silo_out = min(silo_out, nq - iqal0 + 1);
        idumpt = -idumpdt;
        cycle = 0;
        image = true;
        usrmon();
      }
    }
  }
  else if (strcmp(key, "VIMAGE")== 0){
    vectors = true;
    if (!get_n(silofile)){
      cout << "Error : reading SILO filename for imaging" << '\n';
      report();
    }
    else{
      if (get_r(idumpdt) == 0){
        cout << "Error : reading time interval for image output" << '\n';
        image = false;
        report();
      } else {
        /* silo_out takes number of scalars to output           */
        /* If nothing specified: everything                     */
        if (get_i(silo_out) == 0)
           silo_out = nq - iqal0 + 1;
        else
           silo_out = min(silo_out, nq - iqal0 + 1);
        idumpt = -idumpdt;
        cycle = 0;
        image = true;
        usrmon();
      }
    }
  }

  else
    /* Command not known here, pass it back to caller */
    cmd = false;

  return(cmd);
}
//end usrcmd.c

//file usrprt.c
/*
+ usrprt - Print user information.
    Description :
     Prints user information. It is called by the USERS command.
*/
void usrprt()
{
  cout.setf(ios::scientific);
  cout << setprecision(4);
  cout << "no of scalars " << nal << " no of of advecting scalars "
       << scad << '\n';
  cout << "g " << g << '\n';
  cout << "dfloor " << dfloor << '\n';
  cout << "pfloor " << pfloor << '\n';
  cout << "tfloor " << tfloor << '\n';
  cout << "avisp " << avisp << '\n';
  cout << "avise " << avise << '\n';
  cout << "rat " << rat << '\n';
  cout << "heat " << heat << '\n';
  if (viscous)
    cout << "viscous " << " vis " << vis << " tcond " << tcond << 
      " sdiff " << sdiff <<'\n';
  else 
    cout << "inviscid " << '\n';
  if (keps)
    cout << "k-epsilon: kmin " << kmin << "  epsmin " << epsmin <<
      " max turb length scale mtl " << mtl << '\n';
  if (inten)
    cout << "calculating entropy " << '\n';
  if (particle)
    cout << "particles " << " deleted " << delpart << " particles" << 
      '\n';
  else
    cout << "no particles " << '\n';
  if (grav){
    cout << "gravity nrel " << nrel << " normrat " << 
      normrat << " gconst " << gconst;
    if (boundpot)
      cout << " boundpot levb " <<  levb << '\n';
    else
      cout << " noboundpot " << '\n';
  }
  else
    cout << "no gravity " << '\n';
  if (cool)
    cout << "cooling switched ON\n";
  else
    cout << "cooling switched OFF\n";
  if (mon)
    cout << "monitoring " << '\n';
  else 
    cout << "no monitoring " << '\n';

  if (monstats){
    cout << "stats monitoring " << '\n';
    cout << " tstats = " << tstats << '\n';
    cout << " prefx  = " << prefx << '\n';
  }
  else 
    cout << "no stats monitoring " << '\n';

  //if ((strcmp(cas, "WIND") == 0)||(strcmp(cas, "CWB") == 0)||(strcmp(cas, "MANYCLOUD") == 0)){
  cout << "\n";
  cout << "-------- Problem parameters ---------\n";
  cout << "\n";
  if (strcmp(cas, "WIND") == 0){
    cout << "Number of stars        " << nstars << "\n";
    cout << "Ambient density        " << ramb << " (g/cm^3)\n";
    cout << "Ambient temp           " << tamb << " (K)\n";
    cout << "Ambient mu             " << avgmassamb/massh << " (mH)\n";
    cout << "Minimum temperature    " << tmin << " (K)\n";
    cout << "Remap radius           " << remap1 << " (code units)\n";
  }
  if (strcmp(cas, "BESPHERE") == 0){
    cout << "Bonnor-Ebert sphere, radius rc = " << rcloud*rscale/parsec << " (pc) \n";
    if (fractal){
      cout << " \n";
      cout << " Cloud is fractal\n";
      cout << " Fractal dimension,    D  = " << fracdim << "\n";
      cout << " Coarse structure lengthscale/rc = " << 1.0/float(ncoarse) << "\n";
      cout << " Fraction of BE mass      = " << be_massfrac << "\n";
    }
  }
  if (strcmp(cas, "CWB") == 0){
    cout << "Colliding Winds Binary simulation\n";
    cout << " Parameters of Star 1:\n";
    cout << "  Mass                 " << mass[0] << " (Msol)\n";
    cout << "  Mdot                 " << mdot[0]/(msol/yr) << " (Msol/yr)\n";
    cout << "  vinf                 " << vinf[0]/1.0e5 << " (km/s)\n";
    cout << "  wind temp.           " << tempw[0] << " (K)\n";
    cout << "  H mass frac.         " << massfracx[0] << "\n";
    cout << "  He mass frac.        " << massfracy[0] << "\n";
    cout << "  Metals mass frac.    " << massfracz[0] << "\n";
    cout << "  Average mass         " << avgmass1/massh << " (mH)\n";
    cout << "  Remap radius         " << remap1 << " (code units)\n";
    cout << " Parameters of Star 2:\n";
    cout << "  Mass                 " << mass[1] << " (Msol)\n";
    cout << "  Mdot                 " << mdot[1]/(msol/yr) << " (Msol/yr)\n";
    cout << "  vinf                 " << vinf[1]/1.0e5 << " (km/s)\n";
    cout << "  wind temp.           " << tempw[1] << " (K)\n";
    cout << "  H mass frac.         " << massfracx[1] << "\n";
    cout << "  He mass frac.        " << massfracy[1] << "\n";
    cout << "  Metals mass frac.    " << massfracz[1] << "\n";
    cout << "  Average mass         " << avgmass2/massh << " (mH)\n";
    cout << "  Remap radius         " << remap2 << " (code units)\n";
    cout << " ncellsmap             " << ncellsmap << "\n";
    cout << " Wind momentum ratio   " << eta << "\n";
    cout << "  dsep (cm)            " << dsep*rscale << "\n";
    cout << "  dsep                 " << dsep << "\n";
    cout << "  rob                  " << rob << "\n";
    cout << "  rob/dsep             " << rob/dsep << "\n";
    if (mondsb) cout << " DSB files automatically dumped every " << dphase_dsbwr_ap << " in phase\n";
  }

  if ((strcmp(cas, "WIND") == 0)||(strcmp(cas, "CWB") == 0)){
    if (cool){
      cout << "Cooling on (ie radiative) \n";
      cout << " Coolcurve1 = " << coolcurve1 << "\n";
      if (nstars > 1) cout << " Coolcurve2 = " << coolcurve2 << "\n";
      cout << " Chi1 (dsep) = " << (dsep*rscale/1.0e12)*pow(vinf[0]/1.0e8,4.0)/(mdot[0]/(1.0e-7*msol/yr)) << "\n";
      if (nstars > 1) cout << " Chi2 (dsep) = " << (dsep*rscale/1.0e12)*pow(vinf[1]/1.0e8,4.0)/(mdot[1]/(1.0e-7*msol/yr)) << "\n";
      cout << " Chi1 (Rwr) = " << ((dsep-rob)*rscale/1.0e12)*pow(vinf[0]/1.0e8,4.0)/(mdot[0]/(1.0e-7*msol/yr)) << "\n";
      if (nstars > 1) cout << " Chi2 (Rob) = " << (rob*rscale/1.0e12)*pow(vinf[1]/1.0e8,4.0)/(mdot[1]/(1.0e-7*msol/yr)) << "\n";
    }
    else cout << "Cooling off (ie adiabatic) \n";
  }

  if ((strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "MANYCLOUD") == 0)){
    cout << "\n";
    cout << "Shock-cloud simulation\n";
    cout << "\n";
    cout << "Number of clouds       " << ncc << "\n";
    cout << "Ambient density        " << ramb << " (g/cm^3)\n";
    cout << "Ambient temp           " << tamb << " (K)\n";
    cout << "Shock Mach number      " << mach << "\n";
    if (strcmp(cas, "SINGLECLOUD") == 0){
    cout << "Cloud density contrast " << chi << "\n";
    cout << "Cloud density fudge    " << rhofudge << "\n";
    cout << "Cloud parameter cld1   " << cld1 << "\n";
    cout << "Cloud parameter cld2   " << cld2 << "\n";
    }
    else{
    cout << "Minimum cloud density contrast " << chimin << "\n";
    cout << "Maximum cloud density contrast " << chimax << "\n";
    cout << "Minimum cloud radius           " << rcmin << "\n";
    cout << "Maximum cloud radius           " << rcmax << "\n";
    cout << "Minimum cloud separation       " << dc << "\n";
    cout << "Cloud/intercloud mass ratio    " << massratio << "\n";
    cout << "Cloudy region bounded by:\n";
    cout << " xcmin = " << xcmin << "\n";
    cout << " xcmax = " << xcmax << "\n";
    cout << " ycmin = " << ycmin << "\n";
    cout << " ycmax = " << ycmax << "\n";
    if (nd == 3){
    cout << " zcmin = " << zcmin << "\n";
    cout << " zcmax = " << zcmax << "\n";
    }
    }
    cout << "Average particle mass  " << avgmass/massh << " (mH)\n";
    cout << "Minimum temperature    " << tmin << " (K)\n";
    if (cool) cout << "Cooling on (ie radiative) \n";
    else cout << "Cooling off (ie adiabatic) \n";
    cout << "\n";
  }

  if (strcmp(cas, "FILAMENT") == 0){
    cout << "\n";
    cout << "Shock-filament simulation\n";
    cout << "\n";
    cout << "Ambient density             " << ramb << " (g/cm^3)\n";
    cout << "Ambient temp                " << tamb << " (K)\n";
    cout << "Shock Mach number           " << mach << "\n";
    cout << "Filament density contrast   " << chi << "\n";
    cout << "Filament density fudge      " << rhofudge << "\n";
    cout << "Filament parameter cld1     " << cld1 << "\n";
    cout << "Filament parameter cld2     " << cld2 << "\n";
    cout << "Filament length             " << fil_length << "\n";
    cout << "Filament radius             " << fil_rad << "\n";
    cout << "Filament orientation: theta " << fil_theta << "\n";
    cout << "Filament orientation: phi   " << fil_phi << "\n";
    cout << "Average particle mass       " << avgmass/massh << " (mH)\n";
    cout << "Minimum temperature         " << tmin << " (K)\n";
    if (cool) cout << "Cooling on (ie radiative) \n";
    else cout << "Cooling off (ie adiabatic) \n";
    cout << "\n";
  }

  if (strcmp(cas, "LEAKYSHELL") == 0){
    cout << "\n";
    cout << "Leaky shell simulation\n";
    cout << "\n";
    cout << "Ambient density        " << ramb << " (g/cm^3)\n";
    cout << "Ambient temp           " << tamb << " (K)\n";
    cout << "Shell Mach number      " << mach << "\n";
    cout << "Shell thickness        " << shelldz << "\n";
    cout << "Hole radius            " << holedr << "\n";
    cout << "Average particle mass  " << avgmass/massh << " (mH)\n";
    cout << "Minimum temperature    " << tmin << " (K)\n";
    if (cool) cout << "Cooling on (ie radiative) \n";
    else cout << "Cooling off (ie adiabatic) \n";
    cout << "\n";
  }

  if (strcmp(cas, "SHELL_BLOWOUT") == 0){
    cout << "\n";
    cout << "Shell Blowout simulation\n";
    cout << "\n";
    cout << "Ambient density        " << ramb << " (g/cm^3)\n";
    cout << "Shell   density        " << rshell << " (g/cm^3)\n";
    cout << "Bubble  density        " << ramb/100.0 << " (g/cm^3)\n";
    cout << "Ambient pressure       " << pamb/kboltz << " (K/cm^3)\n";
    cout << "Ambient temp           " << tamb << " (K)\n";
    cout << "Shell   temp           " << tamb*ramb/rshell << " (K)\n";
    cout << "Shell   thickness      " << shelldz*rscale/parsec << " (pc)\n";
    cout << "Number of holes        " << nholes << "\n";
    if (nholes == 1){
      cout << "Hole    radius         " << holedr << "\n";
    }
    else if (nholes > 1){
      for (int n = 0; n < nholes; n++){
        cout << " Hole #" << n << ": xyz pos = " << hole_xpos[n] << " " << hole_ypos[n] << " " << hole_zpos[n] << "; radius = " << hole_radius[n] << "\n"; 
      }
    }
    cout << "Shock Mach number      " << mach << "\n";
    if (!tempmu) cout << "Average particle mass  " << avgmass/massh << " (mH)\n";
    cout << "Minimum temperature    " << tmin << " (K)\n";
    if (cool) cout << "Cooling on (ie radiative) \n";
    else cout << "Cooling off (ie adiabatic) \n";
    if (heating) cout << "Heating on (GammaHeat = " << gammaheat << ")\n";
    else cout << "Heating off \n";
    if (tempmu) cout << "Temperature dependent mu\n";
    cout << "\n";
  }

  if (strcmp(cas, "CWB") == 0){
  if (orbit){ 
    cout << "Orbit on \n";
    cout << " Period       = " << period << " (s)\n";
    cout << " Eccentricity = " << ecc << "\n";
    cout << " Phase offset = " << phase_offset << "\n";
    cout << " Star 1 position:\n";
    cout << "  x           = " << xpos[0] << "\n";
    cout << "  y           = " << ypos[0] << "\n";
    cout << "  z           = " << zpos[0] << "\n";
    cout << "  x           = " << xpos[0]*rscale << " (cm)\n";
    cout << "  y           = " << ypos[0]*rscale << " (cm)\n";
    cout << "  z           = " << zpos[0]*rscale << " (cm)\n";
    cout << " Star 2 position:\n";
    cout << "  x           = " << xpos[1] << "\n";
    cout << "  y           = " << ypos[1] << "\n";
    cout << "  z           = " << zpos[1] << "\n";
    cout << "  x           = " << xpos[1]*rscale << " (cm)\n";
    cout << "  y           = " << ypos[1]*rscale << " (cm)\n";
    cout << "  z           = " << zpos[1]*rscale << " (cm)\n";
  }
  else cout << "Orbit off \n";
  if (force) cout << "Radiatively driven winds \n";
  else cout << "Terminal velocity winds \n";
  }
  cout << "Scales: \n";
  cout << "  rscale   " << rscale << '\n';
  cout << "  rhoscale " << rhoscale << '\n';
  cout << "  uscale   " << uscale << '\n';
  cout << "  pscale   " << pscale << '\n';
  cout << "  tscale   " << tscale << '\n';
  cout << "\n";

  return;
}
//end usrprt.c

//file usrwr.c
/*
+ usrwr - Write user information.
    Description :
     Writes  user information to  a model  file. It  is called  by the
     WRITE command.
*/
void usrwr()
{
  write_var("iqd", iqd);
  write_var("iqu0", iqu0);
  write_var("iqe", iqe);
  write_var("delpart", delpart);
  write_var("iqal0", iqal0);
  write_var("iqie", iqie);
  write_var("iqpot", iqpot);
  write_var("keps", keps);
  if (keps){
    write_var("iqk", iqk);
    write_var("iqeps", iqeps);
    write_var("ce1", ce1);
    write_var("ce2", ce2);
    write_var("cs", cs);
    write_var("ct", ct);
    write_var("cd", cd);
    write_var("ce", ce);
    write_var("ceps", ceps);
    write_var("kmin", kmin);
    write_var("epsmin", epsmin);
    write_var("mtl", mtl);
  }
  write_var("g", g);
  write_var("pfloor", pfloor);
  write_var("dfloor", dfloor);
  write_var("avisp", avisp);
  write_var("avise", avise);
  write_var("mon", mon);
  write_var("fxp", fxp);
  write_var("vis", vis);
  write_var("inten", inten);
  write_var("tcond", tcond);
  write_var("sdiff", sdiff);
  write_var("viscous", viscous);
  write_var("gconst", gconst);
  write_var("rat", rat);
  write_var("heat", heat);
  //JMP info
  write_var("tfloor", tfloor);
  write_var("tnlevs", tnlevs);
  write_var("nstars", nstars);
  write_var("remapcells", remapcells);
  write_var("remap1", remap1);
  write_var("remap2", remap2);
  write_var("ramb", ramb);
  write_var("tamb", tamb);
  write_var("xamb", xamb);
  write_var("yamb", yamb);
  write_var("zamb", zamb);
  write_var("avgmassamb", avgmassamb);
  write_var("tmin", tmin);
  write_var("monstats", monstats);
  write_var("tstats", tstats);
  write_var("timest", timest);
  write_var("rscale", rscale);
  write_var("uscale", uscale);
  write_var("rhoscale", rhoscale);
  write_var("pscale", pscale);
  write_var("tscale", tscale);
  write_var("pict_count", pict_count);
  write_var("prefx", prefx);
  write_var("cool", cool);
  write_var("coolcurve1", coolcurve1);
  write_var("coolcurve2", coolcurve2);
  write_var("heating", heating);
  write_var("gammaheat", gammaheat);
  write_var("tempmu", tempmu);
  //Shock-cloud specific
  write_var("ncc", ncc);
  write_var("pamb", pamb);
  write_var("uamb", uamb);
  write_var("kamb", kamb);
  write_var("epsamb", epsamb);
  write_var("mach", mach);
  write_var("chi", chi);
  write_var("rhofudge", rhofudge);
  write_var("cld1", cld1);
  write_var("cld2", cld2);
  write_var("avgmass", avgmass);
  write_var("mcloud", mcloud);
  write_var("vcloud", vcloud);
  write_var("xcmin", xcmin);
  write_var("xcmax", xcmax);
  write_var("ycmin", ycmin);
  write_var("ycmax", ycmax);
  write_var("zcmin", zcmin);
  write_var("zcmax", zcmax);
  write_var("dc", dc);
  write_var("rcmin", rcmin);
  write_var("rcmax", rcmax);
  write_var("chimin", chimin);
  write_var("chimax", chimax);
  write_var("massratio", massratio);
  //Filament specific
  write_var("fil_length", fil_length);
  write_var("fil_rad", fil_rad);
  write_var("fil_theta", fil_theta);
  write_var("fil_phi", fil_phi);
  write_var("fil_com_x", fil_com_x);
  write_var("fil_com_y", fil_com_y);
  write_var("fil_com_z", fil_com_z);

  //RA: slab
  write_var("slradius", slradius);
  //Wind specific
  //Bonnor-Ebert sphere specific
  write_var("rcloud", rcloud);
  write_var("fractal", fractal);
  write_var("fracdim", fracdim);
  write_var("ncoarse", ncoarse);
  write_var("be_massfrac", be_massfrac);
  //CWB specific
  write_var("eta", eta);
  write_var("rob", rob);
  write_var("avgmass1", avgmass1);
  write_var("avgmass2", avgmass2);
  write_var("orbit", orbit);
  write_var("force", force);
  write_var("period", period);
  write_var("ecc", ecc);
  write_var("phase_offset", phase_offset);
  write_var("dsep", dsep);
  write_var("mondsb", mondsb);
  write_var("dphase_dsbwr_ap", dphase_dsbwr_ap);
  write_var("t_dsbwr", t_dsbwr);
  write_var("timedsb", timedsb);
  write_var("ncellsmap", ncellsmap);
  //Now the arrays
  write_var("mass", nstars, mass);
  write_var("mdot", nstars, mdot);
  write_var("vinf", nstars, vinf);
  write_var("tempw", nstars, tempw);
  write_var("massfracx", nstars, massfracx);
  write_var("massfracy", nstars, massfracy);
  write_var("massfracz", nstars, massfracz);
  write_var("xpos", nstars, xpos);
  write_var("ypos", nstars, ypos);
  write_var("zpos", nstars, zpos);
  write_var("xvel", nstars, xvel);
  write_var("yvel", nstars, yvel);
  write_var("zvel", nstars, zvel);

  // Shell-blowout specific data and arrays
  write_var("rshell", rshell);
  write_var("shelldz", shelldz);
  write_var("nholes", nholes);
  write_var("holedr", holedr);
  write_var("hole_xpos", nholes, hole_xpos);
  write_var("hole_ypos", nholes, hole_ypos);
  write_var("hole_zpos", nholes, hole_zpos);
  write_var("hole_radius", nholes, hole_radius);
 
  // SVL SILO code
  write_var("silofile", silofile);
  write_var("image", image);
  write_var("vectors", vectors);
  write_var("silo_out", silo_out);
  write_var("cycle", cycle);
  write_var("idumpt", idumpt);
  write_var("idumpdt", idumpdt);
  //write_var("", );

  return;
}
//end usrwr.c

//file usrrd.c
/*
+ usrrd - Read user information.
    Description :
     Reads user information to a model  file. It is called by the READ
     command.
*/
bool usrrd(char vname[])
{
  bool found;
  found = true;
  if (strcmp(vname, "iqd") == 0)
    read_var(&iqd);
  else if (strcmp(vname, "iqu0") == 0)
    read_var(&iqu0);
  else if (strcmp(vname, "iqe") == 0)
    read_var(&iqe);
  else if (strcmp(vname, "keps") == 0)
    read_var(&keps);
  else if (strcmp(vname, "iqk") == 0)
    read_var(&iqk);
  else if (strcmp(vname, "iqeps") == 0)
    read_var(&iqeps);
  else if (strcmp(vname, "iqal0") == 0)
    read_var(&iqal0);
  else if (strcmp(vname, "iqie") == 0)
    read_var(&iqie);
  else if (strcmp(vname, "delpart") == 0)
    read_var(&delpart);
  else if (strcmp(vname, "iqpot") == 0)
    read_var(&iqpot);
  else if (strcmp(vname, "ce1") == 0)
    read_var(&ce1);
  else if (strcmp(vname, "ce2") == 0)
    read_var(&ce2);
  else if (strcmp(vname, "cs") == 0)
    read_var(&cs);
  else if (strcmp(vname, "ct") == 0)
    read_var(&ct);
  else if (strcmp(vname, "cd") == 0)
    read_var(&cd);
  else if (strcmp(vname, "ce") == 0)
    read_var(&ce);
  else if (strcmp(vname, "ceps") == 0)
    read_var(&ceps);
  else if (strcmp(vname, "kmin") == 0)
    read_var(&kmin);
  else if (strcmp(vname, "epsmin") == 0)
    read_var(&epsmin);
  else if (strcmp(vname, "mtl") == 0)
    read_var(&mtl);
  else if (strcmp(vname, "g") == 0)
      read_var(&g);
  else if (strcmp(vname, "pfloor") == 0)
    read_var(&pfloor);
  else if (strcmp(vname, "dfloor") == 0)
    read_var(&dfloor);
  else if (strcmp(vname, "avisp") == 0)
    read_var(&avisp);
  else if (strcmp(vname, "avise") == 0)
    read_var(&avise);
  else if (strcmp(vname, "mon") == 0)
    read_var(&mon);
  else if (strcmp(vname, "fxp") == 0)
    read_var(&fxp);
  else if (strcmp(vname, "vis") == 0)
    read_var(&vis);
  else if (strcmp(vname, "tcond") == 0)
    read_var(&tcond);
  else if (strcmp(vname, "sdiff") == 0)
    read_var(&sdiff);
  else if (strcmp(vname, "gconst") == 0)
    read_var(&gconst);
  else if (strcmp(vname, "viscous") == 0)
    read_var(&viscous);
  else if (strcmp(vname, "inten") == 0)
    read_var(&inten);
  else if (strcmp(vname, "rat") == 0)
    read_var(&rat);
  else if (strcmp(vname, "heat") == 0)
    read_var(&heat);
  //JMP variables
  else if (strcmp(vname, "tfloor") == 0)
    read_var(&tfloor);
  else if (strcmp(vname, "tnlevs") == 0)
    read_var(&tnlevs);
  else if (strcmp(vname, "nstars") == 0)
    read_var(&nstars);
  else if (strcmp(vname, "remapcells") == 0)
    read_var(&remapcells);
  else if (strcmp(vname, "remap1") == 0)
    read_var(&remap1);
  else if (strcmp(vname, "remap2") == 0)
    read_var(&remap2);
  else if (strcmp(vname, "ramb") == 0)
    read_var(&ramb);
  else if (strcmp(vname, "tamb") == 0)
    read_var(&tamb);
  else if (strcmp(vname, "xamb") == 0)
    read_var(&xamb);
  else if (strcmp(vname, "yamb") == 0)
    read_var(&yamb);
  else if (strcmp(vname, "zamb") == 0)
    read_var(&zamb);
  else if (strcmp(vname, "avgmassamb") == 0)
    read_var(&avgmassamb);
  else if (strcmp(vname, "tmin") == 0)
    read_var(&tmin);
  else if (strcmp(vname, "monstats") == 0)
    read_var(&monstats);
  else if (strcmp(vname, "tstats") == 0)
    read_var(&tstats);
  else if (strcmp(vname, "timest") == 0)
    read_var(&timest);
  else if (strcmp(vname, "rscale") == 0)
    read_var(&rscale);
  else if (strcmp(vname, "uscale") == 0)
    read_var(&uscale);
  else if (strcmp(vname, "rhoscale") == 0)
    read_var(&rhoscale);
  else if (strcmp(vname, "pscale") == 0)
    read_var(&pscale);
  else if (strcmp(vname, "tscale") == 0)
    read_var(&tscale);
  else if (strcmp(vname, "pict_count") == 0)
    read_var(&pict_count);
  else if (strcmp(vname, "prefx") == 0)
    read_var(&*prefx);
  else if (strcmp(vname, "cool") == 0)
    read_var(&cool);
  else if (strcmp(vname, "coolcurve1") == 0)
    read_var(&*coolcurve1);
  else if (strcmp(vname, "coolcurve2") == 0)
    read_var(&*coolcurve2);
  else if (strcmp(vname, "heating") == 0)
    read_var(&heating);
  else if (strcmp(vname, "gammaheat") == 0)
    read_var(&gammaheat);
  else if (strcmp(vname, "tempmu") == 0)
    read_var(&tempmu);
  //Cloud specific
  else if (strcmp(vname, "ncc") == 0)
    read_var(&ncc);
  else if (strcmp(vname, "pamb") == 0)
    read_var(&pamb);
  else if (strcmp(vname, "uamb") == 0)
    read_var(&uamb);
  else if (strcmp(vname, "kamb") == 0)
    read_var(&kamb);
  else if (strcmp(vname, "epsamb") == 0)
    read_var(&epsamb);
  else if (strcmp(vname, "mach") == 0) 
    read_var(&mach);
  else if (strcmp(vname, "chi") == 0)
    read_var(&chi);
  else if (strcmp(vname, "rhofudge") == 0)
    read_var(&rhofudge);
  else if (strcmp(vname, "cld1") == 0)
    read_var(&cld1);
  else if (strcmp(vname, "cld2") == 0)
    read_var(&cld2);
  else if (strcmp(vname, "avgmass") == 0)
    read_var(&avgmass);
  else if (strcmp(vname, "mcloud") == 0)
    read_var(&mcloud);
  else if (strcmp(vname, "vcloud") == 0)
    read_var(&vcloud);
  else if (strcmp(vname, "xcmin") == 0)
    read_var(&xcmin);
  else if (strcmp(vname, "xcmax") == 0)
    read_var(&xcmax);
  else if (strcmp(vname, "ycmin") == 0)
    read_var(&ycmin);
  else if (strcmp(vname, "ycmax") == 0)
    read_var(&ycmax);
  else if (strcmp(vname, "zcmin") == 0)
    read_var(&zcmin);
  else if (strcmp(vname, "zcmax") == 0)
    read_var(&zcmax);
  else if (strcmp(vname, "dc") == 0)
    read_var(&dc);
  else if (strcmp(vname, "rcmin") == 0)
    read_var(&rcmin);
  else if (strcmp(vname, "rcmax") == 0)
    read_var(&rcmax);
  else if (strcmp(vname, "chimin") == 0)
    read_var(&chimin);
  else if (strcmp(vname, "chimax") == 0)
    read_var(&chimax);
  else if (strcmp(vname, "massratio") == 0)
    read_var(&massratio);
  // Filament specific
  else if (strcmp(vname, "fil_length") == 0)
    read_var(&fil_length);
  else if (strcmp(vname, "fil_rad") == 0)
    read_var(&fil_rad);
  else if (strcmp(vname, "fil_theta") == 0)
    read_var(&fil_theta);
  else if (strcmp(vname, "fil_phi") == 0)
    read_var(&fil_phi);
  else if (strcmp(vname, "fil_com_x") == 0)
    read_var(&fil_com_x);
  else if (strcmp(vname, "fil_com_y") == 0)
    read_var(&fil_com_y);
  else if (strcmp(vname, "fil_com_z") == 0)
    read_var(&fil_com_z);
  //RA:slab
  else if (strcmp(vname, "slradius") == 0)
    read_var(&slradius);
  //Wind specific
  //Bonnor-Ebert sphere specific
  else if (strcmp(vname, "rcloud") == 0)
    read_var(&rcloud);
  else if (strcmp(vname, "fractal") == 0)
    read_var(&fractal);
  else if (strcmp(vname, "fracdim") == 0)
    read_var(&fracdim);
  else if (strcmp(vname, "be_massfrac") == 0)
    read_var(&be_massfrac);
  else if (strcmp(vname, "ncoarse") == 0)
    read_var(&ncoarse);
  //CWB specific
  else if (strcmp(vname, "eta") == 0)
    read_var(&eta);
  else if (strcmp(vname, "rob") == 0)
    read_var(&rob);
  else if (strcmp(vname, "avgmass1") == 0)
    read_var(&avgmass1);
  else if (strcmp(vname, "avgmass2") == 0)
    read_var(&avgmass2);
  else if (strcmp(vname, "orbit") == 0)
    read_var(&orbit);
  else if (strcmp(vname, "force") == 0)
    read_var(&force);
  else if (strcmp(vname, "period") == 0)
    read_var(&period);
  else if (strcmp(vname, "ecc") == 0)
    read_var(&ecc);
  else if (strcmp(vname, "phase_offset") == 0)
    read_var(&phase_offset);
  else if (strcmp(vname, "dsep") == 0)
    read_var(&dsep);
  else if (strcmp(vname, "mondsb") == 0)
    read_var(&mondsb);
  else if (strcmp(vname, "dphase_dsbwr_ap") == 0)
    read_var(&dphase_dsbwr_ap);
  else if (strcmp(vname, "t_dsbwr") == 0)
    read_var(&t_dsbwr);
  else if (strcmp(vname, "timedsb") == 0)
    read_var(&timedsb);
  else if (strcmp(vname, "ncellsmap") == 0)
    read_var(&ncellsmap);
  // Shell-blowout specific
  else if (strcmp(vname, "rshell") == 0)
    read_var(&rshell);
  else if (strcmp(vname, "shelldz") == 0)
    read_var(&shelldz);
  else if (strcmp(vname, "nholes") == 0)
    read_var(&nholes);
  else if (strcmp(vname, "holedr") == 0)
    read_var(&holedr);
  //Now read arrays
  else if (strcmp(vname, "mass") == 0)
    read_var(mass);
  else if (strcmp(vname, "mdot") == 0)
    read_var(mdot);
  else if (strcmp(vname, "vinf") == 0)
    read_var(vinf);
  else if (strcmp(vname, "tempw") == 0)
    read_var(tempw);
  else if (strcmp(vname, "massfracx") == 0){
    read_var(massfracx);
    //for (n = 0; n < nstars; n++){
    //  cout << "massfracx[" << n << "] = " << massfracx[n] << "\n";
    //}
    //exit(EXIT_FAILURE);
  }
  else if (strcmp(vname, "massfracy") == 0)
    read_var(massfracy);
  else if (strcmp(vname, "massfracz") == 0)
    read_var(massfracz);
  else if (strcmp(vname, "xpos") == 0)
    read_var(xpos);
  else if (strcmp(vname, "ypos") == 0){
    read_var(ypos);
    //for (n = 0; n < nstars; n++){
    //  cout << "ypos[" << n << "] = " << ypos[n] << "\n";
    //}
    //exit(EXIT_FAILURE);
  }
  else if (strcmp(vname, "zpos") == 0)
    read_var(zpos);
  else if (strcmp(vname, "xvel") == 0)
    read_var(xvel);
  else if (strcmp(vname, "yvel") == 0)
    read_var(yvel);
  else if (strcmp(vname, "zvel") == 0)
    read_var(zvel);
  else if (strcmp(vname, "hole_xpos") == 0)
    read_var(hole_xpos);
  else if (strcmp(vname, "hole_ypos") == 0)
    read_var(hole_ypos);
  else if (strcmp(vname, "hole_zpos") == 0)
    read_var(hole_zpos);
  else if (strcmp(vname, "hole_radius") == 0)
    read_var(hole_radius);
  //else if (strcmp(vname, "") == 0)
  //  read_var(&);
  // SVL SILO code
  else if (strcmp(vname, "silofile") == 0)
    read_var(silofile);
  else if (strcmp(vname, "image") == 0)
    read_var(&image);
  else if (strcmp(vname, "vectors") == 0)
    read_var(&vectors);
  else if (strcmp(vname, "silo_out") == 0)
    read_var(&silo_out);
  else if (strcmp(vname, "cycle") == 0)
    read_var(&cycle);
  else if (strcmp(vname, "idumpt") == 0)
    read_var(&idumpt);
  else if (strcmp(vname, "idumpdt") == 0)
    read_var(&idumpdt);
  // End of SVL SILO code
  else
    /* Unknown here  */
    found = false;
  /* Check for turbulence */
  keps = iqk > 0;
  /* Set gamma constants for nonlinear Riemann solver */
  usrsetg();

  // Re-read the data for the cooling upon calling radiate 
  first_radiate = true;

  // Re-read the temperature dependent mu data
  if (tempmu) read_tempmu_data();

  /*
  // Re-read the stellar data
  if (strcmp(cas, "WIND") == 0){
    ifstream file("stars_info.dat");  
    if (!file) {
      cerr << "Failed to open file\n";
      exit(EXIT_FAILURE);
    } 
    file.seekg(0); // goto start of file
    file >> nstars_check;   // number of stars
    //read stellar data. Format of file is:
     //id, xpos, ypos, zpos, mass, mdot, vinf, twind, massfracx, massfracy, massfracz
    for (n = 0; n < nstars; n++){
      file >> cnt >> xpos[n] >> ypos[n] >> zpos[n] >> mass[n] >> mdot[n] >> vinf[n] >> tempw[n] >> massfracx[n] >> massfracy[n] >> massfracz[n];
    }
    file.close();
    //convert to cgs units
    for (n = 0; n < nstars; n++){ 
      mdot[n] *= msol/yr;
    }
  }
  */

  return(found);
}
//end usrrd.c

//file usrsum.c
/*
+ usrsum - sum variables over the grid.
     Description :
        Calculates the sum of variables over the grid. The sum is over
        levels plmin to plmax.
*/
void usrsum()
{
  int i, ilev;
  float cvol, tot[MAXQ], pa[MAXQ]; 
  struct cell *cptr;
  //JMP added
  bool rzp,zrp,rtp,xyz;
  char cnum[5], statsfile[60], tempfile[60];
  int n;
  float dens,col,dr,dx,dy,dz,r,x,y,z,vr,vx,vy,vz,v, pres;
  float windmass[NSTARSMAX],windke[NSTARSMAX],windie[NSTARSMAX],windvol[NSTARSMAX];
  float mass,vol,inte,tote;
  float ambmass,ambke,ambie,ambvol,ambcol,totcol;
  double volnorm,massnorm,energynorm;

  FILE *fptr; //pointer to file (the "FILE" must be uppercase)

  int id, nrho, nbin, ithcl, j, index, cnt;;
  const int nmax = 2000;
  const int nvmax = 400;

  float rhomin, rhomax, dlogrho, densthcl, rho;
  float rms, zms, zm, vrms, vzms, vzm, volc, a, c, dvr, dvz, rhom, cs, vb;
  float rhom0, massi, cvol2, circ, omega, divdx, dudx[MAXD][MAXD];
  float divdr, divdz, ul, ur;
  float mc, mcol, colamb, mz, mz2, mr2, mvr2, mvz, mvz2, vc, vc2, tcc;

  float mr2_2, mz2_2, mz_2, mvr2_2, mvz2_2, mvz_2, mcol_2, volc_2;
  float rms_2, zms_2, zm_2, vrms_2, vzms_2, vzm_2;
  float a_2, c_2, mc_2, vc_2, vc2_2, rhom_2, dvz_2, dvr_2;

  float mr2_3, mz2_3, mz_3, mvr2_3, mvz2_3, mvz_3, mcol_3, volc_3;
  float rms_3, zms_3, zm_3, vrms_3, vzms_3, vzm_3;
  float a_3, c_3, mc_3, vc_3, vc2_3, rhom_3, dvz_3, dvr_3;

  //float mx,my,mx2,my2,mvx,mvx2,mvy2;
  //float mx_2,my_2,mx2_2,my2_2,mvx_2,mvx2_2,mvy2_2;
  //float mx_3,my_3,mx2_3,my2_3,mvx_3,mvx2_3,mvy2_3;
  float omegax,omegay,omegaz;

  double ke, kie, ie, the, tue, etot, escale, tkie, tthe, ttue;

  float rhob[nmax], massb[nmax], rhob_single[10][nmax], massb_single[10][nmax], rhozdV[nmax], rhoz2dV[nmax], rhor2dV[nmax],
    rhovr2dV[nmax], rhovzdV[nmax], rhovz2dV[nmax], dvol[nmax], velb[nmax];
  //float logt[ntmax], e1[nemax], e2[nemax], ecen[nemax], de[nemax], dlogt, dloge;
  float pre, u, temp, maxtemp;
  //double lambda[ntmax][nemax];
  //double spec[nemax];
  double scale, rhosqV, gr, em, 
    luma, lumb, lumc, lumtot;

  float meanrho, meanvr, meanvz, meanvel, meanpre, totmass;
  float varvel, varrho, varpre, vel, jump, ambp;
  float zmin, zmax;

  //variables for Halpha calculation
  double n_h,n_e,te,logte,logemisshalpha_caseA,logemisshalpha_caseB,
    emsvty,htotcolionrate,htotrecombcaseA,ratio,onesubx,ionfrac,
    dvel,velmin,velmax,emiss,totemiss;
  float velbl[nvmax];
  double halpha_spec[nvmax]; 

  char emissfile[50];

  int ne, nt, idum, isp;

  int vd, stepl, stepr;
  struct cell *left, *right;
  struct join *pjptr;
  struct join *jptr;
  float pl[MAXQ], pr[MAXQ], ql[MAXQ], qr[MAXQ], lu, ru, dvz_dr, dvr_dz;

  int nfgz,nfgr;
  const int max_nfgz=120;
  const int max_nfgr=120;
  int ntf,ndvr,ndvz,ntfz;
  float fgzmax,dfgz,fgrmin,dfgr,v2,turbfrac,maxturbfrac,dvravg,dvzavg,tfavg;
  float vz2,turbfracz,maxturbfracz,tfzavg;
  float mc_2fg[max_nfgr][max_nfgz], mvz_2fg[max_nfgr][max_nfgz], 
    mvr2_2fg[max_nfgr][max_nfgz], mvz2_2fg[max_nfgr][max_nfgz],
    vz_2fg[max_nfgr][max_nfgz], vz2_2fg[max_nfgr][max_nfgz],
    vr2_2fg[max_nfgr][max_nfgz], dvzfg[max_nfgr][max_nfgz],
    dvrfg[max_nfgr][max_nfgz], dv2[max_nfgr][max_nfgz], 
    mv_2fg[max_nfgr][max_nfgz], v_2fg[max_nfgr][max_nfgz];
  //manycloud variables follow
  const int mcmax=10;
  int nf;
  char numc[5], numc2[5];
  float mc_cloud[mcmax],mx_cloud[mcmax],my_cloud[mcmax],mz_cloud[mcmax],
    mx2_cloud[mcmax],my2_cloud[mcmax],mz2_cloud[mcmax],mvx_cloud[mcmax],
    mvy_cloud[mcmax],mvz_cloud[mcmax],mvx2_cloud[mcmax],mvy2_cloud[mcmax],
    mvz2_cloud[mcmax],vc_cloud[mcmax];
  float mc_core[mcmax],mx_core[mcmax],my_core[mcmax],mz_core[mcmax],
    mx2_core[mcmax],my2_core[mcmax],mz2_core[mcmax],mvx_core[mcmax],
    mvy_core[mcmax],mvz_core[mcmax],mvx2_core[mcmax],mvy2_core[mcmax],
    mvz2_core[mcmax],vc_core[mcmax];
  double kiemc[mcmax],themc[mcmax],tuemc[mcmax];
  double kiecld[mcmax],thecld[mcmax],tuecld[mcmax],etotcld[mcmax];
  float mcol_cloud[mcmax],mcol_core[mcmax];
  float xm_cld[mcmax],ym_cld[mcmax],zm_cld[mcmax],x2m_cld[mcmax],
    y2m_cld[mcmax],z2m_cld[mcmax],vxm_cld[mcmax],vx2m_cld[mcmax],
    vy2m_cld[mcmax],vz2m_cld[mcmax],dx_cld[mcmax],dvx_cld[mcmax],
    dvy_cld[mcmax],dvz_cld[mcmax],rhom_cld[mcmax],
    mc_cld[mcmax],vc_cld[mcmax],mx_cld[mcmax],my_cld[mcmax],mz_cld[mcmax],
    mx2_cld[mcmax],my2_cld[mcmax],mz2_cld[mcmax],mvx_cld[mcmax],
    mvy_cld[mcmax],mvz_cld[mcmax],mvx2_cld[mcmax],mvy2_cld[mcmax],
    mvz2_cld[mcmax],dy_cld[mcmax],dz_cld[mcmax];
  float xm_cor[mcmax],ym_cor[mcmax],zm_cor[mcmax],x2m_cor[mcmax],
    y2m_cor[mcmax],z2m_cor[mcmax],vxm_cor[mcmax],vx2m_cor[mcmax],
    vy2m_cor[mcmax],vz2m_cor[mcmax],dx_cor[mcmax],dvx_cor[mcmax],
    dvy_cor[mcmax],dvz_cor[mcmax],rhom_cor[mcmax],
    mc_cor[mcmax],vc_cor[mcmax],mx_cor[mcmax],my_cor[mcmax],mz_cor[mcmax],
    mx2_cor[mcmax],my2_cor[mcmax],mz2_cor[mcmax],mvx_cor[mcmax],
    mvy_cor[mcmax],mvz_cor[mcmax],mvx2_cor[mcmax],mvy2_cor[mcmax],
    mvz2_cor[mcmax],dy_cor[mcmax],dz_cor[mcmax];
  double kietot,thetot,tuetot;
  //RA variables to track shock front positon:
  //  int yint, zint;
  //int ysize = int( nc[1]*pow(2.0, nlevs -1) );
  //bool shockPosY[ysize+10][1];   //b[y][z] 
  //float shockPosX[ysize+10][1];
  //  cerr << ysize << endl;
//x[y][z]
  //initialize array to 0, false
  //int iMax, jMax, iMin, jMin; // max y and z parameters.
  //iMax = jMax = 0;
  //iMin = jMin = 99999;    
  //for (i = 0; i < ysize; i++) {
  // for (j = 0; j < 1; j++) {
  //  shockPosX[i][j] = shockPosY[i][j] = 0;
  //}
  //}
  // int iMax=0;
  //int iSum=0;
  // more general variables would be (improve code accordingly): shockPosX[iy][iz], shockPosY[iy], shockPosZ[iz]
  // however, only sum up to shockPos[imax] as less cells could be refined, possibly (unlikely, shock always at max refinemet. Investigate :
  // In current simulations shock front always maximally refined. Only check highest refinement level - Allows to compare Y
  //double shockMaxX, shockMinX, shockAveX; 
  //shockMaxX = shockMinX = shockAveX =0;
  //float presNoShock, presNoShockMax = 0, presShock, presShockMin = 99999999;

  // Filament specific
  float mx, my, mx2, my2, mvx, mvy, mvr, mvx2, mvy2, xm, ym, rm, xms, yms, vsm, vym, vrm, vxms, vyms, vxm, mr, dvx, dvy;
  float mx_2, my_2, mx2_2, my2_2, mvx_2, mvy_2, mvr_2, mvx2_2, mvy2_2, xm_2, ym_2, rm_2, xms_2, yms_2, vsm_2, vym_2, vrm_2, vxms_2, vyms_2, vxm_2, mr_2, dvx_2, dvy_2;
  float mx_3, my_3, mx2_3, my2_3, mvx_3, mvy_3, mvr_3, mvx2_3, mvy2_3, xm_3, ym_3, rm_3, xms_3, yms_3, vsm_3, vym_3, vrm_3, vxms_3, vyms_3, vxm_3, mr_3, dvx_3, dvy_3;

  xyz = (strcmp(geom, "XYZ") == 0);
  rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
  zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
  rtp = (strcmp(geom, "RTP") == 0);

  /* Return if no model */
  if (nom)
    return;
  /* Zero sum, */
  for (i = 0; i < nq; i++)
    tot[i] = 0.0;


  //Different CASES have their own unique summation. SHELL_BLOWOUT is first, then WIND, then the
  //SINGLECLOUD and MANYCLOUD cases, then the FILAMENT case
  if (strcmp(cas, "SHELL_BLOWOUT") == 0){
    double shell_mass = 0.0;
    double shell_mtm  = 0.0;
    double shell_mass_hi = 0.0;
    double shell_mtm_hi  = 0.0;
    double mflux = 0.0;
    double mtmflux = 0.0;
    double eflux = 0.0;
    for (ilev = 0; ilev <= nlevs-1; ilev++){ //sum over ALL grids
      cptr = lev[ilev]->fcell;
      while (cptr != NULL){
	if (cptr->child[0] == NULL){ //if finest cell then ...
	  cvol = cptr->vol;
	  dens = cptr->qa[iqd];
	  float col1 = cptr->qa[iqal0+1]/dens; // postshock material
	  float col2 = cptr->qa[iqal0+2]/dens; // shell material
	  mass = dens*cvol;
          pre  = cptr->pa[iqe];
	  if (rzp){
	    dr = lev[ilev]->dxbar[0];
	    dz = lev[ilev]->dxbar[1];
	    r = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	    z = cptr->xc[1];
	    vr = cptr->qa[iqu0+1]/dens;
	    vz = cptr->qa[iqu0+2]/dens;
	    v = sqrt(vr*vr + vz*vz);
	  }
	  else if (xyz){
	    dx = lev[ilev]->dxbar[0];
	    dy = lev[ilev]->dxbar[1];
	    dz = lev[ilev]->dxbar[2];
	    x = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	    y = cptr->xc[1];
	    z = cptr->xc[2];
	    r = sqrt(y*y + z*z);
	    vx = cptr->qa[iqu0+1]/dens;
	    vy = cptr->qa[iqu0+2]/dens;
	    vz = cptr->qa[iqu0+3]/dens;
	    v  = sqrt(vx*vx + vy*vy + vz*vz);
	    vr = sqrt(vy*vy + vz*vz);
	    vz = vx;
	    z = x;
	  }
	  else {
	    cout << "usrsum: Not coded for other geometries\n";
	    exit(EXIT_FAILURE);
	  }
          ke = 0.5*dens*v*v;
          ie = pre/(g-1.0);

          // Calculate shell mass etc...
          shell_mass += col2*mass;
          shell_mtm  += col2*mass*vx;
          if (dens > 5.0*ramb){  // calculate mass and mtm of the dense part of the shell
            shell_mass_hi += col2*mass;
            shell_mtm_hi  += col2*mass*vx;
          }

          // Calculate various fluxes through the hole in the shell
          if ((cptr->xc[0] > -lev[ilev]->dxbar[0]) && (cptr->xc[0] < 0.0)){ // select a 1-cell wide strip on the finest grid level
            if (nholes == 1){
              if (((rzp) && (r < holedr)) || ((xyz) && (mabs(cptr->xc[1]) < holedr))){ // in original hole
                double vp = cptr->qa[iqu0+1]/dens;
                double da = 0.0;
                if (xyz){
                  if (nd == 2)      da = dy;    // size of hole (cm)
                  else if (nd == 3) da = dy*dz; // size of hole (cm^2)
                }
                else if (rzp){
                  float r1 = r - 0.5*(lev[ilev]->dxbar[0]);
                  float r2 = r + 0.5*(lev[ilev]->dxbar[0]);
                  da = gkpi*(r2*r2 - r1*r1);      // size of hole (cm^2)
                }
                mflux   += da*cptr->qa[iqu0+1];
                mtmflux += da*cptr->qa[iqu0+1]*vp;
                eflux   += da*(ke + ie)*vp;
              }
            }
            else{              
              for (int n = 0; n < nholes; n++){ // Loop over each hole
                if (nd == 2){
                  float lft_edge = hole_ypos[n] - hole_radius[n];
                  float rgh_edge = hole_ypos[n] + hole_radius[n];
                  if ((lft_edge < cptr->xc[1])&&(rgh_edge > cptr->xc[1])){// in hole
                    double vp = cptr->qa[iqu0+1]/dens;
                    mflux   += (lev[ilev]->dxbar[1])*cptr->qa[iqu0+1];
                    mtmflux += (lev[ilev]->dxbar[1])*cptr->qa[iqu0+1]*vp;
                    eflux   += (lev[ilev]->dxbar[1])*(ke + ie)*vp;
                  }
                }
                else if (nd == 3){
                  float dy = cptr->xc[1] - hole_ypos[n];
                  float dz = cptr->xc[2] - hole_zpos[n];
                  float dr = sqrt(dy*dy + dz*dz);
                  if (dr < hole_radius[n]){// in hole
                    double vp = cptr->qa[iqu0+1]/dens;
                    mflux   += (lev[ilev]->dxbar[1])*cptr->qa[iqu0+1];
                    mtmflux += (lev[ilev]->dxbar[1])*cptr->qa[iqu0+1]*vp;
                    eflux   += (lev[ilev]->dxbar[1])*(ke + ie)*vp;
                  }
                }
              } // loop over holes
            } // nholes?
          } // 1-cell wide strip
        }  // child cell?
	cptr = cptr->nxt;
      }
    }

    //compute global quantities
    if (limpe > 1){               //limpe = 1 if single processor
      mflux   = global_sum(mflux);
      mtmflux = global_sum(mtmflux);
      eflux   = global_sum(eflux);
      shell_mass    = global_sum(shell_mass);
      shell_mass_hi = global_sum(shell_mass_hi);
      shell_mtm     = global_sum(shell_mtm);
      shell_mtm_hi  = global_sum(shell_mtm_hi);
    }

    //normalize data
    if ((limpe == 1)||(mype == 0)){// ie if single processor or master
      if (xyz && (nd == 2)) volnorm = pow(rscale,2); // cm^2 
      else                  volnorm = pow(rscale,3); // cm^3 
      //volnorm = pow(rscale,3); // gives 2D xyz units of: mass in Msol/pc
      massnorm   = volnorm*rhoscale;
      energynorm = volnorm*pscale;
      double mtmnorm    = massnorm*(rscale/tscale);
      double massfluxnorm = massnorm/tscale;
      double mtmfluxnorm  = mtmnorm/tscale;
      double efluxnorm  = energynorm/tscale;
      //append info to a file (read and write access)
      shell_mass    *= massnorm/msol;        //units Msol (3D xyz or 2D zrp) or Msol/cm (2D xyz)
      shell_mass_hi *= massnorm/msol;        //units Msol (3D xyz or 2D zrp) or Msol/cm (2D xyz)
      shell_mtm     *= mtmnorm/(msol*1.0e5); //units Msol km s^-1 (3D xyz or 2D zrp) or (Msol km/s)/cm (2D xyz)
      shell_mtm_hi  *= mtmnorm/(msol*1.0e5); //units Msol km s^-1 (3D xyz or 2D zrp) or (Msol km/s)/cm (2D xyz)
      //vol  = ambvol *volnorm/pow(parsec,3); //units pc^-3 
      //ke   = ambke  *energynorm/1.0e30; //units 1e30 ergs
      //inte = ambie  *energynorm/1.0e30; //units 1e30 ergs
      mflux   *= massfluxnorm/(msol/yr);    // units Msol/yr (3D xyz or 2D zrp) or (Msol/yr)/cm (2D xyz)
      mtmflux *= mtmfluxnorm/(msol*1.0e5);  // units Msol km s^-2 (3D xyz or 2D zrp) or (Msol km s^-2)/cm (2D xyz)
      eflux   *= efluxnorm;                 // units erg/s (3D xyz or 2D zrp) or (erg/s)/cm (2D xyz)
    
      // Put statistics for 2D xyz simulations into nicer units 
      if (xyz && (nd == 2)){
        shell_mass    *= parsec; // units Msol/pc
        shell_mass_hi *= parsec; // units Msol/pc
        shell_mtm     *= parsec; // units (Msol km/s)/pc
        shell_mtm_hi  *= parsec; // units (Msol km/s)/pc
        mflux *= parsec;   // units (Msol/yr)/pc
        mtmflux *= parsec; // units (Msol km s^-2)/pc
        eflux *= parsec;   // units (erg/s)/pc
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"shellstats.dat");
      fptr = fopen(statsfile,"a+");
      if (fptr!=NULL){
	//fputs ("fopen example",pFile); 
	fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		t,t*tscale/yr,shell_mass,shell_mass_hi,shell_mtm,shell_mtm_hi);
	fclose (fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"fluxstats.dat");
      fptr = fopen(statsfile,"a+");
      if (fptr!=NULL){
	//fputs ("fopen example",pFile); 
	fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e\n",
		t,t*tscale/yr,mflux,mtmflux,eflux);
	fclose (fptr);
      }
    } // single or master  
  } // SHELL_BLOWOUT case  
  else if (strcmp(cas, "WIND") == 0){
  //Zero wind totals
    for (n = 1; n<= nstars; n++){
      windmass[n] = 0.0;
      windke  [n] = 0.0;
      windie  [n] = 0.0;
      windvol [n] = 0.0;
    }
    ambcol = 0.0;
    ambmass = 0.0;
    ambke   = 0.0;
    ambie   = 0.0;
    ambvol  = 0.0;

    //Calculate the global statistics   
    //cout << "JMP here\n";
    //cout << "nlevs = " << nlevs << ", plmax = " << plmax <<"\n";
    for (ilev = 0; ilev <= nlevs-1; ilev++){ //sum over ALL grids
      //for (ilev = plmin; ilev <= plmax; ilev++){ //sum over plot grids
      cptr = lev[ilev]->fcell;
      while (cptr != NULL){
	if (cptr->child[0] == NULL){ //if finest cell then ...
	  cvol = cptr->vol;
	  for (i = 0; i < nq; i++)
	    tot[i] += cvol*cptr->qa[i];
	  dens = cptr->qa[iqd];
	  col = cptr->qa[iqal0+1]/dens;
	  mass = dens*cvol;
	  if (rzp){
	    dr = lev[ilev]->dxbar[0];
	    dz = lev[ilev]->dxbar[1];
	    r = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	    z = cptr->xc[1];
	    vr = cptr->qa[iqu0+1]/dens;
	    vz = cptr->qa[iqu0+2]/dens;
	    v = sqrt(vr*vr + vz*vz);
	  }
	  else if (xyz){
	    dx = lev[ilev]->dxbar[0];
	    dy = lev[ilev]->dxbar[1];
	    dz = lev[ilev]->dxbar[2];
	    x = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	    y = cptr->xc[1];
	    z = cptr->xc[2];
	    r = sqrt(y*y + z*z);
	    vx = cptr->qa[iqu0+1]/dens;
	    vy = cptr->qa[iqu0+2]/dens;
	    vz = cptr->qa[iqu0+3]/dens;
	    v  = sqrt(vx*vx + vy*vy + vz*vz);
	    vr = sqrt(vy*vy + vz*vz);
	    vz = vx;
	    z = x;
	  }
	  else {
	    cout << "usrsum: Not coded for other geometries\n";
	    exit(EXIT_FAILURE);
	  }
	  if (strcmp(cas, "WIND") == 0){
	    totcol = 0.0;
	    for (n=1; n<= nstars; n++){
	      col = cptr->qa[iqal0+n]/dens; //fraction of cell material 
	                                    //belonging to wind n
	      windmass[n] += col*mass;
	      ke = 0.5*mass*v*v;
	      windke[n] += col*ke;
	      inte = (cptr->qa[iqe] - ke/cvol)*cvol;
	      //pre = (g-1.0)*inte;
	      //Next 6 lines are not necesary if usrfxq has done its job
	      //if (inten){
	      //  if (inte < 0.01*ke/cvol){
	      //    inte = cptr->qa[iqie]; //correct internal energy
	      //    pre = (g-1.0)*cptr->qa[iqie]; //pre from internal energy
	      //  }
	      //}
	      windie[n]  += col*inte;
	      windvol[n] += col*cvol;
	      totcol += col;
	    }
	    ambcol = 1.0-totcol;
	    ambmass += ambcol*mass;
	    ambke   += ambcol*ke;
	    ambie   += ambcol*inte;
	    ambvol  += ambcol*cvol;
	  }
	}
	cptr = cptr->nxt;
      }
    }

  //compute global quantities
    if (limpe > 1){               //limpe = 1 if single processor
      for (n=1; n<= nstars; n++){    
	windmass[n] = global_sum(windmass[n]);
	windke  [n] = global_sum(windke  [n]);
	windie  [n] = global_sum(windie  [n]);
	windvol [n] = global_sum(windvol [n]);
      }
      ambmass = global_sum(ambmass);
      ambvol  = global_sum(ambvol);
      ambke   = global_sum(ambke);
      ambie   = global_sum(ambie);
    }

    //normalize data
    if ((limpe == 1)||(mype == 0)){// ie if single processor or master
      volnorm    = pow(rscale,3);
      massnorm   = volnorm*rhoscale;
      energynorm = volnorm*pscale;
      //append info to a file (read and write access)
      //open 1 file for every wind
      for (n=1; n<= nstars; n++){    
	mass = windmass[n]*massnorm/msol;     //units Msol
	vol  = windvol [n]*volnorm/pow(parsec,3); //units pc^-3 
	ke   = windke  [n]*energynorm/1.0e30; //units 1e30 ergs
	inte = windie  [n]*energynorm/1.0e30; //units 1e30 ergs
	tote = ke + inte;                     //units 1e30 ergs
	strcpy(statsfile,prefx);
	sprintf(cnum, "%d",n);
	strcat(statsfile,cnum);
	strcat(statsfile,"stats.dat");
	fptr = fopen(statsfile,"a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t*tscale/yr,mass,vol,ke,inte,tote);
	  fclose (fptr);
	}
      }//loop over winds
      //Now append ambient data
      mass = ambmass*massnorm/msol;     //units Msol
      vol  = ambvol *volnorm/pow(parsec,3); //units pc^-3 
      ke   = ambke  *energynorm/1.0e30; //units 1e30 ergs
      inte = ambie  *energynorm/1.0e30; //units 1e30 ergs
      tote = ke + inte;                 //units 1e30 ergs
      strcpy(statsfile,prefx);
      strcat(statsfile,"ambstats.dat");
      fptr = fopen(statsfile,"a+");
      if (fptr!=NULL){
	//fputs ("fopen example",pFile); 
	fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		t,t*tscale/yr,mass,vol,ke,inte,tote);
	fclose (fptr);
      }
    }//single or master
  }// WIND case
  
  //SINGLECLOUD and MANYCLOUD cases
  if ((strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "MANYCLOUD") == 0)){
    //    if (mype == 0) cerr << "SC or MC\n";
    //zero global quantities and initial summation indices
    for (i = 0; i < nmax; i++) {
      massb[i] = 0.0;
      rhob[i] = 0.0;
      for (n = 0; n <= nal; n++) {
	massb_single[n][i] = 0.0;
	rhob_single[n][i] = 0.0;
      }
    }
    
    rms = 0.0;
    zms = 0.0;
    zm  = 0.0;
    vrms= 0.0;
    vzms= 0.0;
    vzm = 0.0;
    volc= 0.0;
    circ= 0.0;
  
    rms_2 = 0.0;
    zms_2 = 0.0;
    zm_2  = 0.0;
    vrms_2= 0.0;
    vzms_2= 0.0;
    vzm_2 = 0.0;
    volc_2= 0.0;
    
    rms_3 = 0.0;
    zms_3 = 0.0;
    zm_3  = 0.0;
    vrms_3= 0.0;
    vzms_3= 0.0;
    vzm_3 = 0.0;
    volc_3= 0.0;
    
    mc  = 0.0;
    mz  = 0.0;
    mz2 = 0.0;
    mr2 = 0.0;
    mvr2= 0.0;
    mvz = 0.0;
    mvz2= 0.0;
    vc  = 0.0;
    vc2 = 0.0;
    rhomax = 0.0;
    
    mc_2  = 0.0;
    mz_2  = 0.0;
    mz2_2 = 0.0;
    mr2_2 = 0.0;
    mvr2_2= 0.0;
    mvz_2 = 0.0;
    mvz2_2= 0.0;
    vc_2  = 0.0;
    vc2_2 = 0.0;
  
    mc_3  = 0.0;
    mz_3  = 0.0;
    mz2_3 = 0.0;
    mr2_3 = 0.0;
    mvr2_3= 0.0;
    mvz_3 = 0.0;
    mvz2_3= 0.0;
    vc_3  = 0.0;
    vc2_3 = 0.0;
    
    kie = 0.0;
    the = 0.0;
    tue = 0.0;

    tkie = 0.0;
    tthe = 0.0;
    ttue = 0.0;
    
    if (nal > 1){//many clouds
      for (n=0; n < mcmax; n++){
	mc_cloud[n] = 0.0;
	mx_cloud[n] = 0.0;
	my_cloud[n] = 0.0;
	mz_cloud[n] = 0.0;
	mx2_cloud[n] = 0.0;
	my2_cloud[n] = 0.0;
	mz2_cloud[n] = 0.0;
	mvx_cloud[n] = 0.0;
	mvy_cloud[n] = 0.0;
	mvz_cloud[n] = 0.0;
	mvx2_cloud[n] = 0.0;
	mvy2_cloud[n] = 0.0;
	mvz2_cloud[n] = 0.0;
	vc_cloud[n] = 0.0;
	mc_core[n] = 0.0;
	mx_core[n] = 0.0;
	my_core[n] = 0.0;
	mz_core[n] = 0.0;
	mx2_core[n] = 0.0;
	my2_core[n] = 0.0;
	mz2_core[n] = 0.0;
	mvx_core[n] = 0.0;
	mvy_core[n] = 0.0;
	mvz_core[n] = 0.0;
	mvx2_core[n] = 0.0;
	mvy2_core[n] = 0.0;
	mvz2_core[n] = 0.0;
	vc_core[n] = 0.0;
	kiemc[n] = 0.0;
	themc[n] = 0.0;
	tuemc[n] = 0.0;
	zm_cld[n] = 0.0;
	zm_cor[n] = 0.0;
	mz_cld[n] = 0.0;
	mz2_cld[n] = 0.0;
	mvz2_cld[n] = 0.0;
	mz_cor[n] = 0.0;
	mz2_cor[n] = 0.0;
	mvz2_cor[n] = 0.0;
	dz_cld[n] = 0.0;
	dz_cor[n] = 0.0;
	dvx_cld[n] = 0.0;
	dvx_cor[n] = 0.0;
	dvy_cld[n] = 0.0;
	dvy_cor[n] = 0.0;
	dvz_cld[n] = 0.0;
	dvz_cor[n] = 0.0;
      }
    }
    
    //cout << rhomax << " " << rhomin << " " << dlogrho << "\n";
    //cout << plmin << " " << plmax << "\n";
    //exit(EXIT_FAILURE);

    //define normalization
    cs = sqrt(g*kboltz*tamb/avgmass);  //sound speed in initial intercloud medium
    vb = mach*cs/uscale;               //speed of shock in IIM (mg_gt units)
    //cout << "vb = " << vb <<"\n";
    //exit(EXIT_FAILURE);
    if (nal == 1){
      rhom0 = ramb*chi/rhoscale;         //initial max density of cloud
      tcc = sqrt(chi)/vb;                //cloud crushing timescale (radius=1)
    }
    else{
      if ((chimin == chimax)&&(rcmin == rcmax)){ //uniform clouds
	rhom0 = ramb*chimin/rhoscale;      //initial max density of cloud
      tcc = sqrt(chimin)/vb;             //cloud crushing timescale (radius=1)
      }
      else {          //chi is not uniform
	rhom0 = 1.0;  //so set these parameters to 1.0
	tcc = 1.0;
      }
    }

    
    //Calculate the other global statistics 
    /* Sum over grids from plmin to plmax */ 
    if (nal == 1){ //individual cloud
    //cout << "JMP here\n";
    //cout << "nlevs = " << nlevs << ", plmin = " << plmin << ", plmax = " << plmax <<"\n";
      for (ilev = 0; ilev <= nlevs-1; ilev++){ //sum over ALL grids
	//for (ilev = plmin; ilev <= plmax; ilev++){ //sum over plot grids
	cptr = lev[ilev]->fcell;
	while (cptr != NULL){
	  if (!cptr->refined){    //if finest cell then...
	    //	    if (mype == 0) cout << "cell # = " << cptr->icell << "\n";
	    //	    if (mype == 10) cerr << "cell # = " << cptr->icell << " mype= " << mype << endl;
	    cvol = cptr->vol;
	    for (i = 0; i < nq; i++)
	      tot[i] += cvol*cptr->qa[i];
	    dens = cptr->qa[iqd];
	    col = cptr->qa[iqal0+1]/dens;
	    colamb = 1.0/chi;
	    if (dens > rhomax) rhomax = dens;
	    mass = dens*cvol;
	    
	    if (rzp){
	      dr = lev[ilev]->dxbar[0];
	      dz = lev[ilev]->dxbar[1];
	      // if (ilev == 2){
          //  cout << "ilev = " << ilev << " dxbar = " << lev[ilev]->dxbar[0];
          //  exit(EXIT_FAILURE);
          //}
	      r = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	      z = cptr->xc[1];
	      vr = cptr->qa[iqu0+1]/dens;
	      vz = cptr->qa[iqu0+2]/dens;
	    }
	    else if (xyz){
	      dx = lev[ilev]->dxbar[0];
	      dy = lev[ilev]->dxbar[1];
	      //	      dz = lev[ilev]->dxbar[2];
	      x = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	      y = cptr->xc[1];
	      if (nd ==3) {
		z = cptr->xc[2];
		dz = lev[ilev]->dxbar[2];
		vz = cptr->qa[iqu0+3]/dens;
	      }
	      else {
		z = 0.0;
		dz = 0.0;
		vz = 0.0;
	      }
	      //	      z = cptr->xc[2];
	      //z sometimes gains large value, e+25 or so in 2D sims.
	      // cerr << "z: " << z << endl;
	      r = sqrt(y*y + z*z);
	      vx = cptr->qa[iqu0+1]/dens;
	      vy = cptr->qa[iqu0+2]/dens;
	      //	      vz = cptr->qa[iqu0+3]/dens;
	      //	  cerr << "vr: sqrt( " << vy << "^2 * " << vz << "^2\n";
	      vr = sqrt(vy*vy + vz*vz);
	      vz = vx;
	      z = x;
              // Comment out the following if we wish to treat a single cloud 3D sim like a 2D RZP sim
	      //if (nd ==3) {
	      //  z = cptr->xc[2];
	      //  dz = lev[ilev]->dxbar[2];
	      //  vz = cptr->qa[iqu0+3]/dens;
	      //}
	      //else {
	      //  z = 0.0;
	      //  dz = 0.0;
	      //  vz = 0.0;
	      //}

	    }
	    else{
	      cout << "usrsum: Not coded for other geometries\n";
	      exit(EXIT_FAILURE);
	    }
	    
	    //Construct moments using the colour variable. Define
	    //cloud material as those with col > 2.0 times the ambient colour
	    //cout << chi << " " << colamb << "\n";
	    //exit(EXIT_FAILURE);
	    //if (mype == 0){
          //  cout << "r = " << r << ", z = " << z << ", dens = " << dens << ", vol = " << cvol << ", mass = " << mass << "\n";
	    //}
	    if (col >= 2.0*colamb){ // cell material is mostly cloud
	      mcol = mass;
	      mc   += mcol;
	      mz   += mcol*z;
	      mr   += mcol*r;
	      mz2  += mcol*z*z;
	      mr2  += mcol*r*r;
	      mvr2 += mcol*vr*vr;
	      mvz  += mcol*vz;
	      mvz2 += mcol*vz*vz;
	      vc   += cvol;
	    }
	    
	    //only count cloud material (col=0, outside r=1.5)
	    if (col >= 2.0*colamb){
	      mcol_2 = mass*col;
	      mc_2   += mcol_2;
	      mz_2   += mcol_2*z;
	      mr_2   += mcol_2*r;
	      mz2_2  += mcol_2*z*z;
	      mr2_2  += mcol_2*r*r;
	      mvr2_2 += mcol_2*vr*vr;
	      mvz_2  += mcol_2*vz;
	      mvz2_2 += mcol_2*vz*vz;
	      vc_2   += cvol*col;
	      i = int(100.0*mabs(vz)/(0.75*vb));
	      massb[i] += mcol_2;
	      i = int(50.0*dens/rhom0);
	      rhob[i] += mcol_2;
	      //cout << i << "," << mcol_2 << "," << mc_2 << "," << massb[i] << "\n";          
	    }
	    
	    //only count cells with mostly cloud material (col=0, outside r=1.5)
	    if (col >= 0.5){ // cell material is mostly cloud
	      mcol_3 = mass*col;
	      mc_3   += mcol_3;
	      mz_3   += mcol_3*z;
	      mr_3   += mcol_3*r;
	      mz2_3  += mcol_3*z*z;
	      mr2_3  += mcol_3*r*r;
	      mvr2_3 += mcol_3*vr*vr;
	      mvz_3  += mcol_3*vz;
	      mvz2_3 += mcol_3*vz*vz;
	      vc_3   += cvol*col;
	    }
	    
	    //integrate the cloud kinetic, thermal, and turbulent energy
	    //(k is the turbulent energy per unit mass, qa[iqk] is rhok,
	    //which is the turbulent energy density with dimension energy/volume)
	    ke = 0.5*dens*(vr*vr + vz*vz); //kinetic energy density
	    kie += ke*col*cvol;
	    ie = cptr->qa[iqe] - ke;
	    the += ie*col*cvol;//thermal energy = internal energy density * volume
	    tue += cptr->qa[iqk]*col*cvol;
	    
	    //integrate the total kinetic, thermal, and turbulent energy
	    tkie += ke*cvol;
	    tthe += ie*cvol;
	    ttue += cptr->qa[iqk]*cvol;
	    
	    //construct the vorticity and the integerated vorticity (circulation)
	    //NOTE:The calculation for omega is different to the runs in paper I 
	    // and II becuase of a change in the definition of dudx
	    if (rzp){
	      omega = -(cptr->dudx[0][1] - cptr->dudx[1][0]); // thetahat (dv_r/dz - d
	      //For models without keps dudx is not calculated. Therefore I'll need
	      //to calculate the gradient myself.
	      circ += omega*dr*dz; // omega . dA
	    }
	    else if (xyz){
	      //construct the vorticity and the integerated vorticity (circulation)
	      omegax = cptr->dudx[2][1] - cptr->dudx[1][2]; // (dv_z/dy - dv_y/dz)
	      omegay = cptr->dudx[0][2] - cptr->dudx[2][0]; // (dv_x/dz - dv_z/dx)
	      omegaz = cptr->dudx[1][0] - cptr->dudx[0][1]; // (dv_y/dx - dv_x/dy)
	      //For models without keps dudx is not calculated. Therefore I'll need
	      //to calculate the gradient myself.
	      circ += omegax*dy*dz + omegay*dx*dz + omegaz*dx*dy; // omega . dA
	    }
	    //fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
	    //	r,z,divdr,divdz,cptr->dudx[0][1],cptr->dudx[1][0],omega,circ);
	  } //finest cell
	  cptr = cptr->nxt;
	} //next cell
      } //cycle over grid levels
      
      //fclose (fptr);
      
      //normalize mass in (velocity and density) bins
      /*      if (mc_2 != 0.0){ // note that with multi processors mc_2 could be 0.0
	for (i = 0; i < nmax; i++)
	  massb[i] /= mc_2;
    for (i = 0; i < nmax; i++)
      rhob[i] /= mc_2;
      }*/
      
      //cout << " proc # = " << mype << "; mc = " << mc << "\n";
      //      cout << "returning from usrsum, proc # = " << mype << "\n";
      //return;
      
      //compute global quantities
      if (limpe > 1){               //limpe = 1 if single processor
	mc    = global_sum(mc );
	vc    = global_sum(vc );
	mr2   = global_sum(mr2);
	mz2   = global_sum(mz2);
	mz    = global_sum(mz );
	mr    = global_sum(mr );
	mvr2  = global_sum(mvr2);
	mvz2  = global_sum(mvz2);
	mvz   = global_sum(mvz);
	mc_2  = global_sum(mc_2 );
	vc_2  = global_sum(vc_2 );
	mr2_2 = global_sum(mr2_2);
	mz2_2 = global_sum(mz2_2);
	mz_2  = global_sum(mz_2 );
	mr_2  = global_sum(mr_2 );
	mvr2_2= global_sum(mvr2_2);
	mvz2_2= global_sum(mvz2_2);
	mvz_2 = global_sum(mvz_2);
	mc_3  = global_sum(mc_3 );
	vc_3  = global_sum(vc_3 );
	mr2_3 = global_sum(mr2_3);
	mz2_3 = global_sum(mz2_3);
	mz_3  = global_sum(mz_3 );
	mr_3  = global_sum(mr_3 );
	mvr2_3= global_sum(mvr2_3);
	mvz2_3= global_sum(mvz2_3);
	mvz_3 = global_sum(mvz_3);
	kie   = global_sum(kie  );
	the   = global_sum(the  );
	tue   = global_sum(tue  );
	tkie  = global_sum(tkie );
	tthe  = global_sum(tthe );
	ttue  = global_sum(ttue );
	circ  = global_sum(circ );
	rhomax= global_max(rhomax);
      }
      //cout << " 2: proc # = " << mype << "; mc = " << mc << "\n";
      //return;
      
      //compute global quantities using the colour approach
      if ((limpe == 1)||(mype == 0)){// ie if single processor or master
	rms = mr2/mc;
	zms = mz2/mc;
	zm  = mz/mc;
	rm  = mr/mc;
	vrms= mvr2/mc;
	vzms= mvz2/mc;
	vzm = mvz/mc;
	
	rms_2 = mr2_2/mc_2;
	zms_2 = mz2_2/mc_2;
	zm_2  = mz_2/mc_2;
	rm_2  = mr_2/mc_2;
	vrms_2= mvr2_2/mc_2;
	vzms_2= mvz2_2/mc_2;
	vzm_2 = mvz_2/mc_2;
	
	rms_3 = mr2_3/mc_3;
	zms_3 = mz2_3/mc_3;
	zm_3  = mz_3/mc_3;
	rm_3  = mr_3/mc_3;
	vrms_3= mvr2_3/mc_3;
	vzms_3= mvz2_3/mc_3;
	vzm_3 = mvz_3/mc_3;
	
	a = sqrt(2.5*rms);           // effective radius normal to axis
	c = sqrt(5.0*(zms - zm*zm)); //                 along axis
	dvz = sqrt(vzms - vzm*vzm);  // axial velocity dispersion
	dvr = sqrt(vrms);            // radial velocity dispersion
	rhom = mc/vc;                // mean density of cloud
	
	a_2 = sqrt(2.5*rms_2);           
	c_2 = sqrt(5.0*(zms_2 - zm_2*zm_2)); 
	dvz_2 = sqrt(vzms_2 - vzm_2*vzm_2); 
	dvr_2 = sqrt(vrms_2);           
	rhom_2 = mc_2/vc_2;               
	
	a_3 = sqrt(2.5*rms_3);           
	c_3 = sqrt(5.0*(zms_3 - zm_3*zm_3)); 
	dvz_3 = sqrt(vzms_3 - vzm_3*vzm_3); 
	dvr_3 = sqrt(vrms_3);           
	rhom_3 = mc_3/vc_3;               

	//normalize global quantities and output to a file
	rhomax /= rhom0;
	rhom /= rhom0;
	mc /= mcloud;
	vc /= vcloud;
	vzm /= vb; 
	dvz /= vb;
	dvr /= vb;
	
	rhom_2 /= rhom0;
	mc_2 /= mcloud;
	vc_2 /= vcloud;
	vzm_2 /= vb; 
	dvz_2 /= vb;
	dvr_2 /= vb;
	
	rhom_3 /= rhom0;
	mc_3 /= mcloud;
	vc_3 /= vcloud;
	vzm_3 /= vb; 
	dvz_3 /= vb;
	dvr_3 /= vb;

	etot = kie + the + tue; //total energy of cloud material
	kie /= etot;
	the /= etot;
	tue /= etot;
	
	//append info to a file (read and write access)
	//if ((limpe == 1)||(mype == 0)){//write if single processor or master
	//FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	//strcpy(prefx,"m10.0chi1e3adialokepsres128circ"); //prefix for movie files
	//strcpy(prefx,"m10.0chi1e3adialokepsres64p21"); //prefix for movie files
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,a,c,c/a,mc,vc,rhomax,rhom,vzm,dvz,dvr,circ/-vb,zm,rm);
	  fclose (fptr);
	}
	
	//append info to a file (read and write access)
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats2.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,a_2,c_2,c_2/a_2,mc_2,vc_2,rhomax,rhom_2,vzm_2,dvz_2,dvr_2,zm_2,rm_2);
	  fclose (fptr);
	}
	
	//append info to a file (read and write access)
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats3.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,a_3,c_3,c_3/a_3,mc_3,vc_3,rhomax,rhom_3,vzm_3,dvz_3,dvr_3,zm_3,rm_3);
	  fclose (fptr);
	}
	
	//append info to a file (read and write access)
	escale = (rhoscale*uscale*uscale)*(rscale/1.0e15)*(rscale/1.0e15)*rscale;
	etot *= escale;
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats4.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,kie,the,tue,etot,tue*etot);
	  fclose (fptr);
	}
	
  //append info to a file (read and write access)
	etot = tkie + tthe + ttue; //total energy of grid material
	tkie /= etot;
	tthe /= etot;
	ttue /= etot;
	etot *= escale;
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats5.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,tkie,tthe,ttue,etot,tkie*etot,tthe*etot,ttue*etot);
	  fclose (fptr);
	}
 //write out mass in (velocity) bins
  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	/*	strcpy(statsfile, prefx);	
	strcat(statsfile,"mv.dat");
	fptr = fopen(statsfile,"w");
	if (fptr!=NULL){
	  mass = 0.0;
	  for (i = 0; i < 200; i++){
	    mass += massb[i];
	    fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/100.0,massb[i],mass);	     
	  }
	  fclose (fptr);
	}
  

  
	//write out mass in (density) bins
	//FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	strcpy(statsfile, prefx);
	strcat(statsfile,"mrho.dat");
	fptr = fopen(statsfile,"w");
	if (fptr!=NULL){
	  mass = 0.0;
	  for (i = 0; i < 2000; i++){
	    mass += rhob[i];
	    fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/50.0,rhob[i],mass);	     
	  }
	  fclose (fptr);
	}
	*/
      } //single processor or master  
    }//single cloud
    
    
    
    else{ //many clouds
      //      cerr << mype << " in manycloud\n";
      /*RA: DEBUG
      //create file for debug output
      fptr = fopen("temp2.dat", "w");
      if (fptr != NULL) {
      fprintf(fptr, "Entering manyclouds. Time now: %.4e\n", t);
      fclose (fptr);
      }
      RA: end DEBUG*/
      //    cerr << "JMP here\n";
      //cerr << "nlevs = " << nlevs << ", plmin = " << plmin << ", plmax = " << plmax <<"\n";
      ilev = 0;
      //cout << mype <<" processor entering " << ilev << endl;
      while (ilev <= nlevs-1) {
      //  for ( ilev = 0; ilev <= nlevs-1; ilev++){ //sum over ALL grids
	if (mype == 0)	cerr << "ilev " << ilev << "  " << mype << "  mype\n";
	//for (ilev = plmin; ilev <= plmax; ilev++){ //sum over plot grids
	cptr = lev[ilev]->fcell;
	while (cptr != NULL){
	  if (!cptr->refined){    //if finest cell then...
	    //	    if (mype == 0) cerr << "cell # = " << cptr->icell << "\n";
	    cvol = cptr->vol;
	    for (i = 0; i < nq; i++)
	      tot[i] += cvol*cptr->qa[i];
	    dens = cptr->qa[iqd];
	    mass = dens*cvol;
	    dx = lev[ilev]->dxbar[0];
	    dy = lev[ilev]->dxbar[1];
	    x = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	    y = cptr->xc[1];
	    /*	if (ilev == nlevs-1) {
	      fptr = fopen("temp2.dat", "a+");
	      if (fptr != NULL) {
	      fprintf(fptr, "x: %.4e y: %.4e\n", x, y);
	      fclose (fptr);
	      }
	      }*/
	    vx = cptr->qa[iqu0+1]/dens;
	    vy = cptr->qa[iqu0+2]/dens;

	    if (nd ==3) {
	      z = cptr->xc[2];
	      dz = lev[ilev]->dxbar[2];
	      vz = cptr->qa[iqu0+3]/dens;
	    }
	    else {
	      z = 0.0;
	      dz = 0.0;
	      vz = 0.0;
	    }
	    //Construct moments using the advected scalars to distinguish
	    //the desired material. Define
	    //cloud material as those with col > 0.01, and core material
	    //as that with col > 0.5
	    
	    //First search over blocks
	    for (n=2; n <= 5; n++){
	      col = cptr->qa[iqal0+n]/dens;
	      //    cout << "dens = " << dens << ", col = " << col << "\n";
	      if (col >= 0.01){//integrate for the cloud (not core) material
		mcol_cloud[n] = mass*col;
		mc_cloud[n]   += mcol_cloud[n];
		mx_cloud[n]   += mcol_cloud[n]*x;
		mx2_cloud[n]  += mcol_cloud[n]*x*x;
		mvx_cloud[n]  += mcol_cloud[n]*vx;
		mvx2_cloud[n] += mcol_cloud[n]*vx*vx;
		mvy2_cloud[n] += mcol_cloud[n]*vy*vy;
		if (nd == 3) mvz2_cloud[n] += mcol_cloud[n]*vz*vz;
		vc_cloud[n]   += cvol*col;
		i = int(100.0*mabs(vx)/(0.75*vb));
		massb[i] += mcol_cloud[n];
		//	cout <<"i: " << i << " massb[i]: " << massb[i] << " x " << x << " y " << y << endl;
		i = int(50.0*dens/rhom0);
		rhob[i] += mcol_cloud[n];
		//	cout << "mype: " << mype << "i: " << i << " rhob[i]: " << rhob[i] << " x " << x << " y " << y << endl;
		i = int(100.0*mabs(vx)/(0.75*vb));
		massb_single[n][i] += mcol_cloud[n];
		i = int(50.0*dens/rhom0);
		rhob_single[n][i] += mcol_cloud[n];
	      }
	      if (col >= 0.5){//integrate for core material
		//cout << "cloud core, n = " << n << "\n";
		//exit(EXIT_FAILURE);
		mcol_core[n] = mass*col;
		mc_core[n]   += mcol_core[n];
		mx_core[n]   += mcol_core[n]*x;
		mx2_core[n]  += mcol_core[n]*x*x;
		mvx_core[n]  += mcol_core[n]*vx;
		mvx2_core[n] += mcol_core[n]*vx*vx;
		mvy2_core[n] += mcol_core[n]*vy*vy;
		if (nd == 3) mvz2_core[n] += mcol_core[n]*vz*vz;
		vc_core[n]   += cvol*col;
	      }
	      //integrate the kinetic, thermal, and turbulent energy of all of the 
	      //cloud material within a given block (ie over many clouds)
	      //(k is the turbulent energy per unit mass, qa[iqk] is rhok,
	      //which is the turbulent energy density with dimension energy/volume)
	      ke = 0.5*dens*(vx*vx + vy*vy + vz*vz); //kinetic energy density of cell
	      kiemc[n] += ke*col*cvol;
	      ie = cptr->qa[iqe] - ke; //thermal energy of cell
	      themc[n] += ie*col*cvol;//therm. energy = int. energy density*volume
	      tuemc[n] += cptr->qa[iqk]*col*cvol;
	    } //end search over blocks
	    
	    for (n=6; n <= nal; n++){ // search over individual clouds
	      //     if (mype == 0) cerr << " Individual Clouds\n";
	      col = cptr->qa[iqal0+n]/dens;
	      if (col >= 0.01){// count cloud material within cell
		mcol_cloud[n] = mass*col;
		mc_cloud[n]   += mcol_cloud[n];
		mx_cloud[n]   += mcol_cloud[n]*x;
		my_cloud[n]   += mcol_cloud[n]*y;
		mx2_cloud[n]  += mcol_cloud[n]*x*x;
		my2_cloud[n]   += mcol_cloud[n]*y*y;
		mvx_cloud[n]  += mcol_cloud[n]*vx;
		mvy_cloud[n] += mcol_cloud[n]*vy;
		mvx2_cloud[n] += mcol_cloud[n]*vx*vx;
		mvy2_cloud[n] += mcol_cloud[n]*vy*vy;
		vc_cloud[n]   += cvol*col;
		i = int(100.0*mabs(vx)/(0.75*vb));
		massb_single[n][i] += mcol_cloud[n];
		i = int(50.0*dens/rhom0);
		rhob_single[n][i] += mcol_cloud[n];
		if (nd == 3){
		  mz_cloud[n]   += mcol_cloud[n]*z;
		  mz2_cloud[n]   += mcol_cloud[n]*z*z;
		  mvz_cloud[n] += mcol_cloud[n]*vz;
		  mvz2_cloud[n] += mcol_cloud[n]*vz*vz;
		}
	      }
	      if (col >= 0.5){ // core material
		mcol_core[n] = mass*col;
		mc_core[n]   += mcol_core[n];
		mx_core[n]   += mcol_core[n]*x;
		my_core[n]   += mcol_core[n]*y;
		mx2_core[n]  += mcol_core[n]*x*x;
		my2_core[n]   += mcol_core[n]*y*y;
		mvx_core[n]  += mcol_core[n]*vx;
		mvy_core[n] += mcol_core[n]*vy;
		mvx2_core[n] += mcol_core[n]*vx*vx;
		mvy2_core[n] += mcol_core[n]*vy*vy;
		vc_core[n]   += cvol*col;
		if (nd == 3){
		  mz_core[n]   += mcol_core[n]*z;
		  mz2_core[n]   += mcol_core[n]*z*z;
		  mvz_core[n] += mcol_core[n]*vz;
		  mvz2_core[n] += mcol_core[n]*vz*vz;
		}
	      }
	      //integrate the kinetic, thermal, and turbulent energy of 
	      //material from individual CLOUDS
	      //(k is the turbulent energy per unit mass, qa[iqk] is rhok,
	      //which is the turbulent energy density with dimension energy/volume)
	      ke = 0.5*dens*(vx*vx + vy*vy + vz*vz); //kinetic energy density of cell
	      kiemc[n] += ke*col*cvol;
	      ie = cptr->qa[iqe] - ke; //thermal energy of cell
	      themc[n] += ie*col*cvol;//therm. energy = int. energy density*volume
	      tuemc[n] += cptr->qa[iqk]*col*cvol;
	    }//loop over individual clouds
	   
	    //So far we have the various energies for cloud material ONLY.
	    //Now integrate the TOTAL kinetic, thermal, and turbulent energy
	    //of the flow
	    tkie += ke*cvol;
	    tthe += ie*cvol;
	    ttue += cptr->qa[iqk]*cvol;
	    
	    //construct the vorticity and the integerated vorticity (circulation)
	    omegax = cptr->dudx[2][1] - cptr->dudx[1][2]; // (dv_z/dy - dv_y/dz)
	    omegay = cptr->dudx[0][2] - cptr->dudx[2][0]; // (dv_x/dz - dv_z/dx)
	    omegaz = cptr->dudx[1][0] - cptr->dudx[0][1]; // (dv_y/dx - dv_x/dy)
	    //For models without keps dudx is not calculated. Therefore I'll need
	    //to calculate the gradient myself.
	    if (nd == 2)
	      circ += omegaz*dx*dy; // omega . dA
	    else circ += omegax*dy*dz + omegay*dx*dz + omegaz*dx*dy; // omega . dA
	  }//finest cell
	  cptr = cptr->nxt;
	} //next cell
	ilev++; //changed "for" loop to "while" loop
	//	if (mype == 0) cerr <<  "Next Level" <<endl;
      } //cycle over grid levels
      //RA:find shockMaxX, shockMinX, shockAveX
      //}

      /*Temporary debug
	sprintf(numc, "%d", mype);
	strcpy(tempfile,prefx);
	strcat(tempfile,"temp");
	strcat(tempfile,numc);
	strcat(tempfile,".dat");
	fptr = fopen(tempfile,"w");
	if (fptr!=NULL){
	//fputs ("fopen example",pFile); 
	//    for (i=0; i< (g_iMax[mype]+20); i++) {
	//ORIGINAL     for (i=0; i< iMax ; i++) {
	//INDEXED 
	for (i=0; i<=iMax; i++) {
	fprintf(fptr,"%f, %i\n", shockPosX[i], shockPosY[i]);
	}
	fclose (fptr);
	}*/
      
      //To compute global quantities we need to integrate over all processors
     if (limpe > 1){               //limpe = 1 if single processor
	for (n=2; n <= nal; n++){
	  //	  if (mype == 0) cerr << "Limpe > 1 " << endl;
	  mc_cld[n] = global_sum(mc_cloud[n]);
	  vc_cld[n] = global_sum(vc_cloud[n]);
	  mx_cld[n] = global_sum(mx_cloud[n]);
	  my_cld[n] = global_sum(my_cloud[n]);
	  mx2_cld[n] = global_sum(mx2_cloud[n]);
	  my2_cld[n] = global_sum(my2_cloud[n]);
	  mvx_cld[n] = global_sum(mvx_cloud[n]);
	  //mvy_cld[n] = global_sum(mvy_cloud[n]);
	  //mvz_cld[n] = global_sum(mvz_cloud[n]);
	  mvx2_cld[n] = global_sum(mvx2_cloud[n]);
	  mvy2_cld[n] = global_sum(mvy2_cloud[n]);
	  mvz2_cld[n] = global_sum(mvz2_cloud[n]);
	  if (nd == 3){
	    mz_cld[n] = global_sum(mz_cloud[n]);
	    mz2_cld[n] = global_sum(mz2_cloud[n]);
	    mvz2_cld[n] = global_sum(mvz2_cloud[n]);
	  }
	  
	  mc_cor[n] = global_sum(mc_core[n]);
	  vc_cor[n] = global_sum(vc_core[n]);
	  mx_cor[n] = global_sum(mx_core[n]);
	  my_cor[n] = global_sum(my_core[n]);
	  mx2_cor[n] = global_sum(mx2_core[n]);
	  my2_cor[n] = global_sum(my2_core[n]);
	  mvx_cor[n] = global_sum(mvx_core[n]);
	  //mvy_cor[n] = global_sum(mvy_core[n]);
	  //mvz_cor[n] = global_sum(mvz_core[n]);
	  mvx2_cor[n] = global_sum(mvx2_core[n]);
	  mvy2_cor[n] = global_sum(mvy2_core[n]);
	  if (nd == 3){
	    mz_cor[n] = global_sum(mz_core[n]);
	    mz2_cor[n] = global_sum(mz2_core[n]);
	    mvz2_cor[n] = global_sum(mvz2_core[n]);
	  }
	  
	  kiecld[n]   = global_sum(kiemc[n]);//KE of cloud material from blocks 
	  thecld[n]   = global_sum(themc[n]);//of clouds and individual clouds
	  tuecld[n]   = global_sum(tuemc[n]);
	}
	kietot  = global_sum(tkie);
	thetot  = global_sum(tthe);
	tuetot  = global_sum(ttue);
	circ  = global_sum(circ );
	/*	for (i = 0; i < 2000; i++){
	    rhob[i] = global_sum( rhob[i] );
	    massb[i] = global_sum( massb[i]);
	    for (n = 0; n <= nal; n++){
	      rhob_single[n][i] = global_sum( rhob_single[n][i] );
	      massb_single[n][i] = global_sum( massb_single[n][i] );
	}  
	}*/ // bins
      }
        else if (limpe == 1){               //limpe = 1 if single processor
	for (n=2; n <= nal; n++){
	  mc_cld[n] = (mc_cloud[n]);
	  vc_cld[n] = (vc_cloud[n]);
	  mx_cld[n] = (mx_cloud[n]);
	  my_cld[n] = (my_cloud[n]);
	  mx2_cld[n] = (mx2_cloud[n]);
	  my2_cld[n] = (my2_cloud[n]);
	  mvx_cld[n] = (mvx_cloud[n]);
	  //mvy_cld[n] = (mvy_cloud[n]);
	  //mvz_cld[n] = (mvz_cloud[n]);
	  mvx2_cld[n] = (mvx2_cloud[n]);
	  mvy2_cld[n] = (mvy2_cloud[n]);
	  mvz2_cld[n] = (mvz2_cloud[n]);
	  if (nd == 3){
	    mz_cld[n] = (mz_cloud[n]);
	    mz2_cld[n] = (mz2_cloud[n]);
	    mvz2_cld[n] = (mvz2_cloud[n]);
	  }
	  
	  mc_cor[n] = (mc_core[n]);
	  vc_cor[n] = (vc_core[n]);
	  mx_cor[n] = (mx_core[n]);
	  my_cor[n] = (my_core[n]);
	  mx2_cor[n] = (mx2_core[n]);
	  my2_cor[n] = (my2_core[n]);
	  mvx_cor[n] = (mvx_core[n]);
	  //mvy_cor[n] = (mvy_core[n]);
	  //mvz_cor[n] = (mvz_core[n]);
	  mvx2_cor[n] = (mvx2_core[n]);
	  mvy2_cor[n] = (mvy2_core[n]);
	  if (nd == 3){
	    mz_cor[n] = (mz_core[n]);
	    mz2_cor[n] = (mz2_core[n]);
	    mvz2_cor[n] = (mvz2_core[n]);
	  }
	  
	  kiecld[n]   = (kiemc[n]);//KE of cloud material from blocks 
	  thecld[n]   = (themc[n]);//of clouds and individual clouds
	  tuecld[n]   = (tuemc[n]);
	}
	kietot  = (tkie);
	thetot  = (tthe);
	tuetot  = (ttue);
	circ  = (circ );
	/*	for (i = 0; i < 2000; i++){
	    rhob[i] = ( rhob[i] );
	    massb[i] = ( massb[i]);
	    for (n = 0; n <= nal; n++){
	      rhob_single[n][i] = global_sum( rhob_single[n][i] );
	      massb_single[n][i] = global_sum( massb_single[n][i] );
	}  
	}*/ // bins
      }
      //compute global quantities using the colour approach
      if ((limpe == 1)||(mype == 0)){// ie if single processor or master
	//	cerr << "master Processor\n";
	for (n=2; n <= 5; n++){//for each advected scalar (blocks of clouds)
	  //	   cerr << "blocks\n";
	    //	    cout << n << " " << mx_cld[n] << " " << mc_cld[n] << "\n";
	  if ((mc_cld[n] != 0.0)&&(vc_cld[n] != 0.0)){
	    //	   cerr << n << " " << mx_cld[n] << " " << mc_cld[n] << " " << vc_cld[n] <<  "\n";
	    //cout << n << " " << mx_cld[n] << " " << mc_cld[n] << "\n";
	    xm_cld[n] = mx_cld[n]/mc_cld[n]; //mass weighted x
	    ym_cld[n] = my_cld[n]/mc_cld[n]; //mass weighted y
	    x2m_cld[n] = mx2_cld[n]/mc_cld[n]; //mass weighted x^2
	    vxm_cld[n] = mvx_cld[n]/mc_cld[n]; //mass weighted vx
	    vx2m_cld[n] = mvx2_cld[n]/mc_cld[n]; //mass weighted vx^2
	    //	    cerr << n << " " << mvy2_cld[n] << " " << mc_cld[n] << "\n";
	    vy2m_cld[n] = mvy2_cld[n]/mc_cld[n]; //mass weighted vy^2
	    //if (x2m_cld[n] != 0.0){
	    //	    cerr << x2m_cld[n] - xm_cld[n]*xm_cld[n] << "x disp\n";
	    dx_cld[n] = sqrt(x2m_cld[n] - xm_cld[n]*xm_cld[n]); //x dispersion
	    //	    cerr << vx2m_cld[n] - vxm_cld[n]*vxm_cld[n] << "vx disp \n";
	    dvx_cld[n] = sqrt(vx2m_cld[n] - vxm_cld[n]*vxm_cld[n]); //vx dispersion
	    dvy_cld[n] = sqrt(vy2m_cld[n]); //vy dispersion
	    if (nd == 3){
	      zm_cld[n] = mz_cld[n]/mc_cld[n]; //mass weighted z
	      vz2m_cld[n] = mvz2_cld[n]/mc_cld[n]; //mass weighted vz^2
	      dvz_cld[n] = sqrt(vz2m_cld[n]); //vz dispersion
	    }
	    if (vc_cld[n] != 0.0) rhom_cld[n] = mc_cld[n]/vc_cld[n]; //mean density of cloud material
	    else rhom_cld[n] = 0.0;
	  }
	  if ((mc_cor[n] != 0.0)&&(vc_cor[n] != 0.0)){
	    xm_cor[n] = mx_cor[n]/mc_cor[n]; //mass weighted x
	    ym_cor[n] = my_cor[n]/mc_cor[n]; //mass weighted y
	    x2m_cor[n] = mx2_cor[n]/mc_cor[n]; //mass weighted x^2
	    vxm_cor[n] = mvx_cor[n]/mc_cor[n]; //mass weighted vx
	    vx2m_cor[n] = mvx2_cor[n]/mc_cor[n]; //mass weighted vx^2
	    vy2m_cor[n] = mvy2_cor[n]/mc_cor[n]; //mass weighted vy^2
      //if (x2m_cor[n] != 0.0){
	    dx_cor[n] = sqrt(x2m_cor[n] - xm_cor[n]*xm_cor[n]); //x dispersion
	    dvx_cor[n] = sqrt(vx2m_cor[n] - vxm_cor[n]*vxm_cor[n]); //vx dispersion
	    dvy_cor[n] = sqrt(vy2m_cor[n]); //vy dispersion
	    if (nd == 3){
	      zm_cor[n] = mz_cor[n]/mc_cor[n]; //mass weighted z
	      vz2m_cor[n] = mvz2_cor[n]/mc_cor[n]; //mass weighted vz^2
	      dvz_cor[n] = sqrt(vz2m_cor[n]); //vz dispersion
	    }
	    if (vc_cor[n] != 0.0) rhom_cor[n] = mc_cor[n]/vc_cor[n]; //mean density of core material
	    else rhom_cor[n] = 0.0;
	  }
	  //normalize global quantities and output to a file
	  vxm_cld[n] /= vb; 
	  dvx_cld[n] /= vb;
	  dvy_cld[n] /= vb;
	  vxm_cor[n] /= vb; 
	  dvx_cor[n] /= vb;
	  dvy_cor[n] /= vb;
	  if (nd == 3){
	    dvz_cld[n] /= vb;
	    dvz_cor[n] /= vb;
	  }
    }//blocks of clouds
	for (n=6; n <= nal; n++){//for each individually monitored cloud)
	  //	  	   cerr << "individuals\n";
	  if ((mc_cld[n] != 0.0)&&(vc_cld[n] != 0.0)){
	    xm_cld[n] = mx_cld[n]/mc_cld[n]; //mass weighted x
	    ym_cld[n] = my_cld[n]/mc_cld[n]; //mass weighted y
	    x2m_cld[n] = mx2_cld[n]/mc_cld[n]; //mass weighted x^2
	    y2m_cld[n] = my2_cld[n]/mc_cld[n]; //mass weighted x^y
	    vxm_cld[n] = mvx_cld[n]/mc_cld[n]; //mass weighted vx
	    vx2m_cld[n] = mvx2_cld[n]/mc_cld[n]; //mass weighted vx^2
	    vy2m_cld[n] = mvy2_cld[n]/mc_cld[n]; //mass weighted vy^2
	    //   cerr << x2m_cld[n] - xm_cld[n]*xm_cld[n] << " x disp\n";
	    dx_cld[n] = sqrt(mabs(x2m_cld[n] - xm_cld[n]*xm_cld[n])); //x dispersion
	    //cerr << y2m_cld[n] - ym_cld[n]*ym_cld[n] << " y disp\n";
	    dy_cld[n] = sqrt(mabs(y2m_cld[n] - ym_cld[n]*ym_cld[n])); //y dispersion
	    //	        cerr << vx2m_cld[n] - vxm_cld[n]*vxm_cld[n] << "vx disp\n";
	    dvx_cld[n] = sqrt(mabs(vx2m_cld[n] - vxm_cld[n]*vxm_cld[n])); //vx dispersion
	    //	        cerr << vy2m_cld[n] << "vy disp\n";
	    dvy_cld[n] = sqrt(vy2m_cld[n]); //vy dispersion
	    //cerr << " End of Dispersions\n";
	    if (nd == 3){
	      zm_cld[n] = mz_cld[n]/mc_cld[n]; //mass weighted z
	      z2m_cld[n] = mz2_cld[n]/mc_cld[n]; //mass weighted z^2
	      vz2m_cld[n] = mvz2_cld[n]/mc_cld[n]; //mass weighted vz^2
	      dz_cld[n] = sqrt(mabs(z2m_cld[n] - zm_cld[n]*zm_cld[n])); //z dispersion
	      dvz_cld[n] = sqrt(vz2m_cld[n]); //vz dispersion
	    }
	    if (vc_cld[n] != 0.0) rhom_cld[n] = mc_cld[n]/vc_cld[n]; //mean density of cloud material
	    else rhom_cld[n] = 0.0;
	    //    cerr << "End of Loop\n";
	  }
	  if ((mc_cor[n] != 0.0)&&(vc_cor[n] != 0.0)){
	    //	    cerr << "Core\n";
	    xm_cor[n] = mx_cor[n]/mc_cor[n]; //mass weighted x
	    ym_cor[n] = my_cor[n]/mc_cor[n]; //mass weighted y
	    x2m_cor[n] = mx2_cor[n]/mc_cor[n]; //mass weighted x^2
	    y2m_cor[n] = my2_cor[n]/mc_cor[n]; //mass weighted y^2
	    vxm_cor[n] = mvx_cor[n]/mc_cor[n]; //mass weighted vx
	    vx2m_cor[n] = mvx2_cor[n]/mc_cor[n]; //mass weighted vx^2
	    vy2m_cor[n] = mvy2_cor[n]/mc_cor[n]; //mass weighted vy^2
	    dx_cor[n] = sqrt(mabs(x2m_cor[n] - xm_cor[n]*xm_cor[n])); //x dispersion
	    dy_cor[n] = sqrt(mabs(y2m_cor[n] - ym_cor[n]*ym_cor[n])); //y dispersion
	    dvx_cor[n] = sqrt(mabs(vx2m_cor[n] - vxm_cor[n]*vxm_cor[n])); //vx dispersion
	    dvy_cor[n] = sqrt(vy2m_cor[n]); //vy dispersion
	    if (nd == 3){
	      zm_cor[n] = mz_cor[n]/mc_cor[n]; //mass weighted z
	      z2m_cor[n] = mz2_cor[n]/mc_cor[n]; //mass weighted z^2
	      vz2m_cor[n] = mvz2_cor[n]/mc_cor[n]; //mass weighted vz^2
	      dz_cor[n] = sqrt(mabs(z2m_cor[n] - zm_cor[n]*zm_cor[n])); //z dispersion
	      dvz_cor[n] = sqrt(vz2m_cor[n]); //vz dispersion
	    }
	    if (vc_cor[n] != 0.0) rhom_cor[n] = mc_cor[n]/vc_cor[n]; //mean density of core material
	    else rhom_cor[n] = 0.0;
	    //  cerr << "end_core\n";
	  }      

	  //normalize global quantities and output to a file
	  vxm_cld[n] /= vb; 
	  dvx_cld[n] /= vb;
	  dvy_cld[n] /= vb;
	  vxm_cor[n] /= vb; 
	  dvx_cor[n] /= vb;
	  dvy_cor[n] /= vb;
	  if (nd == 3){
	    dvz_cld[n] /= vb;
	    dvz_cor[n] /= vb;
	  }
	}//individual clouds
	for (n=1; n <= nal; n++){//for every scalar
	  etotcld[n] = kiecld[n] + thecld[n] + tuecld[n]; //total energy of cloud material
	  if (etotcld[n] != 0.0){
	    kiecld[n] /= etotcld[n]; //ke fraction of cloud material
	    thecld[n] /= etotcld[n];
	    tuecld[n] /= etotcld[n];
	  }
	} //loop over NAL
	etot = kietot + thetot + tuetot; //total energy of ALL material on grid
	kietot /= etot; //ke fraction of ALL material
	thetot /= etot;
	tuetot /= etot;

    //append info to a file (read and write access)
	for (n=2; n <= 5; n++){//for each block of clouds
	  //	  cerr << "Write out blocks\n";
	  nf = 999 + n;
	  sprintf(numc,"%d",nf);
	  strcpy(statsfile,prefx);
	  strcat(statsfile,"blk");
	  strcat(statsfile,numc);
	  strcat(statsfile,"_cld_stats.dat");
	  fptr = fopen(statsfile,"a+"); 
	  //fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	  if (fptr!=NULL){ 
	    //fputs ("fopen example",pFile); 
  	    fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
	 	    t,t/tcc,t*tscale/yr,xm_cld[n],dx_cld[n],mc_cld[n],vc_cld[n],rhom_cld[n],vxm_cld[n],dvx_cld[n],dvy_cld[n],dvz_cld[n]);
	    fclose (fptr);
	  }
	  strcpy(statsfile,prefx);
	  strcat(statsfile,"blk");
	  strcat(statsfile,numc);
	  strcat(statsfile,"_core_stats.dat");
	  fptr = fopen(statsfile,"a+");
	  //fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	  if (fptr!=NULL){
	    //fputs ("fopen example",pFile); 
	    fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		    t,t/tcc,t*tscale/yr,xm_cor[n],dx_cor[n],mc_cor[n],vc_cor[n],rhom_cor[n],vxm_cor[n],dvx_cor[n],dvy_cor[n],dvz_cor[n]);
	    fclose (fptr);
	    }  //SECOND FILE
	   //write out mass in (velocity) bins
	  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	  /*	  strcpy(statsfile,prefx);
	  strcat(statsfile,"blk");
	  strcat(statsfile, numc);
	  strcat(statsfile,"_mv.dat_t");
	  sprintf(numc2,"%d",int(t));
	  strcat(statsfile, numc2);
	  fptr = fopen(statsfile,"w");
	  if (fptr!=NULL){
	    mass = 0.0;
	    for (i = 0; i < 200; i++){
	      mass += massb_single[n][i];
	      fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/100.0,massb_single[n][i],mass);	     
	    }
	    fclose (fptr);
	  }
  

  
	    //write out mass in (density) bins
	  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	  strcpy(statsfile,prefx);
	  strcat(statsfile,"blk");
	  strcat(statsfile, numc);
	  strcat(statsfile,"_mrho.dat_t");
	  strcat(statsfile, numc2);	  
	  fptr = fopen(statsfile,"w");
	  if (fptr!=NULL){
	    mass = 0.0;
	    for (i = 0; i < 2000; i++){
	      mass += rhob_single[n][i];
	      fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/50.0,rhob_single[n][i],mass);	     
	    }
	    fclose (fptr);
	    } //Bins
*/
    } //block of clouds
	for (n=6; n <= nal; n++){//for each individually monitored cloud
	  //	  cerr << "Write out individuals\n";
	    	  	  nf = 999 + n;
	  sprintf(numc,"%d",nf);
	  strcpy(statsfile,prefx);
	  strcat(statsfile,"single");
	  strcat(statsfile,numc);
	  strcat(statsfile,"_cld_stats.dat");
	  fptr = fopen(statsfile,"a+");
	  //fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	  if (fptr!=NULL){
	    //fputs ("fopen example",pFile); 
	    fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		    t,t/tcc,t*tscale/yr,xm_cld[n],ym_cld[n],zm_cld[n],dx_cld[n],dy_cld[n],dz_cld[n],mc_cld[n],vc_cld[n],rhom_cld[n],vxm_cld[n],dvx_cld[n],dvy_cld[n],dvz_cld[n]);
	    fclose (fptr);
	  }
	  strcpy(statsfile,prefx);
	  strcat(statsfile,"single");
	  strcat(statsfile,numc);
	  strcat(statsfile,"_core_stats.dat");
	  fptr = fopen(statsfile,"a+");
	  //fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	  if (fptr!=NULL){
	    //fputs ("fopen example",pFile); 
	    fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		    t,t/tcc,t*tscale/yr,xm_cor[n],ym_cor[n],zm_cor[n],dx_cor[n],dy_cor[n],dz_cor[n],mc_cor[n],vc_cor[n],rhom_cor[n],vxm_cor[n],dvx_cor[n],dvy_cor[n],dvz_cor[n]);
	    fclose (fptr);
	    }//THIRD FILE
	  	  //write out mass in (velocity) bins
	  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	  /*	  strcpy(statsfile,prefx);
	  strcat(statsfile,"single");
	  strcat(statsfile, numc);
	  strcat(statsfile,"_mv.dat_t");
	  sprintf(numc2,"%d",int(t));
	  strcat(statsfile, numc2);
	  fptr = fopen(statsfile,"w");
	  if (fptr!=NULL){
	    mass = 0.0;
	    for (i = 0; i < 200; i++){
	      mass += massb_single[n][i];
	      fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/100.0,massb_single[n][i],mass);	     
	    }
	    fclose (fptr);
	  }
  

	  
	  //write out mass in (density) bins
	  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	  strcpy(statsfile,prefx);
	  strcat(statsfile,"single");
	  strcat(statsfile, numc);
	  strcat(statsfile,"_mrho.dat_t");
	  strcat(statsfile, numc2);	  
	  fptr = fopen(statsfile,"w");
	  if (fptr!=NULL){
	    mass = 0.0;
	    for (i = 0; i < 2000; i++){
	      mass += rhob_single[n][i];
	      fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/50.0,rhob_single[n][i],mass);	     
	    }
	    fclose (fptr);
	    }*/ // bins
	} //individual clouds

	//energies
		for (n=2; n <= nal; n++){//for each block of clouds and individual clouds
	  //	  cerr << "Write out energies\n";
	  	  escale = (rhoscale*uscale*uscale)*(rscale/1.0e15)*(rscale/1.0e15)*rscale;
	  etotcld[n] *= escale;
	  nf = 999 + n;
	  sprintf(numc,"%d",nf);
	  strcpy(statsfile,prefx);
	  strcat(statsfile,"energy");
	  strcat(statsfile,numc);
	  strcat(statsfile,"_cld_stats.dat");
	  fptr = fopen(statsfile,"a+");
	  //fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	  if (fptr!=NULL){
	    //fputs ("fopen example",pFile); 
	    fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		    t,t/tcc,t*tscale/yr,kiecld[n],thecld[n],tuecld[n],etotcld[n],tuecld[n]*etotcld[n]);
	    fclose (fptr);
	  }
	}
	  //loop over clouds
	//Now the TOTAL energies on the ENTIRE grid! (Plus the total circulation)
	etot *= escale;
	strcpy(statsfile,prefx);
	strcat(statsfile,"totalenergy_cld_stats.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,kietot,thetot,tuetot,etot,kietot*etot,thetot*etot,tuetot*etot,circ/-vb);
	  fclose (fptr);
	}
        //FOURTH FILE

	/*	//write out mass in (velocity) bins
	//FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	strcpy(statsfile,"mv.dat");
	sprintf(numc,"%d",int(t));
	strcat(statsfile, numc);
       	fptr = fopen(statsfile,"w");
	if (fptr!=NULL){
	  mass = 0.0;
	  for (i = 0; i < 200; i++){
	    mass += massb[i];
	    fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/100.0,massb[i],mass);	     
	  }
	  fclose (fptr);
	}
  
	//write out mass in (density) bins
	//FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	strcpy(statsfile,"mrho.dat");
	cout << "time: " << t << " numc:" << numc << endl;
	sprintf(numc,"%d",int(t));
	strcat(statsfile, numc);
	fptr = fopen(statsfile,"w");
	if (fptr!=NULL){
	  mass = 0.0;
	  for (i = 0; i < 2000; i++) { 
	    //	    cout << "mype " << mype << "i " << i << " rhob: " << rhob[i] << endl; 	  
	    mass += rhob[i];
	    fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/50.0,rhob[i],mass);	     
	  }
	}
	  fclose (fptr);
	*/ //bins
      } //master processor
      //    cerr << mype << " outside manycloud\n";
    } //many clouds
    //if (mype == 0) cerr << "[/usrsum]" << endl;
  } //SINGLECLOUD or MANYCLOUD
  

  //FILAMENT case
  if (strcmp(cas, "FILAMENT") == 0){
    //    if (mype == 0) cerr << "SC or MC\n";
    //zero global quantities and initial summation indices
    for (i = 0; i < nmax; i++) {
      massb[i] = 0.0;
      rhob[i] = 0.0;
      for (n = 0; n <= nal; n++) {
	massb_single[n][i] = 0.0;
	rhob_single[n][i] = 0.0;
      }
    }
    
    mc = 0.0;
    mc_2 = 0.0;
    mc_3 = 0.0;

    mx  = 0.0;
    my  = 0.0;
    mz  = 0.0;
    mx2 = 0.0;
    my2 = 0.0;
    mz2 = 0.0;
    mr2 = 0.0;

    mvx = 0.0;
    mvy = 0.0;
    mvz = 0.0;
    mvr = 0.0;
    mvx2 = 0.0;
    mvy2 = 0.0;
    mvz2 = 0.0;
    mvr2 = 0.0;
    
    xm = 0.0;
    ym = 0.0;
    zm = 0.0;
    rm = 0.0;
    xms = 0.0;
    yms = 0.0;
    zms = 0.0;
    rms = 0.0;

    vxm = 0.0;
    vym = 0.0;
    vzm = 0.0;
    vrm = 0.0;

    vxms= 0.0;
    vyms= 0.0;
    vzms= 0.0;
    vrms= 0.0;

    mx_2  = 0.0;
    my_2  = 0.0;
    mz_2  = 0.0;
    mx2_2 = 0.0;
    my2_2 = 0.0;
    mz2_2 = 0.0;
    mr2_2 = 0.0;

    mvx_2 = 0.0;
    mvy_2 = 0.0;
    mvz_2 = 0.0;
    mvr_2 = 0.0;
    mvx2_2 = 0.0;
    mvy2_2 = 0.0;
    mvz2_2 = 0.0;
    mvr2_2 = 0.0;
    
    xm_2 = 0.0;
    ym_2 = 0.0;
    zm_2 = 0.0;
    rm_2 = 0.0;
    xms_2 = 0.0;
    yms_2 = 0.0;
    zms_2 = 0.0;
    rms_2 = 0.0;

    vxm_2 = 0.0;
    vym_2 = 0.0;
    vzm_2 = 0.0;
    vrm_2 = 0.0;

    vxms_2= 0.0;
    vyms_2= 0.0;
    vzms_2= 0.0;
    vrms_2= 0.0;

    mx_3  = 0.0;
    my_3  = 0.0;
    mz_3  = 0.0;
    mx2_3 = 0.0;
    my2_3 = 0.0;
    mz2_3 = 0.0;
    mr2_3 = 0.0;

    mvx_3 = 0.0;
    mvy_3 = 0.0;
    mvz_3 = 0.0;
    mvr_3 = 0.0;
    mvx2_3 = 0.0;
    mvy2_3 = 0.0;
    mvz2_3 = 0.0;
    mvr2_3 = 0.0;
    
    xm_3 = 0.0;
    ym_3 = 0.0;
    zm_3 = 0.0;
    rm_3 = 0.0;
    xms_3 = 0.0;
    yms_3 = 0.0;
    zms_3 = 0.0;
    rms_3 = 0.0;

    vxm_3 = 0.0;
    vym_3 = 0.0;
    vzm_3 = 0.0;
    vrm_3 = 0.0;

    vxms_3= 0.0;
    vyms_3= 0.0;
    vzms_3= 0.0;
    vrms_3= 0.0;

    volc = 0.0;
    volc_2 = 0.0;
    volc_3 = 0.0;
    vc = 0.0;
    vc_2 = 0.0;
    vc_3 = 0.0;
    circ= 0.0;
    rhomax = 0.0;
        
    kie = 0.0;
    the = 0.0;
    tue = 0.0;

    tkie = 0.0;
    tthe = 0.0;
    ttue = 0.0;
    
    //define normalization
    cs = sqrt(g*kboltz*tamb/avgmass);  //sound speed in initial intercloud medium
    vb = mach*cs/uscale;               //speed of shock in IIM (mg_gt units)
    rhom0 = ramb*chi/rhoscale;         //initial max density of cloud
    tcc = sqrt(chi)/vb;                //cloud crushing timescale (radius=1)
   
    // Sum over grids from G1 to GMAX (nlevs-1 = GMAX)
    int levmin, levmax;
    if (nlevs == 1){
      levmin = 0;
      levmax = 0;
    }
    else{
      levmin = 1;
      levmax = nlevs-1;
    }

    //Calculate the other global statistics 
    /* Sum over grids from plmin to plmax */ 
    //for (ilev = 0; ilev <= nlevs-1; ilev++){ //sum over ALL grids
    for (ilev = levmin; ilev <= levmax; ilev++){//sum over ALL grids
	//for (ilev = plmin; ilev <= plmax; ilev++){ //sum over plot grids
	cptr = lev[ilev]->fcell;
	while (cptr != NULL){
	  if (!cptr->refined){    //if finest cell then...
	    //	    if (mype == 0) cout << "cell # = " << cptr->icell << "\n";
	    //	    if (mype == 10) cerr << "cell # = " << cptr->icell << " mype= " << mype << endl;
	    cvol = cptr->vol;
	    for (i = 0; i < nq; i++)
	      tot[i] += cvol*cptr->qa[i];
	    dens = cptr->qa[iqd];
	    col = cptr->qa[iqal0+1]/dens;
	    colamb = 1.0/chi;
	    if (dens > rhomax) rhomax = dens;
	    mass = dens*cvol;
	    
	    dx = lev[ilev]->dxbar[0];
	    dy = lev[ilev]->dxbar[1];
	    dz = lev[ilev]->dxbar[2];
	    x = cptr->xc[0];  //xc[] values are the geometric centre of the cell
	    y = cptr->xc[1];
	    z = cptr->xc[2];
	    r = sqrt(y*y + z*z); // Shock propagates along x-axis by definition so r is transverse distance

	    vx = cptr->qa[iqu0+1]/dens;
	    vy = cptr->qa[iqu0+2]/dens;
	    vz = cptr->qa[iqu0+3]/dens;
	    vr = sqrt(vy*vy + vz*vz);
	    
	    //Construct moments using the colour variable. Define
	    //cloud material as those with col > 2.0 times the ambient colour
	    //cout << chi << " " << colamb << "\n";
	    //exit(EXIT_FAILURE);
	    //if (mype == 0){
          //  cout << "r = " << r << ", z = " << z << ", dens = " << dens << ", vol = " << cvol << ", mass = " << mass << "\n";
	    //}
	    if (col >= 2.0*colamb){ // cell material is mostly cloud
	      mcol = mass;
	      mc   += mcol;
	      mx   += mcol*x;
	      my   += mcol*y;
	      mz   += mcol*z;
	      mr   += mcol*r;
	      mx2  += mcol*x*x;
	      my2  += mcol*y*y;
	      mz2  += mcol*z*z;
	      mr2  += mcol*r*r;
	      mvx  += mcol*vx;
	      mvy  += mcol*vy;
	      mvz  += mcol*vz;
	      mvr  += mcol*vr;
	      mvx2 += mcol*vx*vx;
	      mvy2 += mcol*vy*vy;
	      mvz2 += mcol*vz*vz;
	      mvr2 += mcol*vr*vr;
	      vc   += cvol;
	    }
	    
	    //only count cloud material (col=0, outside r=1.5)
	    if (col >= 2.0*colamb){
	      mcol_2 = mass*col;
	      mc_2   += mcol_2;
	      mx_2   += mcol_2*x;
	      my_2   += mcol_2*y;
	      mz_2   += mcol_2*z;
	      mr_2   += mcol_2*r;
	      mx2_2  += mcol_2*x*x;
	      my2_2  += mcol_2*y*y;
	      mz2_2  += mcol_2*z*z;
	      mr2_2  += mcol_2*r*r;
	      mvx_2  += mcol_2*vx;
	      mvy_2  += mcol_2*vy;
	      mvz_2  += mcol_2*vz;
	      mvr_2  += mcol_2*vr;
	      mvx2_2 += mcol_2*vx*vx;
	      mvy2_2 += mcol_2*vy*vy;
	      mvz2_2 += mcol_2*vz*vz;
	      mvr2_2 += mcol_2*vr*vr;
	      vc_2   += cvol;
	      i = int(100.0*mabs(vx)/(0.75*vb));
	      massb[i] += mcol_2;
	      i = int(50.0*dens/rhom0);
	      rhob[i] += mcol_2;
	      //cout << i << "," << mcol_2 << "," << mc_2 << "," << massb[i] << "\n";          
	    }
	    
	    //only count cells with mostly cloud material (col=0, outside r=1.5)
	    if (col >= 0.5){ // cell material is mostly cloud
	      mcol_3 = mass*col;
	      mc_3   += mcol_3;
	      mx_3   += mcol_3*x;
	      my_3   += mcol_3*y;
	      mz_3   += mcol_3*z;
	      mr_3   += mcol_3*r;
	      mx2_3  += mcol_3*x*x;
	      my2_3  += mcol_3*y*y;
	      mz2_3  += mcol_3*z*z;
	      mr2_3  += mcol_3*r*r;
	      mvx_3  += mcol_3*vx;
	      mvy_3  += mcol_3*vy;
	      mvz_3  += mcol_3*vz;
	      mvr_3  += mcol_3*vr;
	      mvx2_3 += mcol_3*vx*vx;
	      mvy2_3 += mcol_3*vy*vy;
	      mvz2_3 += mcol_3*vz*vz;
	      mvr2_3 += mcol_3*vr*vr;
	      vc_3   += cvol;
	    }
	    
	    //integrate the cloud kinetic, thermal, and turbulent energy
	    //(k is the turbulent energy per unit mass, qa[iqk] is rhok,
	    //which is the turbulent energy density with dimension energy/volume)
	    ke = 0.5*dens*(vx*vx + vy*vy + vz*vz); //kinetic energy density
	    kie += ke*col*cvol;
	    ie = cptr->qa[iqe] - ke;
	    the += ie*col*cvol;//thermal energy = internal energy density * volume
	    tue += cptr->qa[iqk]*col*cvol;
	    
	    //integrate the total kinetic, thermal, and turbulent energy
	    tkie += ke*cvol;
	    tthe += ie*cvol;
	    ttue += cptr->qa[iqk]*cvol;
	    
	    //construct the vorticity and the integerated vorticity (circulation)
	    //NOTE:The calculation for omega is different to the runs in paper I 
	    // and II becuase of a change in the definition of dudx
	    omegax = cptr->dudx[2][1] - cptr->dudx[1][2]; // (dv_z/dy - dv_y/dz)
	    omegay = cptr->dudx[0][2] - cptr->dudx[2][0]; // (dv_x/dz - dv_z/dx)
	    omegaz = cptr->dudx[1][0] - cptr->dudx[0][1]; // (dv_y/dx - dv_x/dy)
	    //For models without keps dudx is not calculated. Therefore I'll need
	    //to calculate the gradient myself.
	    circ += omegax*dy*dz + omegay*dx*dz + omegaz*dx*dy; // omega . dA
	    
	    //fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
	    //	r,z,divdr,divdz,cptr->dudx[0][1],cptr->dudx[1][0],omega,circ);
	  } //finest cell
	  cptr = cptr->nxt;
	} //next cell
      } //cycle over grid levels
      
      //fclose (fptr);
      
      //normalize mass in (velocity and density) bins
      /*      if (mc_2 != 0.0){ // note that with multi processors mc_2 could be 0.0
	for (i = 0; i < nmax; i++)
	  massb[i] /= mc_2;
    for (i = 0; i < nmax; i++)
      rhob[i] /= mc_2;
      }*/
      
      //cout << " proc # = " << mype << "; mc = " << mc << "\n";
      //      cout << "returning from usrsum, proc # = " << mype << "\n";
      //return;
      
      //compute global quantities
      if (limpe > 1){               //limpe = 1 if single processor
	mc    = global_sum(mc );
	vc    = global_sum(vc );
	mx    = global_sum(mx );
	my    = global_sum(my );
	mz    = global_sum(mz );
	mr    = global_sum(mr );
	mx2   = global_sum(mx2);
	my2   = global_sum(my2);
	mz2   = global_sum(mz2);
	mr2   = global_sum(mr2);
	mvx   = global_sum(mvx);
	mvy   = global_sum(mvy);
	mvz   = global_sum(mvz);
	mvr   = global_sum(mvr);
	mvx2  = global_sum(mvx2);
	mvy2  = global_sum(mvy2);
	mvz2  = global_sum(mvz2);
	mvr2  = global_sum(mvr2);

	mc_2    = global_sum(mc_2 );
	vc_2    = global_sum(vc_2 );
	mx_2    = global_sum(mx_2 );
	my_2    = global_sum(my_2 );
	mz_2    = global_sum(mz_2 );
	mr_2    = global_sum(mr_2 );
	mx2_2   = global_sum(mx2_2);
	my2_2   = global_sum(my2_2);
	mz2_2   = global_sum(mz2_2);
	mr2_2   = global_sum(mr2_2);
	mvx_2   = global_sum(mvx_2);
	mvy_2   = global_sum(mvy_2);
	mvz_2   = global_sum(mvz_2);
	mvr_2   = global_sum(mvr_2);
	mvx2_2  = global_sum(mvx2_2);
	mvy2_2  = global_sum(mvy2_2);
	mvz2_2  = global_sum(mvz2_2);
	mvr2_2  = global_sum(mvr2_2);

	mc_3    = global_sum(mc_3 );
	vc_3    = global_sum(vc_3 );
	mx_3    = global_sum(mx_3 );
	my_3    = global_sum(my_3 );
	mz_3    = global_sum(mz_3 );
	mr_3    = global_sum(mr_3 );
	mx2_3   = global_sum(mx2_3);
	my2_3   = global_sum(my2_3);
	mz2_3   = global_sum(mz2_3);
	mr2_3   = global_sum(mr2_3);
	mvx_3   = global_sum(mvx_3);
	mvy_3   = global_sum(mvy_3);
	mvz_3   = global_sum(mvz_3);
	mvr_3   = global_sum(mvr_3);
	mvx2_3  = global_sum(mvx2_3);
	mvy2_3  = global_sum(mvy2_3);
	mvz2_3  = global_sum(mvz2_3);
	mvr2_3  = global_sum(mvr2_3);


	kie   = global_sum(kie  );
	the   = global_sum(the  );
	tue   = global_sum(tue  );
	tkie  = global_sum(tkie );
	tthe  = global_sum(tthe );
	ttue  = global_sum(ttue );
	circ  = global_sum(circ );
        if (!keps) circ = 0.0;
	rhomax= global_max(rhomax);
      }
      //cout << " 2: proc # = " << mype << "; mc = " << mc << "\n";
      //return;
      
      //compute global quantities using the colour approach
      if ((limpe == 1)||(mype == 0)){// ie if single processor or master
	xm  = mx/mc;
	ym  = my/mc;
	zm  = mz/mc;
	rm  = mr/mc;
	xms = mx2/mc;
	yms = my2/mc;
	zms = mz2/mc;
	rms = mr2/mc;

	vxm = mvx/mc;
	vym = mvy/mc;
	vzm = mvz/mc;
	vrm = mvr/mc;
	vxms= mvx2/mc;
	vyms= mvy2/mc;
	vzms= mvz2/mc;
	vrms= mvr2/mc;
	
	xm_2  = mx_2/mc_2;
	ym_2  = my_2/mc_2;
	zm_2  = mz_2/mc_2;
	rm_2  = mr_2/mc_2;
	xms_2 = mx2_2/mc_2;
	yms_2 = my2_2/mc_2;
	zms_2 = mz2_2/mc_2;
	rms_2 = mr2_2/mc_2;

	vxm_2 = mvx_2/mc_2;
	vym_2 = mvy_2/mc_2;
	vzm_2 = mvz_2/mc_2;
	vrm_2 = mvr_2/mc_2;
	vxms_2= mvx2_2/mc_2;
	vyms_2= mvy2_2/mc_2;
	vzms_2= mvz2_2/mc_2;
	vrms_2= mvr2_2/mc_2;
	
	xm_3  = mx_3/mc_3;
	ym_3  = my_3/mc_3;
	zm_3  = mz_3/mc_3;
	rm_3  = mr_3/mc_3;
	xms_3 = mx2_3/mc_3;
	yms_3 = my2_3/mc_3;
	zms_3 = mz2_3/mc_3;
	rms_3 = mr2_3/mc_3;

        // Save filament centre-of-mass into global variable in user.h (this is broadcast
        // to the other processors if necessary - see below)
        fil_com_x = xm_3;
        fil_com_y = ym_3;
        fil_com_z = zm_3;

	vxm_3 = mvx_3/mc_3;
	vym_3 = mvy_3/mc_3;
	vzm_3 = mvz_3/mc_3;
	vrm_3 = mvr_3/mc_3;
	vxms_3= mvx2_3/mc_3;
	vyms_3= mvy2_3/mc_3;
	vzms_3= mvz2_3/mc_3;
	vrms_3= mvr2_3/mc_3;
	
	//a = sqrt(2.5*rms);           // effective radius normal to axis
	//c = sqrt(5.0*(zms - zm*zm)); //                 along axis
	dvx = sqrt(vxms - vxm*vxm);  // axial velocity dispersion
	dvy = sqrt(vyms - vym*vym);  // axial velocity dispersion
	dvz = sqrt(vzms - vzm*vzm);  // axial velocity dispersion
	dvr = sqrt(vrms);            // radial velocity dispersion
	rhom = mc/vc;                // mean density of cloud
	
	//a_2 = sqrt(2.5*rms_2);           
	//c_2 = sqrt(5.0*(zms_2 - zm_2*zm_2)); 
	dvx_2 = sqrt(vxms_2 - vxm_2*vxm_2); 
	dvy_2 = sqrt(vyms_2 - vym_2*vym_2); 
	dvz_2 = sqrt(vzms_2 - vzm_2*vzm_2); 
	dvr_2 = sqrt(vrms_2);           
	rhom_2 = mc_2/vc_2;               
	
	//a_3 = sqrt(2.5*rms_3);           
	//c_3 = sqrt(5.0*(zms_3 - zm_3*zm_3)); 
	dvx_3 = sqrt(vxms_3 - vxm_3*vxm_3); 
	dvy_3 = sqrt(vyms_3 - vym_3*vym_3); 
	dvz_3 = sqrt(vzms_3 - vzm_3*vzm_3); 
	dvr_3 = sqrt(vrms_3);           
	rhom_3 = mc_3/vc_3;               

	//normalize global quantities and output to a file
	rhomax /= rhom0;
	rhom /= rhom0;
	mc /= mcloud;
	vc /= vcloud;
	vxm /= vb; 
	vym /= vb; 
	vzm /= vb; 
	dvx /= vb;
	dvy /= vb;
	dvz /= vb;
	dvr /= vb;
	
	rhom_2 /= rhom0;
	mc_2 /= mcloud;
	vc_2 /= vcloud;
	vxm_2 /= vb; 
	vym_2 /= vb; 
	vzm_2 /= vb; 
	dvx_2 /= vb;
	dvy_2 /= vb;
	dvz_2 /= vb;
	dvr_2 /= vb;
	
	rhom_3 /= rhom0;
	mc_3 /= mcloud;
	vc_3 /= vcloud;
	vxm_3 /= vb; 
	vym_3 /= vb; 
	vzm_3 /= vb; 
	dvx_3 /= vb;
	dvy_3 /= vb;
	dvz_3 /= vb;
	dvr_3 /= vb;

	etot = kie + the + tue; //total energy of cloud material
	kie /= etot;
	the /= etot;
	tue /= etot;
	
	//append info to a file (read and write access)
	//if ((limpe == 1)||(mype == 0)){//write if single processor or master
	//FILE *fptr; //pointer to file (the "FILE" must be uppercase)
	//strcpy(prefx,"m10.0chi1e3adialokepsres128circ"); //prefix for movie files
	//strcpy(prefx,"m10.0chi1e3adialokepsres64p21"); //prefix for movie files
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,xm,ym,zm,mc,vc,rhomax,rhom,vxm,vym,vzm,dvx,dvy,dvz,dvr,circ/-vb);
	  fclose (fptr);
	}
	
	//append info to a file (read and write access)
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats2.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
        //cout << "mc_2 = " << mc_2 << "\n";
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,xm_2,ym_2,zm_2,mc_2,vc_2,rhomax,rhom_2,vxm_2,vym_2,vzm_2,dvx_2,dvy_2,dvz_2,dvr_2);
	  fclose (fptr);
	}
	
	//append info to a file (read and write access)
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats3.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,xm_3,ym_3,zm_3,mc_3,vc_3,rhomax,rhom_3,vxm_3,vym_3,vzm_3,dvx_3,dvy_3,dvz_3,dvr_3);
	  fclose (fptr);
	}
	
	//append info to a file (read and write access)
	escale = (rhoscale*uscale*uscale)*(rscale/1.0e15)*(rscale/1.0e15)*rscale;
	etot *= escale;
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats4.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,kie,the,tue,etot,tue*etot);
	  fclose (fptr);
	}
	
        //append info to a file (read and write access)
	etot = tkie + tthe + ttue; //total energy of grid material
	tkie /= etot;
	tthe /= etot;
	ttue /= etot;
	etot *= escale;
	strcpy(statsfile,prefx);
	strcat(statsfile,"stats5.dat");
	fptr = fopen(statsfile,"a+");
	//fptr = fopen ("m10.0_dens1e3_adia_res64_stats.dat","a+");
	if (fptr!=NULL){
	  //fputs ("fopen example",pFile); 
	  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
		  t,t/tcc,t*tscale/yr,tkie,tthe,ttue,etot,tkie*etot,tthe*etot,ttue*etot);
	  fclose (fptr);
	}

      } // master proc

      // Broadcast filament centre of mass to the other processors (if necessary)
      if (limpe > 1){
        broadcast(&fil_com_x);
        broadcast(&fil_com_y);
        broadcast(&fil_com_z);
      }

  } // filament


  /*
  //write out mass in (velocity) bins
  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
  strcpy(statsfile,"mv.dat");
  fptr = fopen(statsfile,"w");
  if (fptr!=NULL){
    mass = 0.0;
    for (i = 0; i < 200; i++){
      mass += massb[i];
      fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/100.0,massb[i],mass);	     
    }
    fclose (fptr);
  }
  */

  /*
  //write out mass in (density) bins
  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
  strcpy(statsfile,"mrho.dat");
  fptr = fopen(statsfile,"w");
  if (fptr!=NULL){
    mass = 0.0;
    for (i = 0; i < 2000; i++){
      mass += rhob[i];
      fprintf(fptr,"%.4e, %.4e, %.4e\n", float(i)/10.0,rhob[i],mass);	     
    }
    fclose (fptr);
  }
  */  



  //Code for calculating X-ray spectra
  /*

  //First read in file storing emissivity data
  strcpy(emissfile,"solar_mekal_logcoarse_emiss0.1_10keV");
  // read in emissivity data: emiss(E) = (rho/mh)**2 Vol Lambda(E,T)
  // open file - for general input/output use fstream,
  //                 for input to the program only use ifstream,
  //                 for output from the program only use ofstream   
  // cout << "Opening " << emissfile << '\n'; 
  ifstream file(emissfile);  
  if (!file) {
    cerr << "Failed to open file\n";
    exit(EXIT_FAILURE);
  } 
  file.seekg(0); // goto start of file 
  for (nt = 1; nt < ntmax; nt++){
    file >> idum;
    file >> logt[nt];
    for (ne = 1; ne < nemax; ne++){
      file >> lambda[nt][ne];
      if (lambda[nt][ne] == 0.0) lambda[nt][ne] == 1.0e-50; //set floor value 
      //lambda[nt][ne] = log10(lambda[nt][ne]); //log emissivity
      //if ((nt==101)&&(ne==nemax-1)){
      //if (nt==24){
      //  cout << nt << " " << logt[nt] << " " << ne << " " << lambda[nt][ne] << '\n'; 
      //}
    }
  }
  //cout << lambda[24][87] << "\n";

  dlogt = logt[2]-logt[1];

  // construct logarithmic energy bins
  e1[1] = 0.1;
  e2[nemax-1] = 10.0;
  dloge = (log10(e2[nemax-1])-log10(e1[1]))/float(nemax-1);
  for (ne = 1; ne < nemax; ne++){
    e1[ne] = pow(10,(log10(e1[1]) + float(ne-1)*dloge));
    e2[ne] = pow(10,(log10(e1[1]) + float(ne)*dloge));
    ecen[ne] = (e1[ne]+e2[ne])/2.0;
    de[ne] = e2[ne]-e1[ne];
  }

  //zero spectrum
  for (ne = 1; ne < nemax; ne++)
    spec[ne] = 0.0;
  maxtemp = 0.0;

  //rescale to different values (best to set absolutely due to
  //possible repeated calls to this routine)
  //pscale *= 800.0;  //ie Tamb goes from 1.0e4 K to 8.0e6 K
  //uscale *= sqrt(800);
  //rscale *= 0.05;   //ie cloud radius goes from 2ps to 0.1pc
  //tscale = rscale/uscale;

  rscale = 0.1*pc;
  uscale = sqrt(800.0)*1.0e6;
  pscale = rhoscale*uscale*uscale;
  tscale = rscale/uscale;


  // Sum over grids from plmin to plmax  
  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
  //fptr = fopen("test.dat","a+");
  for (ilev = plmin; ilev <= plmax; ilev++){
    cptr = lev[ilev]->fcell;
    while (cptr != NULL){
      if (cptr->child == NULL){    //if finest cell then...
        dens = cptr->qa[iqd];
        dr = lev[ilev]->dxbar[0];
        dz = lev[ilev]->dxbar[1];
        r = cptr->xc[0] + 0.5*dr;
        z = cptr->xc[1] + 0.5*dz;
        cvol2 = gkpi*(pow(cptr->xc[0]+dr,2) - pow(cptr->xc[0],2))*dz;
        //cout << "cvol2 = " << cvol2 << "\n";

        col = cptr->qa[iqal0+1]/dens;
        colamb = 1.0/chi;
        if (col >= 2.0*colamb){ // cell material is mostly cloud
	  //	  rhosqV = (dens/massh)*(dens/massh)*col*cvol2;
	  rhosqV = dens*dens*col*cvol2;
        
        ke = 0.0;
        for (id = 1; id <= nd; id++){
          u = cptr->qa[iqu0+id]/dens;
          ke += u*u;
        }
        ke *= 0.5*dens;
        pre = (g-1.0)*(cptr->qa[iqe] - ke);
        temp = (pre*pscale/(dens*rhoscale))*(avgmass/kboltz);
        if (temp > maxtemp) maxtemp = temp;
        //cout << r << " " << z << " " << temp << "\n"; 
        //fprintf(fptr,"%4i, %.4e, %.4e, %.4e, %.4e\n",ilev,r,z,col,temp);
       	if (temp > 1.5e5){  //hot enough for X-ray emission
          isp = (log10(temp)-logt[1])/dlogt + 2; //index of bin just above temp
          if (isp > ntmax-1) isp=ntmax-1;
          else if (isp < 2) isp=2;
	  // Cycle over spectral grid, computing emission using a linear
          //  gradient (logarithmic gradients fail because lambda often 
          //  has a value of zero)  
	  for (ne = 1; ne < nemax; ne++){
            gr=(lambda[isp][ne]-lambda[isp-1][ne])/
                    (pow(10,logt[isp])-pow(10,logt[isp-1]));
            em=lambda[isp-1][ne] + gr*(temp-pow(10,logt[isp-1]));
            //cout << temp << " " << isp << " " << gr << " " << em << " " << pow(10,em) << " " << rhosqV << " " << rhosqV*pow(10,em) << "\n";
            //exit(EXIT_FAILURE);
	    spec[ne] += rhosqV*em;
            //if (ne == 87){
            //  fprintf(fptr,"%4i, %.4e, %.4e, %.4e, %.4e, %.4e %.4e, %.4e, %.4e, %.4e\n",isp,temp,pow(10,logt[isp-1]),pow(10,logt[isp]),lambda[isp-1][ne],lambda[isp][ne],gr,em,rhosqV*em,spec[ne]);
            //}            
          }
        } //temp limit
        } //cloud material
      }
      cptr = cptr->nxt;
    }
  }
  //cout << "Maximum temp = " << maxtemp << " (K)\n";
  //fclose (fptr);

  //Rescale emission (to get into erg/s/keV) 
  scale = pow(rhoscale/massh,2)*pow(rscale,3);
  //cout << "Scale = " << scale << "\n";
  for (ne = 1; ne < nemax; ne++){
    spec[ne] *= scale;
    //if (spec[ne] > 0.0);
    //else spec[ne] = 0.0;
  }
  //Calculate luminosities in certain bands
  lumtot = 0.0;
  luma = 0.0;
  lumb = 0.0;
  lumc = 0.0;
  for (ne = 1; ne < nemax; ne++){
    lumtot += spec[ne]*de[ne]; // erg/s
    if ((e1[ne] >= 0.3)&&(e2[ne] <= 2.0)){
      luma += spec[ne]*de[ne];
    }
    if ((e1[ne] >= 2.0)&&(e2[ne] <= 5.0)){
      lumb += spec[ne]*de[ne];
    }
    if ((e1[ne] >= 5.0)&&(e2[ne] <= 10.0)){
      lumc += spec[ne]*de[ne];
    }
  }

  // Write out total X-ray spectrum from cloud material
  FILE *fptr; //pointer to file (the "FILE" must be uppercase)
  strcpy(prefx,"m10.0chi1e3adiahikepsres128");
  strcpy(statsfile,prefx);
  strcat(statsfile,"spec.dat");
  fptr = fopen(statsfile,"a+");
  fprintf(fptr,"%s%.4e%s%.4e%s%.4e\n","#X-ray spectrum at time: t = ",t,", t/tcc = ",t/tcc,", t (yr) = ",t*tscale/yr);
  fprintf(fptr,"%s\n","Low E       High E      Central E   Flux (erg/s/keV)");
  for (ne = 1; ne < nemax; ne++){
    fprintf(fptr,"%.4e, %.4e, %.4e, %.4e\n",e1[ne],e2[ne],ecen[ne],spec[ne]);	     
  }
  fclose (fptr);

  strcpy(statsfile,prefx);
  strcat(statsfile,"lum.dat");
  fptr = fopen(statsfile,"a+");
  // Write out X-ray luminosity from cloud material
  fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",t,t/tcc,t*tscale/yr,luma,lumb,lumc,lumtot);
  fclose (fptr);

  //End of code for calculating X-ray spectra

  */  



  /*  
  
  //Calculate Halpha emission. 
  //I could use Case B emissivities for HBeta (erg cm3/s)
  //from Table B.5 of Dopita and Sutherland's book, and then Table B.7 for
  //the Case B Balmer line ratio of Halpha/Hbeta.
  //I also need the ionization fraction of H as a function of temperature
  //to determine the electron number density

  //However, I've also had a look at how DKS does his calcs in Proteus.
  //proteusPhysics.cc has some interesting subroutines - see eg
  //Hemissivity, which returns the emissivity for case A or B, and either
  //Lyman or Balmer series lines. Halpha is the n=3 to n=2 (Balmer) line.
  //The values are based on fits to the values given in Storey & Hummer
  //(1995, MNRAS, 272, 41)
 
  //The DKS work seems the best bet so I'll go with this!

  //First construct a velocity grid and zero the emissivity within it
  velmin = -5.0e7; // cm/s
  velmax = 5.0e7;
  dvel = (velmax-velmin)/float(nvmax);
  for (i = 0; i < nvmax; i++){
    halpha_spec[i] = 0.0;
    velbl[i] = velmin + i*dvel; //velocity of low edge of bin
  }

  // Sum over grids from plmin to plmax  
  //FILE *fptr; //pointer to file (the "FILE" must be uppercase)
  //fptr = fopen("test.dat","a+");
  for (ilev = plmin; ilev <= plmax; ilev++){
    cptr = lev[ilev]->fcell;
    while (cptr != NULL){
      if (cptr->child == NULL){    //if finest cell then...
        dens = cptr->qa[iqd];
        dr = lev[ilev]->dxbar[0];
        dz = lev[ilev]->dxbar[1];
        r = cptr->xc[0] + 0.5*dr;
        z = cptr->xc[1] + 0.5*dz;
        vz = cptr->qa[iqu0+2]/dens;
        cvol2 = gkpi*(pow(cptr->xc[0]+dr,2) - pow(cptr->xc[0],2))*dz;
        //cout << "cvol2 = " << cvol2 << "\n";

        col = cptr->qa[iqal0+1]/dens;
        colamb = 1.0/chi;
        //if (col >= 2.0*colamb){ // cell material is mostly cloud
	  //	  rhosqV = (dens/massh)*(dens/massh)*col*cvol2;
	//  rhosqV = dens*dens*col*cvol2;
        
        ke = 0.0;
        for (id = 1; id <= nd; id++){
          u = cptr->qa[iqu0+id]/dens;
          ke += u*u;
        }
        ke *= 0.5*dens;
        pre = (g-1.0)*(cptr->qa[iqe] - ke);
        te = (pre*pscale/(dens*rhoscale))*(avgmass/kboltz);//K

        //Dopita & Sutherland note that CaseB applies in most HII regions to
        //the Lyman continuum photons (see section 9.1.2). However, I will 
        //assume that CaseA applies for Hbeta. This is consistent with
        //calculating ne assuming CaseA.

        //The fit in DKS's routine is specifically for 500 < Te < 30000,
        //but Figure 2 of Sutherland & Dopita (1993) indicates that a 
        //power-law extrapolation works well up to 1e5 K.
        //log10 emsvty = m log10 t_e + n

        emiss = 0.0;
        if ((te > 1.05e4)&&(te < 1.0e5)){
	  logte = log10(te);
          logemisshalpha_caseA = -20.822 + logte*(-0.956);
          //logemisshalpha_caseB = -20.927 + logte*(-0.884);
          emsvty = pow(10.0,logemisshalpha_caseA);// (erg cm^3/s)
          //cout << te << " " << logte << " " << emsvty << "\n";

          x = 0.7; //H mass fraction is 70%
          n_h = x*dens*rhoscale/massh;  //(cm^-3)
          //compute electron number density (see DKS NeFromRho routine,
          //which assumes caseA recombination)
          n_e = (dens*rhoscale/massh)*0.5*(1.0+x); //raw ne (cm^-3)
          htotcolionrate = 2.5e-10*(1.0+(te/7.8945e4))*
                            sqrt(te)*exp(-(1.5789e5/te));
          htotrecombcaseA = 4.11e-13*pow((1.0e-4*te),-0.698);
          ratio = htotcolionrate/htotrecombcaseA;
          onesubx = 1.0/(1.0+ratio);
          ionfrac = 1.0-onesubx;
	  n_e *= ionfrac;  // correct ne (cm^-3)

          emiss = n_e*n_h*emsvty*cvol2*pow(rscale,3); //Halpha emission (erg/s)

	  //cout << n_e << " " << emsvty << " " << emiss << "\n"; 
          //exit(EXIT_FAILURE);           

          //Now place in the appropriate velocity bin
          vz *= uscale; // cm/s
	  i = int((vz-velmin)/dvel);
          //printf("%.4e, %.4e, %.4e\n",te,n_e,vz,emiss);
          //cout << "i = " << i << "\n"; 
          //exit(EXIT_FAILURE);           
          halpha_spec[i] += emiss;
          
	  } //T filter  
	//} //cloud/ambient filter

      }//finest cell
      cptr = cptr->nxt;
    }
  }//loop over cell pointers

  //write out Halpha emission in (velocity) bins
  FILE *fptr; //pointer to file (the "FILE" must be uppercase)
  strcpy(statsfile,"halphaemiss.dat");
  fptr = fopen(statsfile,"w");
  if (fptr!=NULL){
    totemiss = 0.0;
    for (i = 0; i < nvmax; i++){
      totemiss += halpha_spec[i];
      fprintf(fptr,"%.4e, %.4e, %.4e, %.4e, %.4e\n",velbl[i],velbl[i]+0.5*dvel,velbl[i]+dvel,halpha_spec[i],totemiss);	     
    }
    fclose (fptr);
  }
  
  */

  /* Sam's original code follows:
  // Sum of all variables 
  for (ilev = plmin; ilev <= plmax; ilev++){
    cptr = lev[ilev]->fcell;
    while (cptr != NULL){
      if (!cptr->refined){
	cvol = cptr->vol;
	for (i = 0; i < nq; i++)
	  tot[i] += cvol*cptr->qa[i];
      }
      cptr = cptr->nxt;
    }
  }
  // Get global sum
  for (i = 0; i < nq; i++)
    tot[i] = global_sum(tot[i]);
  // Output sum of all variables 
  if (mype == 0){
    cout.setf(ios::scientific);
    for (i = 0; i < nq; i++)
      cout << "variable " << i << " " << setprecision(6) << tot[i]
	   << '\n';
  }
  */

  return;
}
//end usrsum.c  

//file usrmon.c
/*
+ usrmon - Monitors solution.
     Description :
        This is called  every timestep if mon=true. It  can be used to
        output quantities to stdout or a file at selected intervals.
*/
void usrmon()
{

  /* Automated visualisation of simulations */
  // SVL SILO output
  if (image){
    char filename[SZNAM], temp[SZNAM];
    if ((t >= idumpt)||(cycle == 0)){
      /* Give image file name */
      if (mype == 0){
        usrpath(temp);
        strcpy(filename, getenv(temp));
        strcat(filename,"/");
        strcat(filename, silofile);
        sprintf(temp,".%04d.silo", cycle);
        strcat(filename, temp);
      }
      /* Write image file and zip it (using bzip2)*/
      write_silo(plmax, filename);
      if (mype == 0){
        sprintf(temp, "bzip2 ");
        strcat(temp, filename);
        sys_cmd(temp);
      }
      cycle++;
      idumpt = cycle*idumpdt;
    }
  }

  if (!mon)
    /* No monitor - do nothing */
    return;

}
//end usrmon.c

//file usrdef.c
/*
+ usrdef - set defaults.
     Description :
        Sets the user defaults at start up.
*/
void usrdef()
{
  g = 5.0/3.0;
  usrsetg();
  mon = false;
  fxp = true;
  inten = false;
  dfloor = 0.001;
  pfloor = 0.001;
  avisp = 0.2;
  avise = 0.2;
  viscous = false;
  grav = false;
  gconst = 1.0;
  delpart = 0;

  //JMP added
  initialsetup = true;
  first_radiate = true;
  monstats = false;
  mondsb = false;
  fractal = false;
  cool = false;
  heating = false;
  tempmu = false;
  orbit = false;
  force = false;
  dfloor = 1.0e-8;
  pfloor = 1.0e-8;
  avisp = 0.0;
  avise = 0.0;
  ncellsmap = 4; //default remap radius for CWB
  rhofudge = 1.0;  //used to fudge cloud central density when p1 is small 
  timest = 0.0;
  pict_count = 0;
  cust_al[0] = 0; //RA: custom advected scalars
}
//end usrdef.c

//file usrkeps.c
/*
+ usrkeps - set k-epsilon constants.
     Description :
        Sets the k-epsilon constants.
*/
void usrkeps()
{
  ce1 = 1.4;
  ce2 = 1.94;
  cs = 0.09;
  ct = 0.225;
  cd = 0.09;
  ce = 0.0692;
  ceps = 1.0;
  kmin = 0.0;
  epsmin = 0.0;
  mtl = 1.0;
  viscous = true;
  return;
}
//end usrkeps.c

//file usrsetg.c
/*
+ usrsetg - set gamma constants.
     Description :
        Sets   the  gamma   constants  for   the   non-linear  Riemann
        solver. Must be called whenever gamma is changed.
*/
void usrsetg()
{
  g1 = g - 1.0;
  g2 = g/g1;
  g3 = 0.5*(g + 1.0)/g;
  g4 = 0.5*g1/g;
  g5 = 1.0/g4;
  g6 = 1.0/g;
  g7 = 2.0/g1;
  g8 = g1/(g + 1.0);
  return;
}
//end usrsetg.c

//file usroff.c
/*
+ usroff - set ofsets.
     Description :
        Sets the offsets for the variables at start up.
*/
bool usroff()
{
  iqd = 0;
  iqu0 = 0;
  iqe = 1 + nd;
  /* k- epsilon */
  if (keps){
    iqk = iqe + 1;
    iqeps = iqk + 1;
    iqal0 = iqeps;
  }
  else{
    iqk = -1;
    iqal0 = iqe;
  }
  nq = iqal0 + nal + 1;
  if (inten){
    iqie = nq;
    nq = iqie + 1;
  }
  if (grav){
    /* Gravity */
    iqpot = nq;
    nq++;
  }
  else
    iqpot = 0;
  if (nq > MAXQ){
    cout << "Error : nq = " << nq << " > MAXQ = " << MAXQ <<
      " Change MAXQ in struct.h" << '\n';
    return(false);
  }
  return(true);
}
//end file usroff.c

//file usrvel.c
/*
+ usrvel - Returns velocity.
     Description :
        This returns the id direction  velocity in the cell pointed to
        by cptr. Used for plotting velocity vectors and streamlines.
*/
float usrvel(int id, struct cell *cptr)
{
  return(cptr->qa[iqu0+id]/cptr->qa[iqd]);
}
//end usrvel.c

//file usrbfield.c
/*
+ usrbfield - Returns velocity.
     Description :
        This returns the id direction  velocity in the cell pointed to
        by   cptr.   Used    for   plotting   velocity   vectors   and
        streamlines. This is just a copy of usrvel.
*/
float usrbfield(int id, struct cell *cptr)
{
  return(cptr->qa[iqu0+id]/cptr->qa[iqd]);
}
//end usrbfield.c

//file usrgfield.c
/*
+ usrgfield - Returns gravitational acceleration.
     Description :

        This  returns the id  direction gravitational  acceleration in
        the cell  pointed to by cptr.  Used  for plotting acceleration
        vectors and streamlines.
*/
float usrgfield(int id, struct cell *cptr)
{
  return(cptr->pa[iqu0+id]);
}
//end usrgfield.c

//file usrga.c
/*
+ usrga - Returns gravitational acceleration.
     Description :
        This calculates the id direction gravitational acceleration in
        the cell  pointed to by cptr,  with cells left,  right on left
        and right.  The result is  put in cptr->pa[iqu0+id].
*/
void usrga(int id, struct cell *cptr, struct cell *left,
  struct cell *right)
{
  int count;
  float ga;
  /* Initialise */
  count = 0;
  ga = 0.0;
  if (left != NULL){
    ga += cptr->qa[iqpot] - left->qa[iqpot];
    count++;
  }
  if (right != NULL){
    ga +=  right->qa[nq-1] - cptr->qa[nq-1];
    count++;
  }
  if (count == 0)
    cptr->pa[iqu0+id] = 0.0;
  else{
    ga *= - gconst/((float)(count)*
      (cptr->rjoin[id-1]->xj[id-1] - cptr->ljoin[id-1]->xj[id-1]));
    cptr->pa[iqu0+id] = ga;
  }
}
//end usrga.c

//file usrprompt.c
/*
+ usrprompt - set prompt.
     Description:
        This sets the prompt at start up.
*/
void usrprompt(char prompt[])
{
  strcpy(prompt, "MG_G>");
}
//end usrprompt.c

//file usrquit.c
/*
+ usrquit - clean up user dynamic arrays.
     Description:
        This  deletes any dynamic arrays defined by the user.
*/
void usrquit()
{
  return;
}
//end usrquit.c

//file usrpath.c
/*
+ usrpath - set path.
     Description:
        This sets the path at start up.
*/
void usrpath(char path[])
{
  strcpy(path, "mg_g_data");
}
//end usrpath.c

//file usrdmdt.c
/*
+ usrdmdt - Rate of change of particle mass.
     Description :
        This returns  the rate  of change of  particle mass  with mass
        pmass with the fluid state  in pa[]. It is used for lagrangian
        particles.
*/
float usrdmdt(float pmass, float pa[])
{
  float tv;
  /* Zero for now */
  tv = 0.0;
  return(tv);
}
//end usrdmdt.c

//file usrmkpart.c
/*
+ usrmkpart - Make a lagrangian particle.
     Description :
        This  creates a  particle  at xp,  assigns  it to  a cell  and
        returns a  pointer to the  particle. The particle  position is
        set, but  none of  the other properties.  The particle  is not
        made if xp is not on the processor or in the domain.
*/
struct part *usrmkpart(float xp[])
{
  int id, ilev;
  struct cell *cptr;
  struct part *pptr;
  /* Find cell */
  cptr = find_fcell(ilev, xp);
  if (cptr == NULL) 
    /* Not on processor - do nothing */
    return(NULL);
  else if ((limpe > 1) && (cptr->halo)) 
    /* Halo cell - do nothing */
    return(NULL);
  /* Create particle */
  pptr = new part;
  /* Initial NULL pointers (safety) */
  pptr->prev = NULL;
  pptr->nxt = NULL;
  /* Assign to cell */
  pptr->cptr = cptr;
  if (cptr->pptr == NULL)
    cptr->pptr = pptr;
  else{
    cptr->pptr->prev = pptr;
    pptr->nxt = cptr->pptr;
    cptr->pptr = pptr;
  }
  /* Set particle position */
  for (id = 1; id <= nd; id++)
    pptr->xp[id-1] = xp[id-1];
  return(pptr);
}
//end usrmkpart.c

//file usrprate.c
/*
+ usrprate - Rate of change of lagrangian particle variables.
     Description :
        This  should  return  the  time derivatives  of  the  particle
        variables. It is given
          pprt    -- particle  pointer
          pa[]    -- fluid state at particle position (pptr->xp)
        It should return
          dvp[]   -- particle acceleration
          den     -- rate of change of particle energy
          dm      -- rate of change of particle mass
*/
void usrprate(struct part *pptr, float dt, float pa[], float dvp[],
 float &den, float &dm)
{
  float ql, qr;
  int id;
  struct cell *cptr;
  struct mcell *mcptr, *left, *right;
  /* Zero for now */
  for (id = 0; id < nd; id++)
    dvp[id] = 0.0;
  den = 0.0;
  dm = 0.0;
  if (grav){
    /* Calculate left/right gravitational potentials */
    cptr = pptr->cptr;
    mcptr = cptr->mcptr;
    for (id = 1; id <= nd; id++){
      /* Left potential */
      left = mcptr->lcell[id-1];
      if (left != NULL)
	/* left cell exists */
	ql = left->qa;
      else
	/* Left external symmetry boundary */
	ql = mcptr->qa;
      /* Right potential */
      right = mcptr->rcell[id-1];
      if (right != NULL)
	/* Right interior cell */
	qr = right->qa;
      else
	/* Right external symmetry boundary*/
	qr = mcptr->qa;
    /* Apply gravitational acceleration */
      dvp[id-1] -= 0.5*gconst*(qr - ql)/
	(cptr->rjoin[id-1]->xj[id-1] - cptr->ljoin[id-1]->xj[id-1]);
    }
  } 
  return;
}
//end usrprate.c

//file usrparpop.c
/*
+ usrpartpop - Update lagrangian particle population.
     Description :
        This  should  be  used   to  update  the  lagrangian  particle
        population. tcur is the  current model time.
*/
void usrpartpop(int ilev, float tcur)
{
  /* Jet test */
  if (strcmp(cas, "DONALDSON") == 0){
    int id;
    float xp[MAXD];
    static float tcreate = 0.0;
    struct cell *cptr;
    struct part *pptr;
    if (tcur - tcreate > 20.0*dt0){
      tcreate = tcur;
      xp[0] = 0.1;
      xp[1] = 0.01;
      pptr = usrmkpart(xp);
      if (pptr != NULL){
	cptr = pptr->cptr;
	pptr->vp[0] = cptr->qa[iqu0+1]/cptr->qa[iqd];
	pptr->vp[1] = cptr->qa[iqu0+2]/cptr->qa[iqd];
	pptr->t = tcur;
	pptr->mass = 1.0;
	pptr->en = 2.0;
	pptr->props[0] = 3.0;
	pptr->props[1] = 4.0;
      }
      xp[0] = 0.2;
      xp[1] = 0.01;
      pptr = usrmkpart(xp);
      if (pptr != NULL){
	cptr = pptr->cptr;
	pptr->vp[0] = cptr->qa[iqu0+1]/cptr->qa[iqd];
	pptr->vp[1] = cptr->qa[iqu0+2]/cptr->qa[iqd];
	pptr->t = tcur;
	pptr->mass = 1.0;
	pptr->en = 2.0;
	pptr->props[0] = 3.0;
	pptr->props[1] = 4.0;
      }
      xp[0] = 0.3;
      xp[1] = 0.01;
      pptr = usrmkpart(xp);
      if (pptr != NULL){
	cptr = pptr->cptr;
	pptr->vp[0] = cptr->qa[iqu0+1]/cptr->qa[iqd];
	pptr->vp[1] = cptr->qa[iqu0+2]/cptr->qa[iqd];
	pptr->t = tcur;
	pptr->mass = 1.0;
	pptr->en = 2.0;
	pptr->props[0] = 3.0;
	pptr->props[1] = 4.0;
      }
      xp[0] = 0.4;
      xp[1] = 0.01;
      pptr = usrmkpart(xp);
      if (pptr != NULL){
	cptr = pptr->cptr;
	pptr->vp[0] = cptr->qa[iqu0+1]/cptr->qa[iqd];
	pptr->vp[1] = cptr->qa[iqu0+2]/cptr->qa[iqd];
	pptr->t = tcur;
	pptr->mass = 1.0;
	pptr->en = 2.0;
	pptr->props[0] = 3.0;
	pptr->props[1] = 4.0;
      }
    }
  }
  return;
}
//end usrpartpop.c

//file usrpartps.c
/* 
+ usrpartps - construct plot scalar for particle.
     Description :
        s returns  the value  of a particle  property in  the particle
        pointed to  by pptr. The  choice of property is  determined by
        pitem true if the plot pitem is found, false otherwise.
*/
bool usrpartps(char pitem[], struct part *pptr, float &s)
{
  bool found;
  if (strcmp(item, "PPARTMASS") == 0){
    s = pptr->mass;
    found = true;
  }
  else if (strcmp(item, "PPARTRHO") == 0){
    s = pptr->props[0];
    found = true;
  }
  else if (strcmp(item, "PPARTEN") == 0){
    s = pptr->en;
    found = true;
  }
  else if (strcmp(item, "PPARTU1") == 0){
    s = pptr->vp[0];
    found = true;
  }
  else if (strcmp(item, "PPARTU2") == 0){
    if (nd < 2)
      found = false;
    else{
      s = pptr->vp[1];
      found = true;
    }
  }
  else if (strcmp(item, "PPARTU3") == 0){
    if (nd < 3)
      found = false;
    else{
      s = pptr->vp[2];
      found = true;
    }
  }
  else if (strcmp(item, "PPARTUT") == 0){
    int id;
    float tv1;
    s = 0.0;
    for (id = 1; id <= nd; id++){
      tv1 = pptr->vp[id-1];
      s += tv1*tv1;
    }
    s = sqrt(s);
    found = true;
  }
  return(found);
}
//end usrpartps.c

//file usravgmass.c
/*
+ usravgmass - return average mass from WIND simulation
*/
float usravgmass(float pa[])
{
  float x,y,z,avgm;
  //there are nstars advected scalars for the wind colours, which precede the
  //3 advected scalars for the mass fractions  
  x = pa[iqal0+nstars+1]; //H mass fraction
  y = pa[iqal0+nstars+2]; //He mass fraction
  z = pa[iqal0+nstars+3]; //metals mass fraction
  avgm = massh/(2.0*x + 0.75*y + 0.5*z); //assumes complete ionization!
  return(avgm);
}
//end usravgmass.c


//file radiate.c
/*
+ Calculate the energy loss in the cell due to radiative cooling
    Description : The method used in this routine is based upon 
    Strickland & Blondin (1995, ApJ, 449, 727). The technique of operator 
    splitting assumes that the change in pressure during a timestep can be 
    decomposed into a compressional term (handled by the hydrodynamics) and 
    a radiative term (handled here). The pressure change is rewritten as a 
    temperature change (for notes on this see my notebook). This also gives 
    the change in internal energy (dei = delta(temp)*density). The cell 
    temperature is not allowed to fall below twind.
    There is no check for unresolved interfaces between hot and cool gas - 
    VH-1 restricts the cooling in such situations (which I think stops 
    pressure pulses being eaten away as they move across the grid due to the 
    creation of cells with intermediate temperatures, e.g. T = 1.0e6 K).
    This routine returns eloss which is the rate of energy loss per unit time
    per unit volume
*/
float radiate(float pa[], float dt)
{
  bool limit;
  int nt, ntmax, unit, n, id;
  float x1, y1, z1, x2, y2, z2, ke;
  float nden, const_1, const_2, pre, rho, col, col2, tmpold1, tmpold2, 
    lambda, eloss, eloss2, elosslim, limit_frac;  
  /* cout << "radiate\n"; 
     exit(EXIT_FAILURE);*/

  ntmax = 101;
  limit = true;
  limit_frac = 0.3; // limit the fractional energy loss per timestep

  //if (dt == 0.0) return(0.0);// dt=0 if plotting e.g. sa1
  if (first_radiate){
    cout << "First radiate\n";
    for (nt = 1; nt <= ntmax; nt++){
      lamb_tot1[nt] = 0.0;
      lamb_tot2[nt] = 0.0;
    }
    /* read in emissivity data stored such that eloss = -(rho/mh)**2 Lambda */
    /* open file - for general input/output use fstream,
                   for input to the program only use ifstream,
                   for output from the program only use ofstream */   
    // cout << "Opening " << coolcurve1 << '\n'; 
    ifstream file(coolcurve1);  
    if (!file) {
      cerr << "Failed to open file\n";
      exit(EXIT_FAILURE);
    } 
    file.seekg(0); /* goto start of file */
    for (nt = 1; nt <= ntmax; nt++){
      file >> logt[nt] >> lamb_tot1[nt];
      /* cout << nt << " " << logt[nt] << " " << lamb_tot1[nt] << '\n'; */
    }
    file.close();
    //if (nstars > 1){
    if (strcmp(cas, "CWB") == 0){
      ifstream file2(coolcurve2);
      if (!file2) {
        cerr << "Failed to open file\n";
        exit(EXIT_FAILURE);
      }
      file2.seekg(0); /* goto start of file */
      for (nt = 1; nt <= ntmax; nt++){
        file2 >> logt[nt] >> lamb_tot2[nt];
        /* cout << nt << " " << logt[nt] << " " << lamb_tot2[nt] << '\n'; */
      }
      file2.close();
    }
    dlogt = logt[2]-logt[1];
    first_radiate = false;
  }
 
  /* This bit of code is redundant since avgmass1/2 are calculated in usrqis
  //Calculate the avgmass for the current hydro cell
  if (strcmp(cas, "CWB") == 0){//obtain the mass fractions from the stellar data table (since there will be no evolution of these during the simulation)  
    x1 = massfracx[0];
    y1 = massfracy[0];
    z1 = massfracz[0];
    x2 = massfracx[1];
    y2 = massfracy[1];
    z2 = massfracz[1];
    avgmass1 = massh/(2.0*x1 + 0.75*y1 + 0.5*z1);//assumes complete ionization!
    avgmass2 = massh/(2.0*x2 + 0.75*y2 + 0.5*z2);//assumes complete ionization!
  }
  */

  if (strcmp(cas, "WIND") == 0){//obtain the mass fractions from the current cell (since the cell could be in MS wind while the star has since evolved to a WR stage)
    x1 = pa[iqal0+nstars+1];
    y1 = pa[iqal0+nstars+2];
    z1 = pa[iqal0+nstars+3];
    avgmass1 = massh/(2.0*x1 + 0.75*y1 + 0.5*z1);//assumes complete ionization!
  }
  else if ((strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "MANYCLOUD") == 0)||(strcmp(cas, "LEAKYSHELL") == 0)){
    avgmass1 = avgmass;
  }
  else if (strcmp(cas, "CWB") != 0){
    cout << "Not coded for a different CASE in radiate.c\n";
    cout << "Aborting!\n";
    exit(EXIT_FAILURE);
  }
  
  const_1 = avgmass1/kboltz;
  const_2 = avgmass2/kboltz;

  //Calculate density, pressure, colour and temperature of material in cell
  rho = pa[iqd];
  pre = pa[iqe];
  if (inten){
    ke = 0.0;
    for (id = 1; id <= nd; id++){
      ke += pa[iqu0+id]*pa[iqu0+id];
    }
    ke *= 0.5*rho;
    //if (pre < 0.03*ke) pre = (g-1.0)*pa[iqie]; //use internal energy instead
    if (pre < 0.03*ke) pre = pa[iqie]*powf(rho,g); //use entropy (=p/rho^g) instead
  }
  rho *= rhoscale;
  pre *= pscale;
  col = pa[iqal0+1];

  //cout << "rho = " << rho << "\n";
  //cout << "col = " << col << "\n";
  //exit(EXIT_FAILURE);

  /*
  if (strcmp(cas, "WBB") == 0){  
    nden = rho/(1.3*massh);
    tmin = max(1.0e2,min(1.0e4/nden,1.0e4));
  }
  if (strcmp(cas, "CWB") == 0){
    if (clump){
      //if (col2 > 0.5) tmin = twind/ncl; //inside clump
      //else tmin = twind;
      tmin = col2*(twind/ncl) + (1.0-col2)*twind;
    }
    else tmin = twind;
  } 
  */
  //tmin = twind; //messing

  // Calculate the cooling
  // Currently, the WIND case only uses 1 avgmass and 1 cooling curve.
  // In future, we may wish to select different cooling curves (eg solar,
  // WN, WC) based on the cell's mass fractions of H, He and metals)

  // First calculate the temperature of the cell. If T is less than twind 
  // set the energy loss to 0.0 (dei=0.0) and jump out. Else calculate a 
  // first estimate of the cooling rate. By setting Tmin=1.1*teff, the 
  // unshocked wind should not cool. Tmin is also used to exponentially 
  // smooth the cooling function near the drop to zero emissivity. 

  if (strcmp(cas, "WIND") == 0){
    tmpold1 = const_1*pre/rho;     // use avgmass1 to calculate the temperature
    if (tmpold1 <= 1.02*tmin) {
       eloss = 0.0;
    }
    else {  //use lamb_tot1 cooling curve
      coolc(tmpold1,tmin,logt,dlogt,ntmax,lamb_tot1,lambda);
      eloss = -(rho*rho/massh/massh)*lambda;
      // Limit cooling (we use MAX because these terms are negative)
      if (limit) eloss = max(eloss,
      	  (limit_frac*1.5*(tmin-tmpold1)*rho*kboltz/avgmass1/dt/tscale));
    }
  }
  else if (strcmp(cas, "CWB") == 0){ 
    // Assume that the cell contains ONLY one wind if the difference
    // between it and the colour of the winds is less than 5%

    if (col > 0.95) { //assume cell is pure wind 1
      tmpold1 = const_1*pre/rho;
      //cout << "tmpold1 = " << tmpold1 << "\n";
      //cout << "tmin = " << tmin << "\n";
      //cout << "dt = " << dt << "\n";
      //exit(EXIT_FAILURE);
      if (tmpold1 <= 1.02*tmin) {
        eloss = 0.0;
      }
      else {
        coolc(tmpold1,tmin,logt,dlogt,ntmax,lamb_tot1,lambda);
        eloss = -(rho*rho/massh/massh)*lambda;
	// Limit cooling (we use MAX because these terms are negative) 
        if (limit) eloss = max(eloss,
      	  (limit_frac*1.5*(tmin-tmpold1)*rho*kboltz/avgmass1/dt/tscale));
        //cout << "tmpold1 = " << tmpold1 << "\n";
        //cout << "eloss = " << eloss << "\n";
        //exit(EXIT_FAILURE);
      }
    }
    else if (col < 0.05) { //assume cell is pure wind 2
      tmpold2 = const_2*pre/rho;
      if (tmpold2 <= 1.02*tmin) {
         eloss = 0.0;
      }
      else {
        coolc(tmpold2,tmin,logt,dlogt,ntmax,lamb_tot2,lambda);
        eloss = -(rho*rho/massh/massh)*lambda;
        if (limit) eloss = max(eloss,
      	  (limit_frac*1.5*(tmin-tmpold2)*rho*kboltz/avgmass2/dt/tscale));
      }
    }
    else {
      /* mixture of winds - assume in pressure equilibrium with the total
         mass of each cell being divided between the two winds according to 
         their fractional colour. This means that each wind in the cell will 
         have a different temperature. Calculate cooling by treating each 
         wind separately. Cells which are mixed ALWAYS cool unless 
         T <1.1*twind */
      tmpold1 = const_1*pre/rho;
      tmpold2 = const_2*pre/rho;
      if (tmpold1 <= 1.02*tmin) {
        eloss = 0.0;
      }
      else {
        coolc(tmpold1,tmin,logt,dlogt,ntmax,lamb_tot1,lambda);
        eloss = -col*(rho*rho/massh/massh)*lambda;
        if (limit) eloss = max(eloss,
      	  (col*limit_frac*1.5*(tmin-tmpold1)*rho*kboltz/avgmass1/dt/tscale));
      }
      eloss2 = 0.0;
      if (tmpold2 > 1.02*tmin) { /* add cooling of second wind */
        coolc(tmpold2,tmin,logt,dlogt,ntmax,lamb_tot2,lambda);
        eloss2 = -(1.0-col)*(rho*rho/massh/massh)*lambda;
        if (limit) eloss2 = max(eloss2,
	  ((1.0-col)*limit_frac*1.5*(tmin-tmpold2)*rho*kboltz/avgmass2/dt/tscale));
      }
      eloss += eloss2;
    }
  }//end of CASEs

  // Make eloss dimensionless and return
  eloss = eloss*rscale/(pscale*uscale);
  return(eloss);
}


//file radiate_heatcool.c
/*
+ Calculate the energy CHANGE in the cell due to radiative cooling AND heating
    Description : The method used in this routine is based upon a 1st order
    explicit integration with adaptive time-step sub-sampling. 

    This routine returns edot which is the rate of energy loss/gain per 
    unit time per unit volume.
    
    Uses qa to calculate T, and pa to calculate rho,
    since the latter is updated at the half time step, while the cooling
    MUST be based on T at the beginning of the time step.

    Tests: See investigateheatcool.f for additional notes
    Run1: 1e-3 tolerance to jump out produces 415 steps, temp_new = 4.0341
          2e-3 tolerance takes 90 steps: temp_new = 4.0741
    Run2: dt = 2e15, temp_new = 1.3974e4 in 47 steps
    Run3: temp_new = 4.5170e7 in 45 steps
    Run4: temp_new = 2.636e4 in 46 steps
    Run5: temp_new = 2.724e4 in 29 steps
    Run6: temp_new = 4.4835e7 in 5 steps
    Conclusion: Everything seems to be working fine!
    JMP: 20/01/11

  Modifications:
    JMP: 27/01/11 - added temperature dependent mu capability. However,
                    tests shows that it is best to stick with the initial
                    value of mu rather than adjusting mu as T changes (see
                    "Multiple Colliding Winds" book p80-81)
    JMP: 31/01/11 - added xc[] passed variable so that cooling/heating can
                    be position dependent (also useful for debugging).
                    Fixed bug: dtint was compared to dt instead of dt*tscale
    JMP: 01/02/11 - added temp_init and temp_new for debugging
    JMP: 03/02/11 - modified temp calculation for temp dependent gamma
*/

float radiate_heatcool(float qa[], float pa[], float xc[], float dt, float &temp_init, float &temp_new)
{

 /* cout << "radiate_heatcool\n"; 
     exit(EXIT_FAILURE);*/
  bool cool,limit_dtemp,initlambdaju;
  char statsfile[60];
  int ntmax,n,i,id,j,nover;
  double x1, y1, z1, x2, y2, z2, ke;
  double nden, pre, rho, col, col2, temp, temp2, dtemp,
   lambdaju;
  //float temp_new, temp_init;
  double t_min,logtmin,dtstep,dtint,dts;
  double coolterm,heatterm,netheatcool;  
  double mu,mu2,mu_init,avgmass,gam,gam2,gam_init;
  double t_cool,t_cool_init,t_cool_new,edot_to_dtemp;
  double edot,dtemp_frac_cool,dtemp_frac_heat;
  double einit,efin,pdcell,iedcell;
  double xdist,ydist,zdist,r,rs;
  float edotf;
  FILE *fptr; //pointer to file (the "FILE" must be uppercase)

  limit_dtemp = false; //limit the fractional energy loss/gain per step?
  dtemp_frac_cool = 0.3; //maximum fractional energy loss permitted
  dtemp_frac_heat = 0.3; //maximum fractional energy gain permitted

  edotf = 0.0;

  //JMP idealized cooling curve
  ntmax = 10;
  if (first_radiate){
    cout << "First radiate_heatcool\n";

    te[1] = 1.0;
    te[2] = 10.0;
    te[3] = 80.0;
    te[4] = 150.0;
    te[5] = 1.0e3;
    te[6] = 8.0e3;
    te[7] = 1.0e4;
    te[8] = 1.8e4;
    te[9] = 1.0e5;
    te[10] = 3.216e7;
    lambda[1] = 5.12e-31;
    lambda[2] = 5.12e-29;
    lambda[3] = 3.178e-34;
    lambda[4] = 1.060e-28;
    lambda[5] = 1.072e-27;
    lambda[6] = 2.874e-69;
    lambda[7] = 1.989e-45;
    lambda[8] = 4.344e-35;
    lambda[9] = 8.649e-19;
    lambda[10] = 2.490e-27;
    alpha[1] = 2.5;
    alpha[2] = 0.5;
    alpha[3] = 3.236;
    alpha[4] = 0.698;
    alpha[5] = 0.363;
    alpha[6] = 11.014;
    alpha[7] = 5.054;
    alpha[8] = 2.618;
    alpha[9] = -0.639;
    alpha[10] = 0.5;

    //pdmin = 8.733018e7;
    //pdmax = 2.047e17;
    pdmin = pd[0];
    pdmax = pd[npd-1];

    first_radiate = false;
  }//end of first_radiate

  t_min = te[1];    // K
  t_max = 1.0e9;    // K
  //logtmax = log10(t_max);
  logtmin = log10(t_min);

  rho = pa[iqd]*rhoscale;
  //colour of clump material
  col = pa[iqal0+nstars+4]; // X,Y,Z mass fractions are nstars+1->3

  //distance of cell from grid origin
  xdist = xc[0];
  ydist = 0.0;
  zdist = 0.0;
  if (nd > 1) ydist = xc[1];
  if (nd > 2) zdist = xc[2];
  r = sqrt(xdist*xdist + ydist*ydist + zdist*zdist);
  rs = r*rscale;

  //calculation of the temperature depends on whether gamma and mu are
  //temperature dependent. Note that qa MUST be used to calculate the
  //temperature.
  if (tempgamma){      //temperature dependent gamma and mu
    iedcell = usriedqa(qa)*pscale/rhoscale;//calculate temp from ie/d from qa
    temp = usriedtmp(iedcell);
    mu = usriedmu(iedcell);
    avgmass = mu*massh;
    gam = usriedg(iedcell);
  }
  else if (tempmu){    //temperature dependent mu only
    pdcell = usrteqa(qa)*pscale/rhoscale; //calculate temp from p/d from qa
    pdcell = max(pdcell,1.001*pdmin);
    pdcell = min(pdcell,0.999*pdmax); 
    /*
    if (pdcell < pdmin){
      cout << "pdcell < pdmin\n";
      cout << "pdcell = " << usrteqa(qa)*pscale/rhoscale << "\n";
      cout << "pdmin  = " << pdmin << "\n";
      cout << "Aborting from radiate_heatcool.c!\n";
      exit(EXIT_FAILURE);
    }
    else if (pdcell > pdmax){
      cout << "pdcell > pdmax\n";
      cout << "pdcell = " << usrteqa(qa)*pscale/rhoscale << "\n";
      cout << "pdmax  = " << pdmax << "\n";
      cout << "Aborting from radiate_heatcool.c!\n";
      exit(EXIT_FAILURE);
    }
    */
    temp = usrpdtmp(pdcell);
    mu = usrpdmu(pdcell);
    avgmass = mu*massh;
    gam = g;
    //pre = rho*temp*(kboltz/avgmass); //cgs
    //cout << "Initial T = " << temp << "; initial mu = " << mu <<"\n";
  }
  else{                //mu and gamma are temperature independent
  //obtain the mass fractions from the current cell (since the cell could be in MS wind while the star has since evolved to a WR stage)
    x1 = pa[iqal0+nstars+1];
    y1 = pa[iqal0+nstars+2];
    z1 = pa[iqal0+nstars+3];
    avgmass = massh/(2.0*x1 + 0.75*y1 + 0.5*z1);//assumes complete ionization!
    mu = avgmass/massh;
    pdcell = usrteqa(qa)*pscale/rhoscale; //calculate temp from p/d from qa
    temp = (avgmass/kboltz)*pdcell;
    gam = g;
    //temp = 9007.0;
    //mu = 0.6956;
    //cout << "Initial T = " << temp << "; initial mu = " << mu <<"\n";
  }

  temp_init = temp;
  mu_init = mu;
  gam_init = gam;

  //rho = 1.0e-24;
  //temp = 2.5e7;
  //dt = 2.0e17;

  //cout << "rho = " << rho << "; temp = " << temp << "; dt = " << dt << "\n";
  if (temp < 1.01*tmin) return(edotf);
  if (temp > 0.99*t_max) return(edotf);

  nden = rho/massh; //cgs

  initlambdaju = true;
  lambdaju = 0.0;
  i = ntmax + 1;
  while (initlambdaju){
    i--;
    if (temp >= te[i]){
      lambdaju = lambda[i]*pow(temp,alpha[i]);
      initlambdaju = false;
    }
  }
  if ((lambdaju == 0.0)||(lambdaju < 0.0)){
    cout << "Invalid initial value of lambdaju. Aborting!\n";
    exit(EXIT_FAILURE); 
  }
  if ((i < 1)||(i > 10)){
    cout << "radiate_heatcool: invalid initial value of i. Aborting!\n";
    exit(EXIT_FAILURE); 
  }

  //Specify filename for outputting info
  //strcpy(statsfile,"newtemp");

  //Calculate single point cooling/heating time (negative value indicates
  // heating)

  coolterm = nden*nden*lambdaju;
  heatterm = nden*gammaheat;
  netheatcool = coolterm - heatterm;
  t_cool = nden*kboltz*temp/((g-1.0)*mu*netheatcool);
  t_cool_init = t_cool;

  if (t_cool >= 0.0){//if t_cool > 0.0 gas is cooling
    cool = 1;    //true
  }
  else cool = 0; //false

  dtint = 0.0;
  dtstep = 0.1*mabs(t_cool); //initialize safe step

  j = 0;

  //cout << "dt = " << dt << "; dtstep = " << dtstep << "\n";
  dts = dt*tscale;

  while (dtint < dts){
    j++;
    //cout << "j = " << j << "\n";
    dtstep = min(1.5*dtstep,0.2*mabs(t_cool)); // safe step
    if ((dtint + dtstep) > dts) dtstep = dts - dtint;
    nover = 0;
    loop_again:
    temp_new = temp*(1.0 - dtstep/t_cool); // 1st order explicit
    //if (t_cool_init < 0.0){ //ie initial heating
    //  cout << "temp_init = " << temp_init << "; temp_new = " << temp_new << "\n";
    //  cout << "dtstep = " << dtstep << "; t_cool = " << t_cool << "\n";
    //}
    // Update index for position on cooling curve
    if (cool){
      if (temp_new < te[i]){
	//if (i == 1){   // do not allow cooling below Tmin
	if (i <= 1){   // do not allow cooling below Tmin
	  temp_new = te[1];
	  dtint = dts;  //ie jump out
        }
        else i--;
      }
    }
    else{ 
      //if (i != ntmax){
      if (i < ntmax){
        //             write(*,*) temp_new,i,te(i+1)
        if (temp_new > te[i+1]) i++;   
      }
    }
    //check i is within limits
    if (i < 1){
      cout << "radiate_heatcool - i < 1\n";
      cout << "temp_new = " << temp_new << "\n";
      cout << "Aborting!\n";
      exit(EXIT_FAILURE);
    }
    else if (i > 10){
      cout << "radiate_heatcool - i > 10\n";
      cout << "temp_new = " << temp_new << "\n";
      cout << "Aborting!\n";
      exit(EXIT_FAILURE);
    }

    //First calculate the new single point cooling rate and t_cool to check
    // whether an overshoot has occurred
    temp2 = temp_new;
    mu2 = mu_init;
    gam2 = gam_init;
    //DO NOT uncomment the line below - i.e. do not update mu
    //if (tempmu) mu2 = usrtmpmu(temp2); //update mu using the new temp
    if (tempgamma) gam2 = usrtmpg(temp2); //update gamma
    //cout << "T = " << temp2 << "; mu = " << mu2 <<"\n";
    lambdaju = lambda[i]*pow(temp2,alpha[i]);
    coolterm = nden*nden*lambdaju;
    netheatcool = coolterm - heatterm;
    //t_cool_new = nden*kboltz*temp2/((g-1.0)*mu*netheatcool);
    t_cool_new = nden*kboltz*temp2/((gam2-1.0)*mu2*netheatcool);
    if (((cool)&&(t_cool_new < 0.0))||
        ((!cool)&&(t_cool_new > 0.0))){ //overshoot has occured
      nover++;
      dtstep /= 2.0;
      //cout << "Overshoot, nover = " << nover << " : limiting dtstep = " << dtstep <<"\n";
      //jump out if close to equilibrium
      if ((nover > 10)&&(dtstep < 1.0e-2*dts)){
	dtint = dts;
        goto end_of_loop;
      }

      if (nover > 100){
        cout << "Too many overshoots in radiate_heatcool. Aborting!\n";
        exit(EXIT_FAILURE);
      }
  
      goto loop_again;
    }
    t_cool = t_cool_new; //safe to update t_cool
    // Once the temp is changing very slowly (ie is near equilibrium) jump out
    if ((mabs(t_cool) < dts)&&(mabs(temp_new-temp)/temp < 2.0e-3)){
    //if ((abs(t_cool) < dts)&&(j > 10)){
      //the t_cool < dts condition prevents jump out when gas is very
      //low density and t_cool >> dts (which would prevent any cooling of such
      //gas)
      dtint = dts;
    }
    //else if ((cool)&&(t_cool > 10.0*dts)&&(abs(temp_init-temp_new)/temp_init > 0.5)){
    else if ((cool)&&(t_cool > 10.0*dts)&&(j > 10)){
      //t_cool becomes very large as cooling gas approaches equilibrium.
      //If the gas has already radiated more than 50% of its energy (so has
      //already experienced rapid cooling in this step) then jump out.
      dtint = dts;
    }
    //else if ((!cool)&&(abs(t_cool) > 10.0*dts)&&(abs(temp_new-temp_init)/temp_init > 1.5)){
    else if ((!cool)&&(mabs(t_cool) > 10.0*dts)&&(j > 10)){
      //abs(t_cool) becomes very large as heating gas approaches equilibrium.
      //If the gas has already gained more than 50% in energy (so has
      //already experienced rapid heating in this step) then jump out.
      dtint = dts;
    }
    //Jump out if cell temperature approaches lower or upper boundaries
    if ((temp_new < 1.01*tmin)&&(t_cool > 0.0)){
      temp_new = 1.01*tmin;
      dtint = dts;
    }
    if ((temp_new > 0.99*t_max)&&(t_cool < 0.0)){
      temp_new = 0.99*t_max;
      dtint = dts;
    }
    //Jump out with an error if the number of steps is too high
    if (j > 200){
      cout << "Too many loops in radiate_heatcool. Aborting!\n";
      exit(EXIT_FAILURE);
    }
    // If everything else is OK update
    dtint = dtint + dtstep;
  end_of_loop:
    temp = temp_new;
    mu = mu2;
    gam = gam2;
    //write info to file
    //fptr = fopen(statsfile,"a+");
    //if (fptr!=NULL){
    //  fprintf(fptr,"%i, %i, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e, %.4e\n",
    //      j,i,t_cool,dt,dtint,coolterm,heatterm,netheatcool,temp);
    //  fclose (fptr);
    //}
    //cout << j << " " << i << " " << t_cool_init << " " << t_cool << " " << dts << " " << dtint << " " << temp_init << " " << temp << "\n";
  }//do while

  //Given the old temp and the new temp and the timestep, calculate edot for
  //usrsaf. Use the average gamma and mu
  gam = 0.5*(gam_init+gam2);
  mu  = 0.5*(mu_init+mu2);
  avgmass = mu*massh; 
  edot_to_dtemp = dts*(gam-1.0)*avgmass/(rho*kboltz);
  dtemp = temp_new - temp_init;  //dtemp is negative if the gas is cooling
  //if (tempmu){  //THIS code is not needed if mu is fixed during the cooling
  //  einit = rho*(kboltz/(mu_init*massh))*temp_init/(g-1.0);
  //  efin = rho*(kboltz/(mu*massh))*temp_new/(g-1.0);
  //  edot = (efin - einit)/(dt*tscale);  //cgs
  //  cout << "Einit = " << einit << "; Efin = " << efin << "; edot = " << edot << "\n";
  //  exit(EXIT_FAILURE);
  //}
  //else 

  //Limit cooling/heating rate if desired
  if (limit_dtemp){
    if ((cool)&&(mabs(dtemp)>dtemp_frac_cool*temp_init)) 
       dtemp = -dtemp_frac_cool*temp_init;
    else if ((!cool)&&(dtemp>dtemp_frac_heat*temp_init)) 
       dtemp = dtemp_frac_heat*temp_init;
  }

  edot = dtemp/edot_to_dtemp;//cgs (erg/s/per unit vol) (Edot/vol =pdot/(g-1.0))

  // Make edot dimensionless and return (cooling rate per unit volume)
  edotf = edot*rscale/(pscale*uscale);

  //Investigate heating/cooling
  /*
  if (rs < 0.5*rclump*pc){
    cout << "temp_init = " << temp_init << "\n";
    cout << "temp_new = " << temp_new << "\n";
    cout << "dtemp = " << dtemp << "\n";
    cout << "edot = " << edot << "\n";
    cout << "edotf = " << edotf << "\n";
    cout << "t_cool_init = " << t_cool_init << "\n";
    cout << "t_cool = " << t_cool << "\n";
    cout << "dt = " << dt << "\n";
    cout << "dts = " << dts << "\n";
    cout << "Pre = " << pre/pscale << "\n";
    exit(EXIT_FAILURE);
  }
  */

  //cout << "edot = " << edotf << "\n";
  //exit(EXIT_FAILURE);

  // Finally check again for infinities or NaNs
  if ( isinf( edotf )){
    // value is infinity
    cout << "radiate_heatcool.c: edotf is infinity\n";
    exit(EXIT_FAILURE);
    edotf = 0.0;
    return(edotf);
  }

  if ( isnan( edotf )){
    // value is not a number
    cout << "radiate_heatcool.c: edotf is a NAN\n";
    cout << "dtemp = " << dtemp << "\n";
    cout << "temp_init = " << temp_init << "\n";
    cout << "temp_new = " << temp_new << "\n";
    cout << "t_cool_init = " << t_cool_init << "\n";
    exit(EXIT_FAILURE);
    edotf = 0.0;
    return(edotf);
  }

  return(edotf);
}
//end of file radiate_heatcool.c



//file coolc.c
/*
+ Calculate the cooling coefficient from the supplied lookup table
    Description : New cooling curve subroutine based on the use of lookup 
    tables rather than the old smooth parameterised functions. The data for 
    the look up table is read directly from disk as a simple ascii file of
    log10(T [K]) vs lambda (in units of erg cm**3 s**-1). This is 
    considerably easier than trying to create parametrised cooling
    curves as done previously as a sum of Gaussian and exponential terms
*/

void coolc(float t, float tmin, float logt[], float dlogt, int ntmax, 
  float lamb_tot[], float &lambda)

{
  int it, nt;
  float lt, dte;

  lambda = 0.0;  
  lt = log10(t); /* log10 of input temperature */
  /* calculate temperature index it from lt. Add a small number just to
     make sure we end up with the correct temperature index if we happen to
     fall right on a bin boundary */
  dte = 1.0e-6 + (lt-logt[1])/dlogt;
  if (dte >= 0.0) it = int(dte)+1;
  else it = int(dte);
  /* Assume lambda is zero if the temperature is outside the range specified by
     the lookup table (indices 1 to nt). On the lower end we assume the 
     emissivity is the lowest available emissivity - i.e. lamb_tot(1)
     The emissivity can either be the same value over an entire temperature
     bin, or a linear interpolation between the emissivities at adjacent
     temperatures. It is currently setup for a linear interpolation */
  if (it <= 1) lambda = lamb_tot[1];
  else if (it >= ntmax) lambda = lamb_tot[ntmax];
  else {
     lambda = lamb_tot[it] + (lt-logt[it]) * 
       (lamb_tot[it+1] - lamb_tot[it])/ dlogt;
  }
  /* Smooth out drop to zero emissivity at tmin */
  lambda = (1.0 - (tmin*tmin/t/t)) * lambda;
  lambda = max(0.0, lambda);
  return;
}


//file radiate_heatcool.c
/*
+ Calculate the energy CHANGE in the cell due to radiative cooling AND heating
    Description : The method used in this routine is based upon a 4th order
    Runge Kutta integration with adaptive time-step sub-sampling. 
    See other_programs/Cooling/lowTcool/numerical_cooling.cpp.

    This routine returns edot which is the rate of energy loss/gain per 
    unit time per unit volume.
    
    Uses qa to calculate T, and pa to calculate rho,
    since the latter is updated at the half time step, while the cooling
    MUST be based on T at the beginning of the time step.

    JMP: 04/02/15 - Original version
*/
float radiate_heatcool(float pa[], float dt)
{

  double gamm = g-1.0;
  void (*derivs)(double, double*, double*);       // returns values for the function and its derivative
  void (*integ) (double y[], double dydx[], int n, double *x, double htry, double eps, double yscal[], double *hdid, double *hnext, void (*derivs)(double, double*, double*)); // function pointer to the desired integrator 

  derivs = heatcool;          // set function pointer to the correct derivative function
  integ = rkqc;               // set function pointer to the desired integrator (rkqc, stiff or stifbs)

  if (first_radiate){
    cout << "First radiate_heatcool RK4\n";
    readin_cooling_curve();
    first_radiate = false;
  }

  // store values in file-scope variables
  rho = pa[iqd]*rhoscale; 
  pre = pa[iqe]*pscale;  
  temp = usrpdtmp(pre/rho);
  mu = usrpdmu(pre/rho);

  double inte = pre/gamm; // internal energy density

  int nequns = 1;
  double y[nequns];                  // initial starting values of dependent variables
  y[0] = inte;                       // initial internal energy
#ifdef INTERMEDIATE_STORAGE  
  yp[0][0]  = inte;
#endif
  
  double lower_bound = 0.0;// initial start time
  double upper_bound = dt*tscale;// desired end time
  double eps = 4.0e-3;    // desired accuracy (default)
  //double eps = 1.0e-5;    // desired accuracy
  //double h = 1.0e9;       // initial guess at stepsize 
  double hmin = 1.0e-18;  // minimum allowed stepsize
  int num_ok = 0; int num_bad = 0;// number of OK and bad steps taken
  int nsteps;             // steps stored 
  int* nok = &num_ok; int* nbad = &num_bad; 

  // Initial guess at stepsize 
  double lambda_cool = calc_lambda(temp);
  double t_cool = inte/edot(rho,lambda_cool,gammaheat); // single point cooling/heating time (NEGATIVE value indicates cooling).
  double h = 0.1*fabs(t_cool);

  // Set Edot to zero if T is near Tmin and cooling, or is near Tmax and heating...
  if ((temp < 1.01*tmin)&&(t_cool < 0.0)){// if near Tmin and is cooling...
#ifdef DEBUG
    cout << "radiate_heatcool: T very near Tmin - jumping out\n";
    cout << "T = " << temp << "; Tmin = " << tmin << "; rho = " << rho << "; avgmass/mH = " << mu << "; tcool = " << t_cool << "\n"; 
    quit();
#endif
    return (0.0); 
  }
  else if ((temp > 0.99*tmax)&&(t_cool > 0.0)){ // if near Tmax and is heating...
#ifdef DEBUG
    cout << "radiate_heatcool: T very near Tmax - jumping out\n";
    double rho_mH = rho/massh;
    double coolterm = rho_mH*rho_mH*lambda_cool;
    double heatterm = rho_mH*gammaheat;
    cout << "T = " << temp << "; Tmax = " << tmax << "; rho = " << rho << "; avgmass/mH = " << mu << "; tcool = " << t_cool << "; lambda_cool = " << lambda_cool << "; edotcool = " << coolterm << "; edotheat = " << heatterm << "\n"; 
    quit();
#endif
    return (0.0); 
  }
  else{
    // Integrate from lower_bound to upper_bound. The final result is returned in y.
    odeint(y,nequns,lower_bound,upper_bound,eps,&h,hmin,nok,nbad,derivs,integ); 

    nsteps = num_ok + num_bad;
    //cout << " Number of good steps = " << num_ok << "; number of fixed/adjusted steps = " << num_bad << "; total number of steps = " << nsteps << "\n";
    nsteps = min(nsteps+1,kmax); // add the initial state to the number of steps taken

    double inte_final = y[0];
    double e_dot = (inte_final - inte)/(dt*tscale);

    // Make edot dimensionless and return
    e_dot *= rscale/(pscale*uscale);
    return float(e_dot);
  }
}


// Return the net heating/cooling rate (negative value indicates cooling) (erg/s/cm^3)
// (rho/mH) is used to obtain the cooling/heating rates (see Koyama & Inutsuka 2002) and my MEKAL notes.
inline double edot(double rho, double lambda_cool, double gamma_heat){

  double rho_mH = rho/massh;
  double coolterm = rho_mH*rho_mH*lambda_cool;
  double heatterm = rho_mH*gamma_heat;
  //if (temp > 1.0e7) heatterm = 0.0; // prevent excessive heating for low density hot gas

  return heatterm - coolterm;
}

// Return the internal energy density. Uses the file-scope value of mu. 
inline double eint(double rho, double temp){
  return rho*kboltz*temp/((g-1.0)*mu*massh);
}


// Given the temperature, determine the value of lambda for the cooling rate.
// Perform a simple interpolation.
double calc_lambda(double temp){

  double lambda_cool;
  double logtemp = log10(temp);
  int kk = (logtemp-logtmin)/dlogt;
  if (kk > ntmax-1) lambda_cool = lambda[ntmax-1]; 
  else if (kk < 0)  lambda_cool = lambda[0];
  else{
    // temperature lies between logt[kk] and logt[kk+1]
    double grad = (loglambda[kk+1] - loglambda[kk])/dlogt; 
    lambda_cool = pow(10,loglambda[kk] + grad*(logtemp - logt[kk]));
  }
  return lambda_cool;
}


void readin_cooling_curve(){

  // Read in cooling curve(s). It is assumed that the temperature binning is 
  // uniform in log space i.e. dlog(T) is a constant 

  ifstream file(coolcurve1); 
  if (!file) {
    if (mype == 0) cerr << "readin_cooling_curve: failed to open file " << coolcurve1 << "\n";
    quit();
  } 
  if (mype == 0) cout << "Reading in cooling curve " << coolcurve1 << "\n"; 
  for (int n = 0; n < ntmax; n++){
    //cout << "n = " << n << "\n";
    file >> logt[n] >> lambda[n];
    te[n] = pow(10,logt[n]);
    loglambda[n] = log10(lambda[n]);
  }
  file.close();

  tmin = te[0];        // K (do not allow cooling below 10 K in the Koyama & Inutsuka (2002) scheme
  tmax = te[ntmax-1]; // K
  logtmax = logt[ntmax-1];
  // logtmin and dlogt must refer to the cooling curve from file, otherwise the calculations below
  // will be incorrect.
  logtmin = logt[0]; 
  dlogt = (logtmax-logtmin)/float(ntmax-1);
  if (mype == 0) cout << "tmin = " << tmin << " (K); tmax = " << tmax << " (K); logtmin = " << logtmin << "; dlogt = " << dlogt << "\n";

}


// Given x0 and function values in f, calculate the function derivates returning
// the values in fprime. Uses the file-scope values of mu and rho.
// Given eint (passed in f) this function calculates edot (returned in fprime). 
void heatcool(double x0, double* f, double* fprime){

  double eint = *f;

  // Calculate the current temperature from the current internal energy  
  double temp = eint*(g-1.0)*mu*massh/(rho*kboltz);

  // Calculate single point cooling/heating time (NEGATIVE value indicates cooling).
  // (rho/mH) is used to obtain the cooling/heating rates (see Koyama & Inutsuka 2002) and 
  // my MEKAL notes
  double rho_mH = rho/massh;
  double lambda_cool = calc_lambda(temp);
  double coolterm = rho_mH*rho_mH*lambda_cool;
  double heatterm = rho_mH*gammaheat;
  double edot = heatterm - coolterm;

  // If the temperature is at or above the range covered by the cooling table
  // do not let edot be positive (which would cause further heating)
  if (temp >= te[ntmax-1]) edot=min(edot,0.0);
  // Alternatively if the temperature is at or below the range covered by the cooling table
  // do not let edot be negative (which would cause further cooling)
  else if (temp <= te[0]) edot=max(edot,0.0);

  *fprime = edot;

  return;
}



// Driver with adaptive stepsize control. Integrate the starting values ystart[nvar] 
// from x1 to x2 with accuracy eps, storing intermediate results in kmax, kount, dxsav,
// xp and yp. h1 should be set as a guessed first stepsize, hmin as the minimum allowed 
// stepsize (can be zero). On output nok and nbad are the number of good and bad (but 
// retried and fixed) steps taken, and ystart is replaced by values at the end of the 
// integration interval. derivs() is the user-supplied function for calculating the RHS 
// derivative, while rkqc() is the name of the stepper function to be used.  
// From Numerical Recipes. JMP 22/12/14 
void odeint(double ystart[], int nvar, double x1, double x2, double eps, double* h1, double hmin, int* nok, int* nbad, void (*derivs)(double, double*, double*), void (*rkqc)(double*, double*, int, double*, double, double, double*, double*, double*, void (*derivs)(double, double*, double*)))
{

  // For radiative cooling, odeint integrates the internal energy density (= pre/gamm) as y.
  // File scope values of rho,pre,temp and mu exist...

  const int maxstp = 100000;
  const double Tiny = 1.0e-30;

  bool finalstep = false;
  double xsav=0.0;
  double x,hnext,hdid,h;

  // Create dynamic memory
  double* yscal = new double[nvar];
  double* y     = new double[nvar];
  double* dydx  = new double[nvar];

  //cout << "odeint:\n";
  //cout << "nvar = " << nvar << "; ystart[0] = " << ystart[0] << "; x1 = " << x1 << "; x2 = " << x2 << "; eps = " << eps << "; h1 = " << h1 << "; hmin = " << hmin << "\n";
  //cout << " nok = " << *nok << "; nbad = " << *nbad << "\n";
  //odeint(&y[0],1,x[0],upper_bound,eps,h,hmin,nok,nbad,derivs,rkqc);

  x=x1;
  h=(x2 > x1) ? fabs(*h1) : -fabs(*h1);
  *nok = (*nbad) = 0; kount = 0;
  for (int i=0;i<nvar;i++) y[i]=ystart[i];
  if (kmax > 0) xsav=x-dxsav*2.0; // Assures storage of first step

  for (int nstp=0;nstp<maxstp;nstp++){ // Take at most MAXSTP steps
    //cout << "nstp = " << nstp << "; x = " << x << "; y = " << y[0] << "\n";
    (*derivs)(x,y,dydx);
    for (int i=0;i<nvar;i++)
      // Scaling used to monitor accuracy. This general-purpose scaling can be modified if necessary. (see p711 of Numerical Recipes)
      // Appropriate scaling is absolutely necessary for stiff ODEs. 
      yscal[i]=fabs(y[i])+fabs(dydx[i]*h)+Tiny;
    //yscal[i]=max(1.0,fabs(y[i])); // Here the components of "C" are unity - see Eq. 16.6.20
#ifdef INTERMEDIATE_STORAGE
      if (fabs(x-xsav) > fabs(dxsav)){ // Store intermediate results...
	if (kount < kmax-2){           // ... as long as there is enough space... (the "kmax-2" leaves enough space to store the final step) 
          xp[++kount]=x;
	  for (int i=0;i<nvar;i++) yp[i][kount]=y[i];
          //cout << "kount = " << kount << "; x = " << x << "\n";
          xsav=x;
	}
      }
#endif
    if ((x+h-x2)*(x+h-x1) > 0.0) h=x2-x; // If stepsize can overshoot decrease
    (*rkqc)(y,dydx,nvar,&x,h,eps,yscal,&hdid,&hnext,derivs); // h is the stepsize initially tried, hdid is the stepsize actually done (which may be larger or smaller)
    if (hdid == h) ++(*nok); else ++(*nbad);
    if (nstp == maxstp - 1) finalstep = true;

#ifdef JMP_JUMPOUTS
    // Calculate the temperature from the current value of the internal energy (in y[0]).
    // Use the file-scope values of mu and rho...
    double tempc = y[0]*(g-1.0)*mu*massh/(rho*kboltz); // current temperature
    double lambda_cool = calc_lambda(tempc);
    double tcool = eint(rho,tempc)/edot(rho,lambda_cool,gammaheat);
    if (fabs(tcool) > 10.0*x2){
      //cout << "nstp = " << nstp << "; Jumping out!\n";
      finalstep = true;
    }
    if ((tempc < 1.01*tmin)&&(tcool < 0.0)){// if near Tmin and is cooling...
      y[0] = 1.01*eint(rho,tmin);
      finalstep = true;
#ifdef DEBUG
      cout << "T very near Tmin - jumping out\n";
      cout << "Initial T = " << temp << "; current T = " << tempc << "; Tmin = " << tmin << "; rho = " << rho << "; avgmass/mH = " << mu << "; tcool = " << tcool << "; t = " << x << "\n"; 
      quit();
#endif
    }
    if ((tempc > 0.99*tmax)&&(tcool > 0.0)){ // if near Tmax and is heating...
      y[0] = 0.99*eint(rho,tmax);
      finalstep = true;
#ifdef DEBUG
      cout << "T very near Tmax - jumping out\n";
      cout << "Initial T = " << temp << "; current T = " << tempc << "; Tmax = " << tmax << "; rho = " << rho << "; avgmass/mH = " << mu << "; tcool = " << tcool << "; t = " << x << "\n"; 
      quit();
#endif
    }
#endif

    if (((x-x2)*(x2-x1) >= 0.0)||finalstep){ // Are we done?
      for (int i=0;i<nvar;i++) ystart[i]=y[i];
#ifdef INTERMEDIATE_STORAGE
      if (kmax){ // Save final step
	xp[++kount]=x;
        for (int i=0;i<nvar;i++) yp[i][kount]=y[i];
      }
#endif
      *h1 = hdid; // Record final stepsize

      // Free dynamic memory
      delete[] dydx;
      delete[] y;
      delete[] yscal;

      return; // Normal exit
    }
    if (fabs(hnext) <= hmin){
      //str_error("Step size too small in ODEINT");
      cout << "Step size too small in ODEINT. Aborting!\n";
      return;
    }
    h=hnext;
  }
  str_error("Too many steps in routine ODEINT");
  //cout << "Reached maximum number of steps in ODEINT. Aborting!\n";
  return;

}



void str_error(string error_text)
// Error handler
{
  cout << "Run-time error: " << error_text << "\n";
  quit();
}



// Fifth-order Runge-Kutta step with monitoring of local truncation error to ensure accuracy and 
// adjust stepsize. Input are the dependent variable vector y[nvar] and its derivative dydx[nvar] 
// at the starting value of the independent variable x. Also input are the stepsize to be attempted 
// htry, the required accuracy eps, and the vector yscal[nvar] against which the error is scaled. 
// On output, y and x are replaced by their new values, hdid is the stepsize that was actually 
// accomplished, and hnext is the estimated next stepsize. derivs() is the user supplied function 
// that computes the RHS derivatives. From Numerical Recipes. JMP 22/12/14
// NOTE: This routine is slightly different from that in the FORTRAN version of Numerical Recipes,
// where rkqs() calls rkck(). However, it appears to perform perfectly well. 
void rkqc(double y[], double dydx[], int n, double* x, double htry, double eps, double yscal[], double* hdid, double* hnext, void (*derivs)(double, double*, double*))
{

  const double Pgrow = -0.20;
  const double Pshrnk = -0.25;
  const double Fcor = 1.0/15.0;
  const double Safety = 0.9;
  const double Errcon = 6.0e-4;

  double xsav,hh,h,temp,errmax;

  // Create dynamic memory
  double* dysav = new double[n];
  double*  ysav = new double[n];
  double* ytemp = new double[n];
	
  xsav=(*x);
  for (int i=0;i<n;i++){
    ysav[i]=y[i];
    dysav[i]=dydx[i];
  }
  h=htry;
  for (;;){
    hh=0.5*h;
    rk4(ysav,dysav,n,xsav,hh,ytemp,derivs);
    *x=xsav+hh;
    (*derivs)(*x,ytemp,dydx);
    rk4(ytemp,dydx,n,*x,hh,y,derivs);
    *x=xsav+h;
    if (*x == xsav) str_error("Step size too small in routine RKQC");
    rk4(ysav,dysav,n,xsav,h,ytemp,derivs);
    errmax=0.0;
    for (int i=0;i<n;i++){
      ytemp[i]=y[i]-ytemp[i];
      temp=fabs(ytemp[i]/yscal[i]);
      if (errmax < temp) errmax=temp;
    }
    errmax /= eps;
    if (errmax <= 1.0){
      *hdid=h;
      *hnext=(errmax > Errcon ? Safety*h*exp(Pgrow*log(errmax)) : 4.0*h);
      break;
    }
    h=Safety*h*exp(Pshrnk*log(errmax));
  }

  for (int i=0;i<n;i++) y[i] += ytemp[i]*Fcor;

  // Delete dynamic memory
  delete[] ytemp;
  delete[] dysav;
  delete[] ysav;

  return;
}



// Given values for the variables y[nvar] and their derivatives dydx[nvar] known at x, 
// use the 4th order Runge-Kutta method to advance the solution over an interval h and 
// return the incremented variables as yout[nvar], which need not be a distinct array 
// from y. The user supplies the function derivs(x,y,dydx), which returns derivatives 
// dydx at x. From Numerical Recipes, Section 16.1. JMP 22/12/14
void rk4(double y[], double dydx[], int n, double x, double h, double yout[], void (*derivs)(double, double*, double*))
{

  // Create dynamic memory arrays
  double* dym = new double[n];
  double* dyt = new double[n];
  double*  yt = new double[n];

  double hh=h*0.5;
  double h6=h/6.0;
  double xh=x+hh;
	
  for (int i=0;i<n;i++) yt[i]=y[i]+hh*dydx[i];
  
  (*derivs)(xh,yt,dyt);	
  for (int i=0;i<n;i++) yt[i]=y[i]+hh*dyt[i];
 
  (*derivs)(xh,yt,dym);
  for (int i=0;i<n;i++) {
    yt[i]=y[i]+h*dym[i];
    dym[i] += dyt[i];
  }

  (*derivs)(x+h,yt,dyt);
  for (int i=0;i<n;i++)
    yout[i]=y[i]+h6*(dydx[i]+dyt[i]+2.0*dym[i]);

  // Free dynamic memory
  delete[] dym;
  delete[] dyt;
  delete[] yt;

  return;
}



//file find_wind.c
/*
+ Determine whether the passed cell coords are in the wind of star 1 or
  star 2 in a CWB calculation
    Description :
       Returns variable wnd as either 1 (wind 1) or 2 (wind 2).
*/

int find_wind(float xc[], float d1_stag)
{
  int wnd;
  float xs1,ys1,zs1,xs2,ys2,zs2;
  float crx, cry, crz, cr1, cr2, cb, rb, locx, locy, locz, loca,
          mx1, mx2, mx3;

  // Calculate entirely on whether the z-coord is greater than the
  //   stagnation point coord, rounded to the nearest coarse cell
  /*if (xc[1] < 27.5) wnd = 1;
  else wnd = 2;
  return(wnd);*/

  // Load data from arrays
  xs1 = xpos[0];
  ys1 = ypos[0];
  zs1 = zpos[0];
  xs2 = xpos[1];
  ys2 = ypos[1];
  zs2 = zpos[1];

  // Determine distance of wind cell from star 2
  if (nd == 2){
    crx = xs2 - xc[0];
    cry = ys2 - xc[1];
    crz = 0.0;
    cr2 = sqrt(crx*crx + cry*cry);
  }
  else if (nd == 3){
    crx = xs2 - xc[0];
    cry = ys2 - xc[1];
    crz = zs2 - xc[2];
    cr2 = sqrt(crx*crx + cry*cry + crz*crz);
  }	
  // Calculate the components giving the equation for the line of centres
  // Since the stars are always in the plane z=0 only compute the x and y 
  // components and the magnitude, loca
  locx = xs1 - xs2;
  locy = ys1 - ys2;
  locz = 0.0;
  loca = dsep;
      /* Determine distance of wind cell from star 1 */
  if (nd == 2){
    crx = xs1 - xc[0];
    cry = ys1 - xc[1];
    crz = 0.0;
    cr1 = sqrt(crx*crx + cry*cry);
  }
  else if (nd == 3){
    crx = xs1 - xc[0];
    cry = ys1 - xc[1];
    crz = zs1 - xc[2];
    cr1 = sqrt(crx*crx + cry*cry + crz*crz);
  }	

  if ((cr2 > dsep) && (cr2 > cr1)){
     wnd = 1;
     return(wnd);
  }
  // Matrix elements
  mx1 = cry*locz - crz*locy;
  mx2 = crz*locx - crx*locz;
  mx3 = crx*locy - cry*locx;

  // cb = distance of wind cell from line of centres
  cb = sqrt(mx1*mx1 + mx2*mx2 + mx3*mx3)/loca;
  // rb = distance of point B on line of centres from star 1
  rb = sqrt(cr1*cr1 - cb*cb);
  // Now find point on mtm balance curve which has the same distance
  // along the line of centres from star 1 (ie y(1,n) = rb)

  if (rb > d1_stag){
    wnd = 2;
  }
  else{
    wnd = 1;
  }

  /*if (rb > cdz[1]){
    for (int n = 1; n <= nmaxcd-1; n++){
      if ((cdz[n] < rb)&&(cdz[n+1] >= rb)){
        if (cdr[n]>cb) wnd = 2;
        else wnd = 1;
      }
    }
  }
  else wnd = 1; */

  return(wnd);
  }
//end find_wind.c



//file orbit_calc.c
/*
+ Calculate the stellar orbits (positions and velocities of the stars)
    based on the model time
    Description : JMP 05/03/03  - Original version
                  t and xs1,ys1,zs1 etc are passed in
*/

void orbit_calc()
{
  float torbit,time_offset,phase,phi;
  float dsepx,dsepy,dsepz;
  float m1,m2,xs1,ys1,zs1,xs2,ys2,zs2,xv1,yv1,zv1,xv2,yv2,zv2;
  double E,dE,sii,coi,theta,ang;
  double rrel;    // radius vector
  double gamma;   // angle between velocity and radius vectors
  double Edot;    // rate of change of E
  double a1,a2;   // semi-major axis of barycentric orbits
  double M;       // effective orbit mass
  double v;
  double eccd;

  int ndsep;
  //int ncellsmap;
  float maxremap,maxremap1,maxremap2,drmin;

  eccd = double(ecc);

  if (period == 0.0){
    cout << "Orbital period is zero. Aborting!";
    exit(EXIT_FAILURE);
  }

  time_offset = period*phase_offset;

  torbit = t*tscale + time_offset;  // time in seconds
  phase = torbit/period;

  phi = 2.0*gkpi*phase;
  E = phi;           // first guess at eccentric anomaly

  // Now use Newton-Raphson solver to calculate E
  // (dsin, dcos, and dabs are double precision versions)

  dE = (phi-E+eccd*sin(E))/(1.0-eccd*cos(E));
  while (mabs(dE) > 1.0e-10){
    E = E + dE;
    dE = (phi-E+eccd*sin(E))/(1.0-eccd*cos(E));
  }
  
  sii = sqrt(1.0 - eccd*eccd)*sin(E)/(1.0 - eccd*cos(E));
  coi = (cos(E)-eccd)/(1.0-eccd*cos(E));
  theta = atan2(sii,coi);
  if (theta < 0.0) theta= 2.0*gkpi + theta;   // 0 < theta < 2pi

  rrel = (1.0 - eccd*cos(E));   // radius vector

  // First compute barycentric orbital position of mass m1

  //M = m2**3/(m1+m2)**2
  //write(*,*) 'M (Msol) = ',M/Msol
  //a1 = (G*M*P**2/(4.0*pi**2))**(1.0/3.0)
  m1 = mass[0];
  m2 = mass[1];
  M = (pow(m2,3)/pow(m1+m2,2));
  //cout << "M (Msol) = " << M << "\n";

  a1 = pow(gconstcgs*M*period*period/(4.0*gkpi*gkpi),1.0/3.0)*pow(msol,0.3333);
  //cout << "gconstcgs = " << gconstcgs << "\n";
  //cout << "period = " << period << "\n";
  //cout << "gkpi = " << gkpi << "\n";
  //cout << "a1 = " << a1 << "\n";
  //cout << "rrel = " << rrel << "\n";
  //exit(EXIT_FAILURE);

  xs1 = a1*rrel*cos(theta);
  ys1 = a1*rrel*sin(theta);

  xs1 = -xs1/rscale;   // Setup correctly for mg_g
  ys1 = -ys1/rscale;
  zs1 = 0.0;

  // Compute velocity of star 1

  v = sqrt(gconstcgs*M*(2.0/rrel/a1 - 1.0/a1));       // Appears correct

  // Compute angle between velocity and radius vectors

  gamma = gkpi/2.0 + acos(sqrt((1.0-eccd*eccd)/(rrel*(2.0-rrel))));

  if (theta <= gkpi) ang = gkpi-gamma+theta;
  else ang = theta+gamma;

  xv1 = -v*cos(ang)/uscale;
  yv1 = -v*sin(ang)/uscale;
  zv1 = 0.0;

  // Now compute barycentric orbital position of mass m2
  M = (pow(m1,3)/pow(m1+m2,2))*msol;
  a2 = pow(gconstcgs*M*period*period/(4.0*gkpi*gkpi),1.0/3.0);

  xs2 = a2*rrel*cos(theta);
  ys2 = a2*rrel*sin(theta);

  xs2 = xs2/rscale;   // Setup correctly for cobra
  ys2 = ys2/rscale;
  zs2 = 0.0;

  // Compute velocity of star 2

  v = sqrt(gconstcgs*M*(2.0/rrel/a2 - 1.0/a2));       // Appears correct

  // Compute angle between velocity and radius vectors

  gamma = gkpi/2.0 + acos(sqrt((1.0-eccd*eccd)/(rrel*(2.0-rrel))));

  if (theta <= gkpi) ang = gkpi-gamma+theta;
  else ang = theta+gamma;

  xv2 = v*cos(ang)/uscale;
  yv2 = v*sin(ang)/uscale;
  zv2 = 0.0;

  //cout << "t = " << t << ", E = " << E << ", xs1 = " << xs1 << ", ys1 = " << ys1 << "\n";

  //Update arrays
  xpos[0] = xs1;
  ypos[0] = ys1;
  zpos[0] = zs1;
  xpos[1] = xs2;
  ypos[1] = ys2;
  zpos[1] = zs2;
  xvel[0] = xv1;
  yvel[0] = yv1;
  zvel[0] = zv1;
  xvel[1] = xv2;
  yvel[1] = yv2;
  zvel[1] = zv2;

  dsepx = xpos[1]-xpos[0];
  dsepy = ypos[1]-ypos[0];
  dsepz = zpos[1]-zpos[0];

  //Update dsep and rob (needed for controlling refinement in usrqsol)
  dsep = sqrt(dsepx*dsepx + dsepy*dsepy + dsepz*dsepz);
  rob = dsep*sqrt(eta)/(1.0 + sqrt(eta));
  
  //ncellsmap is now a user defined parameter on the command line
  //ncellsmap = 4; //number of cells in remap radius
  //ncellsmap = 8; //number of cells in remap radius
  //ncellsmap = 16; //number of cells in remap radius
  
  maxremap1 = 0.4*(dsep-rob);  //update remap radius
  maxremap2 = 0.4*rob;
  maxremap = min(maxremap1,maxremap2);

  //Increase or decrease maximum number of levels as needed

  drmin = (xr[0]-xl[0])/float(nc[0])/pow(2.0,nlevs-1); //finest cell size
  remap1 = ncellsmap*drmin;  //adjust remap radius to region close to star
  remap2 = ncellsmap*drmin;  //so that orbital dynamics affects as much of
                              //the wind as possible

  ndsep = int(dsep/drmin);  //number of cells between stars

  if (ncellsmap*drmin > maxremap){ //add a new grid level
    if (nlevs == MAXLEV){
      cout << "Error : nlevs > MAXLEV" << '\n';
      report();
    }
    else
      inc_lev();
  }
  else if ((ncellsmap*drmin < 0.5*maxremap)&&(ndsep > 80)){ //delete finest grid
    if (nlevs == 1){
      cout << "Error : only one level" << '\n';
      report();
    }
    else
      dec_lev();
  }

  //updated values
  //ncellsmap = 8;
  drmin = (xr[0]-xl[0])/float(nc[0])/pow(2.0,nlevs-1); //finest cell size
  remap1 = ncellsmap*drmin;  //adjust remap radius to region close to star
  remap2 = ncellsmap*drmin;  //so that orbital dynamics affects as much of
                              //the wind as possible
  return;
}





//file map_wind.c
/*
+ Set up the initial stellar winds
    Description :
       It is able to handle both single and binary calculations, in 2D or 
       3D, and assumes instantaneous acceleration (note that a mass-loss 
       rate and terminal velocity must be defined). init declares whether
       this routine is being called for initialization (in which case the
       winds should be setup over the entire grid domain) or for remapping
       the base of the wind during the calculation. wnd declares which
       wind the cell is in.
       If the pa array is updated, the update flag is set true
*/

void map_wind(bool init, int wnd, float xc[], float pa[], bool &update)
{
  int n;
  float xs, ys, zs, xv, yv, zv, dx, dy, dz, fr, th, ph, arg,
    dens, vel, pres, ur, ux, uy, uz, remap, delx, dely, dist;
  float e;
  float avgm,twind,frs,mdots,vinfs;

  // initialize radii and angles to stars
  fr = 0.0;
  th = 0.0;
  ph = 0.0;

  //cout << "mapwind, remap1 = " << remap1 << "\n";

  //cout << "mapwind: z = " << xc[1] << " wnd = " << wnd << '\n';
  /*if (init==false){
    cout << "wnd = " << wnd;
    cout << remap1;
    cout << xs;
    cout << ys;
    exit(EXIT_FAILURE);
    }*/
  // load desired wind values
  if (wnd == 1){
    xs = xpos[0];
    ys = ypos[0];
    zs = zpos[0];
    xv = xvel[0];
    yv = yvel[0];
    zv = zvel[0];
    mdots = mdot[0];
    vinfs = vinf[0];
    remap = remap1;
    //calculate avgm: assumes complete ionization!
    avgm = massh/(2.0*massfracx[0] + 0.75*massfracy[0] + 0.5*massfracz[0]);
    twind = tempw[0];
  }
  else {
    xs = xpos[1];
    ys = ypos[1];
    zs = zpos[1];
    xv = xvel[1];
    yv = yvel[1];
    zv = zvel[1];
    mdots = mdot[1];
    vinfs = vinf[1];
    remap = remap2;
    avgm = massh/(2.0*massfracx[1] + 0.75*massfracy[1] + 0.5*massfracz[1]);
    twind = tempw[1];
  }
  dx = xc[0] - xs;
  dy = xc[1] - ys;
  if (nd == 3) dz = xc[2] - zs;

  /*  cout << ewnd1 << " " << ewnd2 << '\n';
      exit(EXIT_FAILURE); */
  /*  else{
    cout << "Invalid value of wnd:" << wnd << '\n';
    break;
    }*/
  // calculate the force radius and angles to the centre of this star
  //   For 2D, th_1 = asin(r/fr_1)
  if (nd == 2){
    fr = sqrt(dx*dx + dy*dy);  // units of rscale
    th = asin(dx/fr);
  }
  else if (nd == 3){
    fr = sqrt(dx*dx + dy*dy + dz*dz);
    arg = xc[2]/fr;
    arg = min(arg,1.0);
    arg = max(arg,-1.0);
    ph = asin(arg);
    arg = dx/(fr*cos(ph));
    arg = min(arg,1.0);
    arg = max(arg,-1.0);
    th = acos(arg);
  }
  frs = fr*rscale;
 
  // if initializing, or if not then inside the remap radius, map wind 
  //   into cell. init = 0 denotes false 
   
  if ((init != 0) || ((init == 0) && (fr <= remap))){

    //    if ((init == 0) && (fr <= remap)) cout << remap;
    if ((init !=0) && (fr > 4.0*remap)){
      //map on low density ambient medium to blow away
      update = true;
      //dens = 1.0e3*dfloor*rhoscale;
      //dens = 0.01;
      dens = 0.03*mdot[0]/(4.0*gkpi*frs*frs*vinf[0]);
      vel = 0.0;
      dens /= rhoscale;
      dens = max(0.1,dens);
      pres = dens*rhoscale*(kboltz/1.0e-24)*twind;
      pres /= pscale;
      ux = 0.0; 
      uy = 0.0;
      uz = 0.0;
      ur =0.0;
    }
    else{  
    update = true;
    dens = mdots/(4.0*gkpi*frs*frs*vinfs);
    //if (clump) dens = dens/(mcmw+1.0);//smooth wind is only part of mdot
    vel  = vinfs;
    pres = dens*(kboltz/avgm)*twind;
    dens /= rhoscale;
    vel  /= uscale;
    pres /= pscale;
    if (dens < dfloor) dens = dfloor;
 
    if (nd == 2){
      ur = mabs(vel*sin(th));
      uz = mabs(vel*cos(th));
      if (xc[1] < ys){
        uz = -uz;
      }
    }
    else if (nd == 3){
      ux = mabs(vel*cos(th)*cos(ph));
      uy = mabs(vel*sin(th)*cos(ph));
      uz = mabs(vel*sin(ph));
      if (xc[0] < xs) ux = -ux;
      if (xc[1] < ys) uy = -uy;
      if (xc[2] < zs) uz = -uz;
      ux += xv; //add orbital velocity components
      uy += yv;
      uz += zv;
    }
    }

    pa[iqd] = dens;
    pa[iqe] = pres;
    //if (inten) pa[iqie] = pres/(g-1.0);  //internal energy
    if (inten) pa[iqie] = pres/powf(dens,g);  //entropy  
    // cout << pa[iqe] << "\n"; 
    if (nd == 2){
      pa[iqu0+1] = ur;
      pa[iqu0+2] = uz;
    }
    else if (nd == 3){
      pa[iqu0+1] = ux;
      pa[iqu0+2] = uy;
      pa[iqu0+3] = uz;
    }
    //Distinguish between the winds using an advected scalar
    //Colour: wind1 = 1.0, wind2 = 0.0     (eta car)
    if (wnd == 1) {
      pa[iqal0+1] = 1.0;
      //pa[iqal0+2] = 0.0;
      //pa[iqal0+2+1] = massfracx[0];
      //pa[iqal0+2+2] = massfracy[0];
      //pa[iqal0+2+3] = massfracz[0];
    }
    else {
      pa[iqal0+1] = 0.0;
      //pa[iqal0+2] = 1.0;
      //pa[iqal0+2+1] = massfracx[1];
      //pa[iqal0+2+2] = massfracy[1];
      //pa[iqal0+2+3] = massfracz[1];
    }
    //cout << "Updated wind: x = " << xc[0] << ", y = " << xc[1] << "\n";
    //cout << remap1 <<"\n";


    //Distinguish between the clump and interclump using an advected scalar
    //Colour2: wind = 0.0, clump = 1.0
    //if (nal >= 2){
    //pa[iqal0+1] = 0.0;   //only uncomment this, not the if statement
      //}
    //if (nal >=2) pa[iqal0+2] = 0.0; //initialize scalar to filter primary wind compression
  } // initializing, or inside remap radius
  // map on clumps

  /*if ((init)&&(xc[1]<30.0)&&(xc[1]>25.0)){
    fr = xc[0] + ys2;  
    if (pa[iqal0+1]>0.5) { //wind1
      //pa[iqd] = 4.0*pa[iqd];
      pa[iqd] = 4.0*mdot1/(4.0*gkpi*fr*fr*vinf1);
      pa[iqe] = 1.0e3*pa[iqe];
      if (nal > 1) pa[iqal0+2] = 1.0;
    }
    else {                 //wind2
      pa[iqd] = 4.0*pa[iqd];
      //pa[iqd] = 4.0*mdot2/(4.0*gkpi*fr*fr*vinf2);
      pa[iqe] = 1.0e4*pa[iqe];
    }
    }*/

  /*
  if ((init)&&(clump)){ //only map on clumps if initializing
    //cout << "numc = " << numc << "\n";
    //cout << "clx[1] = " << clx[1] << "\n";
    //exit(EXIT_FAILURE);
    
  // First cycle over the clumps stored for wind 1 to see if 
  //   any cell lies within a clump radius       
      dx = xc[0] - xs1;
      dy = xc[1] - ys1;
      fr = sqrt(dx*dx + dy*dy);  //distance of cell from star 1
      for (n = 1; n <= numc; n++){
        delx = xc[0]-clx[n]; 
        dely = xc[1]-cly[n];
        dist = sqrt(delx*delx + dely*dely);//distance of cell from clump center
        if (dist <= fr*rcl){       //cell inside clump
          //cout << "dx = " << dx << " dy = " << dy << '\n';
          //cout << "delx = " << delx << " dely = " << dely << '\n';
          //exit(EXIT_FAILURE);
          pa[iqd] = pa[iqd]*ncl;  //increase wind density
          pa[iqal0+1] = 1.0; //=1 for clump sims
          //pa[iqal0+2] = 1.0; //=1 for clump sims
          //cout << "mapped clumps in wind 1\n";
          //update = true;
        }
      }
      // Now cycle of clumps in wind 2
      dx = xc[0] - xs2;
      dy = xc[1] - ys2;
      fr = sqrt(dx*dx + dy*dy);  //distance of cell from star 2
      for (n = 1; n <= numc2; n++){
        delx = xc[0]-clx2[n];
        dely = xc[1]-cly2[n];
        dist = sqrt(delx*delx + dely*dely);//distance of cell from clump center
        if (dist <= fr*rcl2){
          pa[iqd] = pa[iqd]*ncl;
          pa[iqal0+1] = 1.0; //=1 for clump sims
          //pa[iqal0+2] = 1.0;
          //cout << "mapped clumps in wind 2\n";
          //update = true;
        }
      }    
      //}
  }
  */

  return;
  }
//end map_wind.c  



//file usrmonitor.c
/*
+ usrmonitor - create files for monitoring the solution
     Description:
        This creates files for monitoring the solution.
*/
void usrmonitor()
{
  char statsfile[100],numc[5];
  int n,nf;
  bool xyz,rzp,zrp,rtp;

  FILE *fptr;

  mon = true; // turn on monitoring of statistics (this is not movies!)

  xyz = (strcmp(geom, "XYZ") == 0);
  rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
  zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
  rtp = (strcmp(geom, "RTP") == 0);

  if (strcmp(cas, "SHELL_BLOWOUT") == 0){
  if (mon){
    //for shell material
    strcpy(statsfile,prefx);
    strcat(statsfile,"shellstats.dat");
    fptr = fopen(statsfile,"w");
    if (fptr!=NULL){
      fprintf(fptr,"%s\n","#Monitoring of mass and mtm of shell");
      fprintf(fptr,"%s\n","#  Time        Time       Mass       Mass-HI       Mtm       Mtm-HI  "); 
      if (xyz && (nd == 2)){
        fprintf(fptr,"%s\n","# (mg_gt)      (yr)    (Msol/pc)    (Msol/pc)(Msol km/s/pc)(Msol km/s/pc)");
      }
      else{
        fprintf(fptr,"%s\n","# (mg_gt)      (yr)      (Msol)       (Msol)   (Msol km/s)(Msol km/s)");       
      }
      fclose(fptr);
    }
    //for flux through original hole
    strcpy(statsfile,prefx);
    strcat(statsfile,"fluxstats.dat");
    fptr = fopen(statsfile,"w");
    if (fptr!=NULL){
      fprintf(fptr,"%s\n","#Monitoring of flux through original shell hole(s)");
      fprintf(fptr,"%s\n","#  Time        Time     Mass flux    Mtm flux   Energy flux "); 
      if (xyz && (nd == 2)){
        fprintf(fptr,"%s\n","# (mg_gt)      (yr)   (Msol/yr/pc)(Msol km/s^2/pc)(erg/s/pc)");
      }
      else{
        fprintf(fptr,"%s\n","# (mg_gt)      (yr)     (Msol/yr)  (Msol km/s^2) (erg/s)");
      }
      fclose(fptr);
    }
  } //if mon

  }// SHELL_BLOWOUT

  if (strcmp(cas, "WIND") == 0){
  if (mon){
    for (n=1; n <= nstars; n++){//for each individual wind
      sprintf(numc,"%d",n);
      strcpy(statsfile,prefx);
      strcat(statsfile,numc);
      strcat(statsfile,"stats.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#Monitoring of mass, volume, and energy of wind");
        fprintf(fptr,"%s\n","#  Time        Time       Mass        Vol         KE         IE         TotE"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (yr)      (Msol)       (pc^3)   (1e30 ergs)(1e30 ergs)(1e30 ergs)");
        fclose(fptr);
      }
    }
    //for ambient medium
    strcpy(statsfile,prefx);
    strcat(statsfile,"ambstats.dat");
    fptr = fopen(statsfile,"w");
    if (fptr!=NULL){
      fprintf(fptr,"%s\n","#Monitoring of mass, volume, and energy of ambient medium");
      fprintf(fptr,"%s\n","#  Time        Time       Mass         Vol        KE         IE         TotE"); 
      fprintf(fptr,"%s\n","# (mg_gt)      (yr)      (Msol)       (pc^3)   (1e30 ergs)(1e30 ergs)(1e30 ergs)");
      fclose(fptr);
    }
  } //if mon
  }//case WIND

  else if ((strcmp(cas, "SINGLECLOUD") == 0)||(strcmp(cas, "MANYCLOUD") == 0)){
    if (mon){


      if (nal == 1){
      strcpy(statsfile,prefx);
      strcat(statsfile,"stats.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#CLOUD monitoring using colour discrimination");
        fprintf(fptr,"%s\n","#  Time        Time        Time         a          c           c/a         Mass       Volume      Max rho   Mean rho     Mean vz      dvz         dvr         circ   <z>	<r>"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)       (rc)        (rc)     (no units)   (M(t)/M(0)) (V(t)/V(0)    (rhoi)   (arb units)    (vb)        (vb)        (vb)       (/a0 vb)		");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats2.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#CLOUD monitoring using colour2 discrimination");
        fprintf(fptr,"%s\n","#  Time        Time        Time         a          c           c/a         Mass       Volume      Max rho   Mean rho     Mean vz      dvz         dvr   <z>	<r>"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)       (rc)        (rc)     (no units)   (M(t)/M(0)) (V(t)/V(0)    (rhoi)   (arb units)    (vb)        (vb)        (vb)		");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats3.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#CLOUD monitoring using colour3 discrimination");
        fprintf(fptr,"%s\n","#  Time        Time        Time         a          c           c/a         Mass       Volume      Max rho   Mean rho     Mean vz      dvz         dvr   <z>	<r>"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)       (rc)        (rc)     (no units)   (M(t)/M(0)) (V(t)/V(0)    (rhoi)   (arb units)    (vb)        (vb)        (vb)		");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats4.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#Monitoring of cloud energy");
        fprintf(fptr,"%s\n","#  Time        Time        Time      KE/Etot     IE/Etot     TE/Etot      Etot        TE"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)     (no units)  (no units)  (no units) (1e30 ergs)(1e30 ergs)");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats5.dat");
      fptr = fopen(statsfile,"w");
      //fptr = fopen("m10.0_dens1e3_adia_res64_stats.dat","w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#Monitoring of total energy and fractions on grid");
        fprintf(fptr,"%s\n","#  Time        Time        Time      KE/Etot     IE/Etot     TE/Etot      Etot        KE        IE         TE"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)     (no units)  (no units)  (no units) (1e30 ergs)(1e30 ergs)(1e30 ergs)(1e30 ergs)");
        fclose(fptr);
      }
      }
      else{ //many clouds

      //RA: shock front
      strcpy(statsfile, prefx);
      strcat(statsfile, "shock_stats.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL) {
	fprintf(fptr, "%s\n", "#SHOCK front position monitoring");
	fprintf(fptr, "%s\n", "# Time        Time        Time        xAve        xMin         xMax");
	fprintf(fptr, "%s\n", "# (mg_gt)      (tcc)       (yr) ");
	fclose(fptr);
      }

        for (n=2; n <= 5; n++){//for each block of clouds
        nf = 999 + n;
        sprintf(numc,"%d",nf);
        strcpy(statsfile,prefx);
        strcat(statsfile,"blk");
        strcat(statsfile,numc);
        strcat(statsfile,"_cld_stats.dat");
        fptr = fopen(statsfile,"w");
        if (fptr!=NULL){
          fprintf(fptr,"%s\n","#Monitoring of CLOUD properties in a BLOCK region");
          fprintf(fptr,"%s\n","#  Time        Time        Time         x          dx           Mass       Volume      Mean rho     Mean vx      dvx         dvy       dvz"); 
          fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)          (code units)     (M(t)/M(0)) (V(t)/V(0))    (arb units)    (vb)        (vb)        (vb)        (vb)");
          fclose(fptr);
        }

        strcpy(statsfile,prefx);
        strcat(statsfile,"blk");
        strcat(statsfile,numc);
        strcat(statsfile,"_core_stats.dat");
        fptr = fopen(statsfile,"w");
        if (fptr!=NULL){
          fprintf(fptr,"%s\n","#Monitoring of CORE properties in a BLOCK region");
          fprintf(fptr,"%s\n","#  Time        Time        Time         x          dx           Mass       Volume      Mean rho     Mean vx      dvx         dvy       dvz"); 
          fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)          (code units)     (M(t)/M(0)) (V(t)/V(0))    (arb units)    (vb)        (vb)        (vb)        (vb)");
          fclose(fptr);
        }
        } //block of clouds

        for (n=6; n <= nal; n++){//for each individually monitored cloud
          nf = 999 + n;
          sprintf(numc,"%d",nf);
          strcpy(statsfile,prefx);
          strcat(statsfile,"single");
          strcat(statsfile,numc);
          strcat(statsfile,"_cld_stats.dat");
          fptr = fopen(statsfile,"w");
          if (fptr!=NULL){
            fprintf(fptr,"%s\n","#Monitoring of individual CLOUD properties");
            fprintf(fptr,"%s\n","#  Time        Time        Time         x           y            z          dx          dy          dz         Mass       Volume      Mean rho     Mean vx      dvx         dvy        dvz"); 
            fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)                                (code units)                                 (M(t)/M(0))  (V(t)/V(0)  (arb units)    (vb)        (vb)        (vb)       (vb)");
            fclose(fptr);
          }

          sprintf(numc,"%d",nf);
          strcpy(statsfile,prefx);
          strcat(statsfile,"single");
          strcat(statsfile,numc);
          strcat(statsfile,"_core_stats.dat");
          fptr = fopen(statsfile,"w");
          if (fptr!=NULL){
            fprintf(fptr,"%s\n","#Monitoring of individual CORE properties");
            fprintf(fptr,"%s\n","#  Time        Time        Time         x           y            z          dx          dy          dz         Mass       Volume      Mean rho     Mean vx      dvx         dvy        dvz"); 
            fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)                                (code units)                                 (M(t)/M(0))  (V(t)/V(0)  (arb units)    (vb)        (vb)        (vb)       (vb)");
            fclose(fptr);
          }
        } //individual clouds

        //energies
        for (n=2; n <= nal; n++){//for each block of clouds and individual clouds
          nf = 999 + n;
          sprintf(numc,"%d",nf);
          strcpy(statsfile,prefx);
          strcat(statsfile,"energy");
          strcat(statsfile,numc);
          strcat(statsfile,"_cld_stats.dat");
          fptr = fopen(statsfile,"w");
          if (fptr!=NULL){
            fprintf(fptr,"%s\n","#Monitoring of cloud energy");
            fprintf(fptr,"%s\n","#  Time        Time        Time      KE/Etot     IE/Etot     TE/Etot      Etot        TE"); 
            fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)     (no units)  (no units)  (no units) (1e30 ergs)(1e30 ergs)");
            fclose(fptr);
          }
        }//loop over clouds and regions/blocks

        //Now the TOTAL energies on the ENTIRE grid!
        strcpy(statsfile,prefx);
        strcat(statsfile,"totalenergy_cld_stats.dat");
        fptr = fopen(statsfile,"w");
        if (fptr!=NULL){
          fprintf(fptr,"%s\n","#Monitoring of total energy and fractions on grid");
          fprintf(fptr,"%s\n","#  Time        Time        Time      KE/Etot     IE/Etot     TE/Etot      Etot        KE        IE         TE        circ"); 
          fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)     (no units)  (no units)  (no units) (1e30 ergs)(1e30 ergs)(1e30 ergs)(1e30 ergs)  (/vb)");
          fclose(fptr);
        }
      }//single or MANY CLOUD
    } //if mon

  }//case SINGLECLOUD or MANYCLOUD

  else if ((strcmp(cas, "FILAMENT") == 0)){
    if (mon){
      if (nal == 1){ // 1 filament only
      strcpy(statsfile,prefx);
      strcat(statsfile,"stats.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#FILAMENT monitoring using colour discrimination");
        fprintf(fptr,"%s\n","#  Time        Time        Time        xms        yms         zms         Mass       Volume      Max rho   Mean rho     Mean vx    Mean vy     Mean vz     dvx         dvy         dvz        dvr         circ   "); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)       (rc)        (rc)        (rc)      (M(t)/M(0)) (V(t)/V(0)    (rhoi)   (arb units)    (vb)        (vb)        (vb)      (vb)        (vb)        (vb)      (vb)        (/a0 vb)");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats2.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#FILAMENT monitoring using colour2 discrimination");
        fprintf(fptr,"%s\n","#  Time        Time        Time        xms        yms         zms         Mass       Volume      Max rho   Mean rho     Mean vx    Mean vy     Mean vz     dvx         dvy         dvz        dvr         circ   "); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)       (rc)        (rc)        (rc)      (M(t)/M(0)) (V(t)/V(0)    (rhoi)   (arb units)    (vb)        (vb)        (vb)      (vb)        (vb)        (vb)      (vb)        (/a0 vb)");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats3.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#FILAMENT monitoring using colour3 discrimination");
        fprintf(fptr,"%s\n","#  Time        Time        Time        xms        yms         zms         Mass       Volume      Max rho   Mean rho     Mean vx    Mean vy     Mean vz     dvx         dvy         dvz        dvr         circ   "); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)       (rc)        (rc)        (rc)      (M(t)/M(0)) (V(t)/V(0)    (rhoi)   (arb units)    (vb)        (vb)        (vb)      (vb)        (vb)        (vb)      (vb)        (/a0 vb)");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats4.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#Monitoring of FILAMENT energy");
        fprintf(fptr,"%s\n","#  Time        Time        Time      KE/Etot     IE/Etot     TE/Etot      Etot        TE"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)     (no units)  (no units)  (no units) (1e30 ergs)(1e30 ergs)");
        fclose(fptr);
      }

      strcpy(statsfile,prefx);
      strcat(statsfile,"stats5.dat");
      fptr = fopen(statsfile,"w");
      if (fptr!=NULL){
        fprintf(fptr,"%s\n","#Monitoring of total energy and fractions on grid");
        fprintf(fptr,"%s\n","#  Time        Time        Time      KE/Etot     IE/Etot     TE/Etot      Etot        KE        IE         TE"); 
        fprintf(fptr,"%s\n","# (mg_gt)      (tcc)       (yr)     (no units)  (no units)  (no units) (1e30 ergs)(1e30 ergs)(1e30 ergs)(1e30 ergs)");
        fclose(fptr);
      }
      }
    }
  }

  return;
}
//end usrmonitor.c


//file usrpdtmp.c
/*
+ usrpdtmp - returns the temperature (K) from p/rho
     Description:
        It uses a pre-tabulated table based on the allen1973.txt data 
        in Fyris Alpha.
*/

float usrpdtmp(float pdcell)
{
  int n;
  float grad,temp;
  bool loop;
       
  temp = 0.0;
  //set temp to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) temp = pow(10,pdlogt[0]);
  else if (pdcell > pd[npd-1]) temp = pow(10,pdlogt[npd-1]);
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
        grad = (pdlogt[n+1]-pdlogt[n])/(log10(pd[n+1])-log10(pd[n])); 
        temp = pow(10,pdlogt[n] + grad*(log10(pdcell) - log10(pd[n])));
        loop = false;
      }
      n++;
    }    
  }
  return temp;
}
//end usrpdtmp.c


//file usrpdmu.c
/*
+ usrpdmu - returns the average particle mass from p/rho
     Description:
        It uses a pre-tabulated table based on the allen1973.txt data 
        in Fyris Alpha.
*/

float usrpdmu(float pdcell)
{
  int n;
  float grad,avgm;
  bool loop;
         
  avgm = 0.0;
  //set avgm to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) avgm = pdmu[0];
  else if (pdcell > pd[npd-1]) avgm = pdmu[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
        grad = (pdmu[n+1]-pdmu[n])/(pd[n+1]-pd[n]); 
        avgm = pdmu[n] + grad*(pdcell - pd[n]);
        loop = false;
      }
      n++;
    }    
  }

  if (strcmp(cas, "SHELL_BLOWOUT") == 0){
    if (avgm > 1.29) avgm = 1.29; // see p7 of "Multiple Colliding Winds III" book
  }

  return avgm;
}
//end usrpdmu.c


//file usrpdg.c
/*
+ usrpdg - returns the ratio of specific heats (gamma) from p/rho
     Description:
        It uses an ad-hoc table noted on p175-176 of "Notes on Constructing
        a Hydro Code"
*/

float usrpdg(float pdcell)
{
  int n;
  float grad,gammasp;
  bool loop;

  //set gamma to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) gammasp = pdg[0];
  else if (pdcell > pd[npd-1]) gammasp = pdg[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
    if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
      grad = (pdg[n+1]-pdg[n])/(pd[n+1]-pd[n]); 
      gammasp = pdg[n] + grad*(pdcell - pd[n]);
      loop = false;
    }
    n++;
    }    
  }

  return gammasp;
}
//end usrpdg.c


//file usriedg.c
/*
+ usriedg - returns the ratio of specific heats (gamma) from ie/rho
     Description:
        It uses an ad-hoc table noted on p175-176 of "Notes on Constructing
        a Hydro Code"
*/

float usriedg(float iedcell)
{
  int n;
  float grad,gammasp;
  bool loop;

  //set gamma to lower or upper limits, or interpolate from the table
  if (iedcell < ied[0]) gammasp = pdg[0];
  else if (iedcell > ied[npd-1]) gammasp = pdg[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
    if ((iedcell >= ied[n])&&(iedcell < ied[n+1])){
      grad = (pdg[n+1]-pdg[n])/(ied[n+1]-ied[n]); 
      gammasp = pdg[n] + grad*(iedcell - ied[n]);
      loop = false;
    }
    n++;
    }    
  }
  return gammasp;
}
//end usriedg.c


//file usrpdh2.c
/*
+ usrpdh2 - returns the number fraction of H in H2
     Description:
        It uses an ad-hoc table noted on p175-176 of "Notes on Constructing
        a Hydro Code"
*/

float usrpdh2(float pdcell)
{
  int n;
  float grad,frach2;
  bool loop;
         
  frach2 = 0.0;
  //set frach2 to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) frach2 = pdh2[0];
  else if (pdcell > pd[npd-1]) frach2 = pdh2[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
        grad = (pdh2[n+1]-pdh2[n])/(pd[n+1]-pd[n]); 
        frach2 = pdh2[n] + grad*(pdcell - pd[n]);
        loop = false;
      }
      n++;
    }    
  }

  return frach2;
}
//end usrpdh2.c

//file usrpdhi.c
/*
+ usrpdhi - returns the number fraction of H in HI
     Description:
        It uses an ad-hoc table noted on p175-176 of "Notes on Constructing
        a Hydro Code"
*/

float usrpdhi(float pdcell)
{
  int n;
  float grad,frachi;
  bool loop;
         
  frachi = 0.0;
  //set frachi to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) frachi = pdhi[0];
  else if (pdcell > pd[npd-1]) frachi = pdhi[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
    if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
      grad = (pdhi[n+1]-pdhi[n])/(pd[n+1]-pd[n]); 
      frachi = pdhi[n] + grad*(pdcell - pd[n]);
      loop = false;
    }
    n++;
    }    
  }

  return frachi;
}
//end usrpdhi.c


//file usrpdhii.c
/*
+ usrpdhii - returns the number fraction of H in HII
     Description:
        It uses a pre-tabulated table based on the allen1973.txt data 
        in Fyris Alpha.
*/

float usrpdhii(float pdcell)
{
  int n;
  float grad,frachii;
  bool loop;
         
  frachii = 0.0;
  //set frachii to lower or upper limits, or interpolate from the table
  if (pdcell < pd[0]) frachii = pdhii[0];
  else if (pdcell > pd[npd-1]) frachii = pdhii[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
    if ((pdcell >= pd[n])&&(pdcell < pd[n+1])){
      grad = (pdhii[n+1]-pdhii[n])/(pd[n+1]-pd[n]); 
      frachii = pdhii[n] + grad*(pdcell - pd[n]);
      loop = false;
    }
    n++;
    }    
  }

  return frachii;
}
//end usrpdhii.c




//file usrtmpmu.c
/*
+ usrpdtmp - returns the value of mu from the temperature (K)
     Description:
        It uses a pre-tabulated table based on the allen1973.txt data 
        in Fyris Alpha.
*/

float usrtmpmu(float temp)
{
  int n;
  float grad,mu,logtemp;
  bool loop;
         
  mu = 0.0;
  logtemp = log10(temp);

  //set mu to lower or upper limits, or interpolate from the table
  if (logtemp < pdlogt[0]) mu = pdmu[0];
  else if (logtemp > pdlogt[npd-1]) mu = pdmu[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((logtemp >= pdlogt[n])&&(logtemp < pdlogt[n+1])){
        grad = (log10(pdmu[n+1])-log10(pdmu[n]))/(pdlogt[n+1]-pdlogt[n]); 
        mu = pow(10,log10(pdmu[n]) + grad*(logtemp - pdlogt[n]));
        loop = false;
      }
    n++;
    }    
  }

  if (strcmp(cas, "SHELL_BLOWOUT") == 0){
    if (mu > 1.29) mu = 1.29; // see p7 of "Multiple Colliding Winds III" book
  }

  return mu;
}
//end usrtmpmu.c


//file usrtmpg.c
/*
+ usrpdtmp - returns the value of g from the temperature (K)
     Description:
        It uses a pre-tabulated table based on the allen1973.txt data 
        in Fyris Alpha.
*/

float usrtmpg(float temp)
{
  int n;
  float grad,gam,logtemp;
  bool loop;
         
  gam = 0.0;
  logtemp = log10(temp);

  //set gam to lower or upper limits, or interpolate from the table
  if (logtemp < pdlogt[0]) gam = pdg[0];
  else if (logtemp > pdlogt[npd-1]) gam = pdg[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((logtemp >= pdlogt[n])&&(logtemp < pdlogt[n+1])){
        grad = (log10(pdg[n+1])-log10(pdg[n]))/(pdlogt[n+1]-pdlogt[n]); 
        gam = pow(10,log10(pdg[n]) + grad*(logtemp - pdlogt[n]));
        loop = false;
      }
    n++;
    }    
  }

  return gam;
}
//end usrtmpg.c


//file usrteqa.c
/*
+ usrteqa - Temperature calculated from qa.
*/
float usrteqa(float qa[])
{
  int id;
  float rho,pre,te,ke,pdfloor;

  rho = qa[iqd];

  //if (inten){
    //qa[iqie] stores rho.s = rho.(p/rho^g) = p/(rho^(g-1))
    //ALWAYS use pressure from entropy (since entropy is properly reset in
    //usrfxq.c at the end of each step - JMP 04/02/11)
    //pre = qa[iqie]*powf(rho, g-1.0); 
  //}
  //else{
  //ALWAYS calculate p from qa[iqe] since it should be fixed up in usrfxq.c
    ke = 0.0;
    for (id = 1; id <= nd; id++){
      ke += qa[iqu0+id]*qa[iqu0+id];
    }
    ke *= 0.5/rho;
  
    pre = (g-1.0)*(qa[iqe] - ke);
    //}

  te = pre/rho;
 
 //Check for NANs or infinities //JMP 08/02/11
  if ( isinf( te )){
    // value is infinity
    //cout << "usrteqa.c te is infinity\n";
    //exit(EXIT_FAILURE);
    te = pfloor/dfloor;
    if (tempmu) te = pd[0]/(pscale/rhoscale);
    return(te);
  }

  if ( isnan( te )){
    // value is not a number
    //cout << "usrteqa.c te is NAN\n";
    //exit(EXIT_FAILURE);
    te = pfloor/dfloor;
    if (tempmu) te = pd[0]/(pscale/rhoscale);
    return(te);
  }

  //set minimum te to a floor temperature 
  if (tempmu) pdfloor = pd[0]/(pscale/rhoscale);
  else{
    pdfloor = 1.0*kboltz/(2.363*massh)/(pscale/rhoscale); //(1.0K, with avgmass = 2.363)
  }
  te = max(te,pdfloor);
  //cout << "pdfloor = " << pdfloor*pscale/rhoscale <<"\n";
  //cout << "pd[0] = " << pd[0] << "\n";
  //exit(EXIT_FAILURE);

  //Check for zero or negative temperature //JMP 22/02/11
  if (te == 0.0){
    cout << "te in usrteqa.c is zero. Aborting!\n";
    cout << "rho = " << rho << "; ke = " << ke << "; pre = " << pre << "; qa[iqe] = " << qa[iqe] << "\n";
    exit(EXIT_FAILURE);
  }
  else if (te < 0.0){
    cout << "te in usrteqa.c is negative. Aborting!\n";
    cout << "rho = " << rho << "; ke = " << ke << "; pre = " << pre << "; qa[iqe] = " << qa[iqe] << "\n";
    exit(EXIT_FAILURE);
  }

  return(te);
}
//end usrteqa.c


//file usriedqa.c
/*
+ usriedqa - internal energy over density calculated from qa.
*/
float usriedqa(float qa[])
{
  int id;
  float rho,pre,ke,ie,ied;
  rho = qa[iqd];
  ke = 0.0;
  for (id = 1; id <= nd; id++){
    ke += qa[iqu0+id]*qa[iqu0+id];
  }
  ke *= 0.5/qa[iqd];
  
  ie = qa[iqe] - ke;
  
  /*
  if (inten){
    //qa[iqie] stores rho.s = rho.(p/rho^g) = p/(rho^(g-1))
    //if (pre < 0.03*ke) 
    pre = qa[iqie]*powf(qa[iqd], g-1.0); //use conserved entropy variable instead
    ie 
  }
  */

  ied = ie/rho;
  return(ied);
}
//end usriedqa.c


  
//file usriedtmp.c
/*
+ usriedtmp - returns the temperature (K) from ie/rho
     Description:
        It uses a pre-tabulated table based on the allen1973.txt data 
        in Fyris Alpha.
*/

float usriedtmp(float iedcell)
{
  int n;
  float grad,temp;
  bool loop;
       
  temp = 0.0;
  //set temp to lower or upper limits, or interpolate from the table
  if (iedcell < ied[0]) temp = pow(10,pdlogt[0]);
  else if (iedcell > ied[npd-1]) temp = pow(10,pdlogt[npd-1]);
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((iedcell >= ied[n])&&(iedcell < ied[n+1])){
        grad = (pdlogt[n+1]-pdlogt[n])/(log10(ied[n+1])-log10(ied[n])); 
        temp = pow(10,pdlogt[n] + grad*(log10(iedcell) - log10(ied[n])));
        loop = false;
      }
      n++;
    }    
  }
  return temp;
}
//end usriedtmp.c


//file usriedmu.c
/*
+ usriedmu - returns the average particle mass from ie/rho
     Description:
        It uses a pre-tabulated table based on the allen1973.txt data 
        in Fyris Alpha.
*/

float usriedmu(float iedcell)
{
  int n;
  float grad,avgm;
  bool loop;
         
  avgm = 0.0;
  //set avgm to lower or upper limits, or interpolate from the table
  if (iedcell < ied[0]) avgm = pdmu[0];
  else if (iedcell > ied[npd-1]) avgm = pdmu[npd-1];
  else{
    n = 0;
    loop = true;
    while (loop){
      if ((iedcell >= ied[n])&&(iedcell < ied[n+1])){
        grad = (pdmu[n+1]-pdmu[n])/(ied[n+1]-ied[n]); 
        avgm = pdmu[n] + grad*(iedcell - ied[n]);
        loop = false;
      }
      n++;
    }    
  }

  if (strcmp(cas, "SHELL_BLOWOUT") == 0){
    if (avgm > 1.29) avgm = 1.29; // see p7 of "Multiple Colliding Winds III" book
  }

  return avgm;
}
//end usriedmu.c


void quit()
{

  exit(EXIT_SUCCESS);
}


void read_tempmu_data(){
  ifstream file("tempdependent_mu_and_gamma.dat");
  if (!file) {
    cerr << "Failed to open file tempdependent_mu_and_gamma.dat\n";
    exit(EXIT_FAILURE);
  }
  file.seekg(0); //goto start of file
  //file >> header; 
  // read data.  Format of file is:
  //logt, mu, gamma, f(H2), f(HI), f(HII), p/rho
  npd = 91;
  for (int n=0; n < npd; n++){
    file >> pdlogt[n] >> pdmu[n] >> pdg[n] >> pdh2[n] >> pdhi[n] >> pdhii[n] >> pd[n] >> ied[n];
    //cout << "n = " << n << "; pdlogt = " << pdlogt[n] << "; pd = " << pd[n] << "; pdmu = " << pdmu[n] << "; pdg = " << pdg[n] << "; pdh2 = " << pdh2[n] << "; pdhi = " << pdhi[n] << "; pdhii = " << pdhii[n] << "; ied = " << ied[n] <<"\n";
  }
  file.close();
  //if (mype==0) cout << "Read in temperature dependent mu data...\n";      
  return;
}


// Calculate the density and pressure jump for an adiabatic shock
// given the Mach number and the ratio of specific heats
void adiabatic_shk_jump(float mach, float& jumpr, float& jumpp){

  float machsq = mach*mach;
  float gp1 = g+1.0;
  float gm1 = g-1.0;

  jumpr = gp1*machsq/(gm1*machsq + 2.0); // density jump
  jumpp = (2.0*g*machsq - gm1)/gp1;      // pressure jump
  return;
}


// Calculate the density and pressure jump for an isothermal shock
// given the ADIABATIC Mach number and the ratio of specific heats
void isothermal_shk_jump(float mach, float& jumpr, float& jumpp){

  // Here mach is the adiabatic mach number, M_A.
  // The compression is also = (M_0)^2, where M_O is the isothermal mach number (M_O^2 = g M_A^2)
  jumpr = g*mach*mach;
  jumpp = jumpr;
  return;
}


// Returns the density and colour of a cell near a cloud with a soft edge.
// The cell position, the ambient density, the density contrast,
// and a fudge factor (usually = 1.0) are specified.
void cloud_soft_edge(float xc[], float ramb, float chi, float rhofudge, float& rho, float& col){

  float r2 = 0.0; // radius squared from cloud centre
  for (int n = 0; n < nd; n++) r2 += xc[n]*xc[n];

  float tv = r2 - 1.0;
  //tv = 1.0e3;
  tv = exp((double)min(cld1,cld2*tv));
  tv = 0.5*(1.0 + (tv - 1.0)/(tv + 1.0));
  rho = chi*rhofudge*ramb + (ramb - chi*rhofudge*ramb)*tv;

  col = rho/(chi*ramb);    //soft edge for colour

  //if (r2 <= 1.0)   col = 1.0;  //hard edge for interior colour
  //if (r2 > 5.0625) col = 1.0;  //soft edge limit for colour
  //if (r2 > 9.0)    col = 1.0;  //soft edge limit for colour for chi=1e5

  return;
}



// Calculate values for the pre-shock primitive array, pa. The pre-shock
// density and pressure are passed in as ramb and ambp, and the preshock 
// primitives returned as pa. The ambient medium is assumed to have zero velocity
// on the grid. The ambient sound speed is also returned.
// Two geometries are currently implemented:
//  1) xyz - shock propagating towards positive x
//  2) rzp - shock propagating towards negative z
// JMP 25/02/15 - Original version 
void setup_preshock_pa(float pa[], float ramb, float ambp, float& cs){

  // Geometry flags
  bool xyz = (strcmp(geom, "XYZ") == 0);
  bool rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
  bool zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
  bool rtp = (strcmp(geom, "RTP") == 0);

  // Preshock values
  pa[iqd] = ramb; // preshock density
  pa[iqe] = ambp; // preshock pressure
  cs = sqrt(g*ambp/ramb);  

  if (keps){
    pa[iqk] = kmin; //kmin
    pa[iqeps] = epsmin;  //epsmin
  }

  //cout << "ambp = " << ambp << "\n";
  //quit();

  return;
}


// Calculate values for the post-shock primitive array. The pre-shock
// primitives are passed in as pa, and the postshock primitives
// returned as pa. The ambient medium is assumed to have zero velocity
// on the grid.
// Two geometries are currently implemented:
//  1) xyz - shock propagating towards positive x
//  2) rzp - shock propagating towards negative z
// JMP 25/02/15 - Original version 
void setup_postshock_pa(float pa[], float cs, float mach){

  float jumpr,jumpp;
  float vs = mach*cs; //shock speed

  // Geometry flags
  bool xyz = (strcmp(geom, "XYZ") == 0);
  bool rzp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[0], "R") == 0));
  bool zrp = ((strcmp(geom, "RZP") == 0) && (strcmp(coord[1], "R") == 0));
  bool rtp = (strcmp(geom, "RTP") == 0);

  // Preshock values
  float ramb = pa[iqd]; // preshock density
  float pamb = pa[iqe]; // preshock pressure


  adiabatic_shk_jump (mach,jumpr,jumpp);  //postshock quantities for adiabatic shocks
  //isothermal_shk_jump(mach,jumpr,jumpp);  //postshock quantities for isothermal shocks
  
  pa[iqd] = ramb*jumpr;    //postshock rho
  pa[iqe] = pamb*jumpp;    //postshock pressure
  if (xyz){
    pa[iqu0+1] = (1.0 - 1.0/jumpr)*mach*cs; //conservation of mass flux (XYZ)
  }
  else if (rzp){
    pa[iqu0+2] = -(1.0 - 1.0/jumpr)*mach*cs; //conservation of mass flux(RZ)
  }
  else{
    if (mype == 0) cout << "setup_postshock_pa: Not coded for this geometry yet! Aborting!\n";
    quit();
  }
  //cout << mach*cs/uscale;
  //exit(EXIT_FAILURE);
  pa[iqal0+1] = 0.0;

  return;
}


// Sets up the filament. Given a cell position xc[], and the key filament 
// parameters (orientation, central position, density contrast), determine
// if the cell lies within the filament and set appropriate primitive values
// (returned in pa).
// JMP - Original version 25/02/15 
void setup_filament(float xc[], float fil_theta, float fil_phi, float rx0, float ry0, float rz0, float ramb, float chi, float pa[]){

  // change angles to radians
  float fil_theta_rad = fil_theta*gkpi/180.0;
  float fil_phi_rad   = fil_phi  *gkpi/180.0;

  // Calculate filament direction vector
  float a = cos(fil_phi_rad)*sin(fil_theta_rad);            // x-direction
  float b = sin(fil_phi_rad)*sin(fil_theta_rad);            // y-direction
  float c =                  cos(fil_theta_rad);            // z-direction

  // Now determine whether the cell lies within the filament and adjust its parameters appropriately
  // Distance of a point (xc[0],xc[1],xc[2]) from a line (specified in vector form as r = r0 + tA where t is 
  // a constant and A is a direction vector with components (a,b,c)). r0 is the coord of the centre of the filament.
  float rpx = xc[0] - rx0;
  float rpy = xc[1] - ry0;
  float rpz = xc[2] - rz0;
  float x = rpy*c - rpz*b;
  float y = -(rpx*c - rpz*a);
  float z = rpx*b - rpy*a;
  float dist = sqrt(x*x + y*y + z*z); // radial distance from filament symmetry axis
  float dist2 = sqrt(rpx*rpx + rpy*rpy + rpz*rpz); // distance from centre of filament
  float dist3 = sqrt(dist2*dist2 - dist*dist);     // projected distance along symmetry axis from centre of filament
  float dist4 = dist3 - fil_length/2.0;            // projected distance beyond filament
  float distr = sqrt(dist*dist + dist4*dist4);     // spherical distance from centre of filament cap
  bool inside_fil = false;
  // Give filament a soft edge up to 1.5x its radius
  if (dist <= 1.5*fil_rad){
    if (dist3 <= fil_length/2.0){
      inside_fil = true;
    }
    else if ((dist3 <= fil_length/2.0 + 1.5*fil_rad)&&(distr <= 1.5*fil_rad)){ // allow a "smooth cap/end" to the filament
      inside_fil = true;
      dist = distr; //spherical distance from centre of filament cap
    }
    if (inside_fil){
      float tv = dist*dist/(fil_rad*fil_rad) - 1.0;
      tv = exp((double)min(cld1,float(cld2*tv)));
      tv = 0.5*(1.0 + (tv - 1.0)/(tv + 1.0));
      pa[iqd] = chi*ramb + (ramb - chi*ramb)*tv;
      //soft edge for colour
      pa[iqal0+1] = pa[iqd]/(chi*ramb);    //soft edge for colour
     
      // internal flow
      //pa[iqu0+1] = ufil;
      //pa[iqu0+2] = vfil;
      //pa[iqu0+3] = wfil;
    }
  }          

  return;
}
